/*
 * Generated by gdbus-codegen 2.39.0. DO NOT EDIT.
 *
 * The license of this code is the same as for the source it was derived from.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "gvfsdbus.h"

#include <string.h>
#ifdef G_OS_UNIX
#  include <gio/gunixfdlist.h>
#endif

typedef struct
{
  GDBusArgInfo parent_struct;
  gboolean use_gvariant;
} _ExtendedGDBusArgInfo;

typedef struct
{
  GDBusMethodInfo parent_struct;
  const gchar *signal_name;
  gboolean pass_fdlist;
} _ExtendedGDBusMethodInfo;

typedef struct
{
  GDBusSignalInfo parent_struct;
  const gchar *signal_name;
} _ExtendedGDBusSignalInfo;

typedef struct
{
  GDBusPropertyInfo parent_struct;
  const gchar *hyphen_name;
  gboolean use_gvariant;
} _ExtendedGDBusPropertyInfo;

typedef struct
{
  GDBusInterfaceInfo parent_struct;
  const gchar *hyphen_name;
} _ExtendedGDBusInterfaceInfo;

typedef struct
{
  const _ExtendedGDBusPropertyInfo *info;
  guint prop_id;
  GValue orig_value; /* the value before the change */
} ChangedProperty;

static void
_changed_property_free (ChangedProperty *data)
{
  g_value_unset (&data->orig_value);
  g_free (data);
}

static gboolean
_g_strv_equal0 (gchar **a, gchar **b)
{
  gboolean ret = FALSE;
  guint n;
  if (a == NULL && b == NULL)
    {
      ret = TRUE;
      goto out;
    }
  if (a == NULL || b == NULL)
    goto out;
  if (g_strv_length (a) != g_strv_length (b))
    goto out;
  for (n = 0; a[n] != NULL; n++)
    if (g_strcmp0 (a[n], b[n]) != 0)
      goto out;
  ret = TRUE;
out:
  return ret;
}

static gboolean
_g_variant_equal0 (GVariant *a, GVariant *b)
{
  gboolean ret = FALSE;
  if (a == NULL && b == NULL)
    {
      ret = TRUE;
      goto out;
    }
  if (a == NULL || b == NULL)
    goto out;
  ret = g_variant_equal (a, b);
out:
  return ret;
}

G_GNUC_UNUSED static gboolean
_g_value_equal (const GValue *a, const GValue *b)
{
  gboolean ret = FALSE;
  g_assert (G_VALUE_TYPE (a) == G_VALUE_TYPE (b));
  switch (G_VALUE_TYPE (a))
    {
      case G_TYPE_BOOLEAN:
        ret = (g_value_get_boolean (a) == g_value_get_boolean (b));
        break;
      case G_TYPE_UCHAR:
        ret = (g_value_get_uchar (a) == g_value_get_uchar (b));
        break;
      case G_TYPE_INT:
        ret = (g_value_get_int (a) == g_value_get_int (b));
        break;
      case G_TYPE_UINT:
        ret = (g_value_get_uint (a) == g_value_get_uint (b));
        break;
      case G_TYPE_INT64:
        ret = (g_value_get_int64 (a) == g_value_get_int64 (b));
        break;
      case G_TYPE_UINT64:
        ret = (g_value_get_uint64 (a) == g_value_get_uint64 (b));
        break;
      case G_TYPE_DOUBLE:
        {
          /* Avoid -Wfloat-equal warnings by doing a direct bit compare */
          gdouble da = g_value_get_double (a);
          gdouble db = g_value_get_double (b);
          ret = memcmp (&da, &db, sizeof (gdouble)) == 0;
        }
        break;
      case G_TYPE_STRING:
        ret = (g_strcmp0 (g_value_get_string (a), g_value_get_string (b)) == 0);
        break;
      case G_TYPE_VARIANT:
        ret = _g_variant_equal0 (g_value_get_variant (a), g_value_get_variant (b));
        break;
      default:
        if (G_VALUE_TYPE (a) == G_TYPE_STRV)
          ret = _g_strv_equal0 (g_value_get_boxed (a), g_value_get_boxed (b));
        else
          g_critical ("_g_value_equal() does not handle type %s", g_type_name (G_VALUE_TYPE (a)));
        break;
    }
  return ret;
}

/* ------------------------------------------------------------------------
 * Code for interface org.gtk.vfs.Daemon
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:GVfsDBusDaemon
 * @title: GVfsDBusDaemon
 * @short_description: Generated C code for the org.gtk.vfs.Daemon D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gtk-vfs-Daemon.top_of_page">org.gtk.vfs.Daemon</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gtk.vfs.Daemon ---- */

static const _ExtendedGDBusArgInfo _gvfs_dbus_daemon_method_info_get_connection_OUT_ARG_address1 =
{
  {
    -1,
    (gchar *) "address1",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_daemon_method_info_get_connection_OUT_ARG_address2 =
{
  {
    -1,
    (gchar *) "address2",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_daemon_method_info_get_connection_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_daemon_method_info_get_connection_OUT_ARG_address1,
  &_gvfs_dbus_daemon_method_info_get_connection_OUT_ARG_address2,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_daemon_method_info_get_connection =
{
  {
    -1,
    (gchar *) "GetConnection",
    NULL,
    (GDBusArgInfo **) &_gvfs_dbus_daemon_method_info_get_connection_OUT_ARG_pointers,
    NULL
  },
  "handle-get-connection",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_daemon_method_info_cancel_IN_ARG_serial =
{
  {
    -1,
    (gchar *) "serial",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_daemon_method_info_cancel_IN_ARG_pointers[] =
{
  &_gvfs_dbus_daemon_method_info_cancel_IN_ARG_serial,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_daemon_method_info_cancel =
{
  {
    -1,
    (gchar *) "Cancel",
    (GDBusArgInfo **) &_gvfs_dbus_daemon_method_info_cancel_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-cancel",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_daemon_method_info_mount_IN_ARG_automount =
{
  {
    -1,
    (gchar *) "automount",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_daemon_method_info_mount_IN_ARG_dbus_id =
{
  {
    -1,
    (gchar *) "dbus_id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_daemon_method_info_mount_IN_ARG_obj_path =
{
  {
    -1,
    (gchar *) "obj_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_daemon_method_info_mount_IN_ARG_pointers[] =
{
  &_gvfs_dbus_daemon_method_info_mount_IN_ARG_automount,
  &_gvfs_dbus_daemon_method_info_mount_IN_ARG_dbus_id,
  &_gvfs_dbus_daemon_method_info_mount_IN_ARG_obj_path,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_daemon_method_info_mount_OUT_ARG_reply =
{
  {
    -1,
    (gchar *) "reply",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_daemon_method_info_mount_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_daemon_method_info_mount_OUT_ARG_reply,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_daemon_method_info_mount =
{
  {
    -1,
    (gchar *) "Mount",
    (GDBusArgInfo **) &_gvfs_dbus_daemon_method_info_mount_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_daemon_method_info_mount_OUT_ARG_pointers,
    NULL
  },
  "handle-mount",
  FALSE
};

static const _ExtendedGDBusMethodInfo * const _gvfs_dbus_daemon_method_info_pointers[] =
{
  &_gvfs_dbus_daemon_method_info_get_connection,
  &_gvfs_dbus_daemon_method_info_cancel,
  &_gvfs_dbus_daemon_method_info_mount,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _gvfs_dbus_daemon_interface_info =
{
  {
    -1,
    (gchar *) "org.gtk.vfs.Daemon",
    (GDBusMethodInfo **) &_gvfs_dbus_daemon_method_info_pointers,
    NULL,
    NULL,
    NULL
  },
  "daemon",
};


/**
 * gvfs_dbus_daemon_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gtk-vfs-Daemon.top_of_page">org.gtk.vfs.Daemon</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
gvfs_dbus_daemon_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_gvfs_dbus_daemon_interface_info.parent_struct;
}

/**
 * gvfs_dbus_daemon_override_properties:
 * @klass: The class structure for a #GObject<!-- -->-derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #GVfsDBusDaemon interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
gvfs_dbus_daemon_override_properties (GObjectClass *klass, guint property_id_begin)
{
  return property_id_begin - 1;
}



/**
 * GVfsDBusDaemon:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Daemon.top_of_page">org.gtk.vfs.Daemon</link>.
 */

/**
 * GVfsDBusDaemonIface:
 * @parent_iface: The parent interface.
 * @handle_cancel: Handler for the #GVfsDBusDaemon::handle-cancel signal.
 * @handle_get_connection: Handler for the #GVfsDBusDaemon::handle-get-connection signal.
 * @handle_mount: Handler for the #GVfsDBusDaemon::handle-mount signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Daemon.top_of_page">org.gtk.vfs.Daemon</link>.
 */

typedef GVfsDBusDaemonIface GVfsDBusDaemonInterface;
G_DEFINE_INTERFACE (GVfsDBusDaemon, gvfs_dbus_daemon, G_TYPE_OBJECT);

static void
gvfs_dbus_daemon_default_init (GVfsDBusDaemonIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * GVfsDBusDaemon::handle-get-connection:
   * @object: A #GVfsDBusDaemon.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Daemon.GetConnection">GetConnection()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_daemon_complete_get_connection() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-get-connection",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusDaemonIface, handle_get_connection),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * GVfsDBusDaemon::handle-cancel:
   * @object: A #GVfsDBusDaemon.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_serial: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Daemon.Cancel">Cancel()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_daemon_complete_cancel() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-cancel",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusDaemonIface, handle_cancel),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT);

  /**
   * GVfsDBusDaemon::handle-mount:
   * @object: A #GVfsDBusDaemon.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_automount: Argument passed by remote caller.
   * @arg_dbus_id: Argument passed by remote caller.
   * @arg_obj_path: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Daemon.Mount">Mount()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_daemon_complete_mount() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-mount",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusDaemonIface, handle_mount),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_BOOLEAN, G_TYPE_STRING, G_TYPE_STRING);

}

/**
 * gvfs_dbus_daemon_call_get_connection:
 * @proxy: A #GVfsDBusDaemonProxy.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Daemon.GetConnection">GetConnection()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_daemon_call_get_connection_finish() to get the result of the operation.
 *
 * See gvfs_dbus_daemon_call_get_connection_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_daemon_call_get_connection (
    GVfsDBusDaemon *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "GetConnection",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_daemon_call_get_connection_finish:
 * @proxy: A #GVfsDBusDaemonProxy.
 * @out_address1: (out): Return location for return parameter or %NULL to ignore.
 * @out_address2: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_daemon_call_get_connection().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_daemon_call_get_connection().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_daemon_call_get_connection_finish (
    GVfsDBusDaemon *proxy,
    gchar **out_address1,
    gchar **out_address2,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(ss)",
                 out_address1,
                 out_address2);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_daemon_call_get_connection_sync:
 * @proxy: A #GVfsDBusDaemonProxy.
 * @out_address1: (out): Return location for return parameter or %NULL to ignore.
 * @out_address2: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Daemon.GetConnection">GetConnection()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_daemon_call_get_connection() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_daemon_call_get_connection_sync (
    GVfsDBusDaemon *proxy,
    gchar **out_address1,
    gchar **out_address2,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "GetConnection",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(ss)",
                 out_address1,
                 out_address2);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_daemon_call_cancel:
 * @proxy: A #GVfsDBusDaemonProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Daemon.Cancel">Cancel()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_daemon_call_cancel_finish() to get the result of the operation.
 *
 * See gvfs_dbus_daemon_call_cancel_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_daemon_call_cancel (
    GVfsDBusDaemon *proxy,
    guint arg_serial,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Cancel",
    g_variant_new ("(u)",
                   arg_serial),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_daemon_call_cancel_finish:
 * @proxy: A #GVfsDBusDaemonProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_daemon_call_cancel().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_daemon_call_cancel().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_daemon_call_cancel_finish (
    GVfsDBusDaemon *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_daemon_call_cancel_sync:
 * @proxy: A #GVfsDBusDaemonProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Daemon.Cancel">Cancel()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_daemon_call_cancel() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_daemon_call_cancel_sync (
    GVfsDBusDaemon *proxy,
    guint arg_serial,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Cancel",
    g_variant_new ("(u)",
                   arg_serial),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_daemon_call_mount:
 * @proxy: A #GVfsDBusDaemonProxy.
 * @arg_automount: Argument to pass with the method invocation.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Daemon.Mount">Mount()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_daemon_call_mount_finish() to get the result of the operation.
 *
 * See gvfs_dbus_daemon_call_mount_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_daemon_call_mount (
    GVfsDBusDaemon *proxy,
    gboolean arg_automount,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Mount",
    g_variant_new ("(bso)",
                   arg_automount,
                   arg_dbus_id,
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_daemon_call_mount_finish:
 * @proxy: A #GVfsDBusDaemonProxy.
 * @out_reply: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_daemon_call_mount().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_daemon_call_mount().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_daemon_call_mount_finish (
    GVfsDBusDaemon *proxy,
    GVariant **out_reply,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a{sv})",
                 out_reply);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_daemon_call_mount_sync:
 * @proxy: A #GVfsDBusDaemonProxy.
 * @arg_automount: Argument to pass with the method invocation.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @out_reply: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Daemon.Mount">Mount()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_daemon_call_mount() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_daemon_call_mount_sync (
    GVfsDBusDaemon *proxy,
    gboolean arg_automount,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    GVariant **out_reply,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Mount",
    g_variant_new ("(bso)",
                   arg_automount,
                   arg_dbus_id,
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a{sv})",
                 out_reply);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_daemon_complete_get_connection:
 * @object: A #GVfsDBusDaemon.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @address1: Parameter to return.
 * @address2: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Daemon.GetConnection">GetConnection()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_daemon_complete_get_connection (
    GVfsDBusDaemon *object,
    GDBusMethodInvocation *invocation,
    const gchar *address1,
    const gchar *address2)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(ss)",
                   address1,
                   address2));
}

/**
 * gvfs_dbus_daemon_complete_cancel:
 * @object: A #GVfsDBusDaemon.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Daemon.Cancel">Cancel()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_daemon_complete_cancel (
    GVfsDBusDaemon *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_daemon_complete_mount:
 * @object: A #GVfsDBusDaemon.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @reply: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Daemon.Mount">Mount()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_daemon_complete_mount (
    GVfsDBusDaemon *object,
    GDBusMethodInvocation *invocation,
    GVariant *reply)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@a{sv})",
                   reply));
}

/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusDaemonProxy:
 *
 * The #GVfsDBusDaemonProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusDaemonProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusDaemonProxy.
 */

struct _GVfsDBusDaemonProxyPrivate
{
  GData *qdata;
};

static void gvfs_dbus_daemon_proxy_iface_init (GVfsDBusDaemonIface *iface);

G_DEFINE_TYPE_WITH_CODE (GVfsDBusDaemonProxy, gvfs_dbus_daemon_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (GVfsDBusDaemonProxy)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_DAEMON, gvfs_dbus_daemon_proxy_iface_init));

static void
gvfs_dbus_daemon_proxy_finalize (GObject *object)
{
  GVfsDBusDaemonProxy *proxy = GVFS_DBUS_DAEMON_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (gvfs_dbus_daemon_proxy_parent_class)->finalize (object);
}

static void
gvfs_dbus_daemon_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_daemon_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_daemon_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_gvfs_dbus_daemon_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], GVFS_DBUS_TYPE_DAEMON);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_DAEMON);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
gvfs_dbus_daemon_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  GVfsDBusDaemonProxy *proxy = GVFS_DBUS_DAEMON_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_daemon_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_daemon_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
gvfs_dbus_daemon_proxy_init (GVfsDBusDaemonProxy *proxy)
{
  proxy->priv = gvfs_dbus_daemon_proxy_get_instance_private (proxy);
  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), gvfs_dbus_daemon_interface_info ());
}

static void
gvfs_dbus_daemon_proxy_class_init (GVfsDBusDaemonProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = gvfs_dbus_daemon_proxy_finalize;
  gobject_class->get_property = gvfs_dbus_daemon_proxy_get_property;
  gobject_class->set_property = gvfs_dbus_daemon_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = gvfs_dbus_daemon_proxy_g_signal;
  proxy_class->g_properties_changed = gvfs_dbus_daemon_proxy_g_properties_changed;

}

static void
gvfs_dbus_daemon_proxy_iface_init (GVfsDBusDaemonIface *iface)
{
}

/**
 * gvfs_dbus_daemon_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Daemon.top_of_page">org.gtk.vfs.Daemon</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_daemon_proxy_new_finish() to get the result of the operation.
 *
 * See gvfs_dbus_daemon_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_daemon_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_DAEMON_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Daemon", NULL);
}

/**
 * gvfs_dbus_daemon_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_daemon_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_daemon_proxy_new().
 *
 * Returns: (transfer full) (type GVfsDBusDaemonProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusDaemon *
gvfs_dbus_daemon_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_DAEMON (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_daemon_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Daemon.top_of_page">org.gtk.vfs.Daemon</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_daemon_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusDaemonProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusDaemon *
gvfs_dbus_daemon_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_DAEMON_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Daemon", NULL);
  if (ret != NULL)
    return GVFS_DBUS_DAEMON (ret);
  else
    return NULL;
}


/**
 * gvfs_dbus_daemon_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like gvfs_dbus_daemon_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_daemon_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See gvfs_dbus_daemon_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_daemon_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_DAEMON_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Daemon", NULL);
}

/**
 * gvfs_dbus_daemon_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_daemon_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_daemon_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type GVfsDBusDaemonProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusDaemon *
gvfs_dbus_daemon_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_DAEMON (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_daemon_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like gvfs_dbus_daemon_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_daemon_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusDaemonProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusDaemon *
gvfs_dbus_daemon_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_DAEMON_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Daemon", NULL);
  if (ret != NULL)
    return GVFS_DBUS_DAEMON (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusDaemonSkeleton:
 *
 * The #GVfsDBusDaemonSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusDaemonSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusDaemonSkeleton.
 */

struct _GVfsDBusDaemonSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_gvfs_dbus_daemon_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  GVfsDBusDaemonSkeleton *skeleton = GVFS_DBUS_DAEMON_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint num_extra;
  guint n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], GVFS_DBUS_TYPE_DAEMON);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_DAEMON);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_gvfs_dbus_daemon_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  GVfsDBusDaemonSkeleton *skeleton = GVFS_DBUS_DAEMON_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_daemon_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_gvfs_dbus_daemon_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  GVfsDBusDaemonSkeleton *skeleton = GVFS_DBUS_DAEMON_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_daemon_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _gvfs_dbus_daemon_skeleton_vtable =
{
  _gvfs_dbus_daemon_skeleton_handle_method_call,
  _gvfs_dbus_daemon_skeleton_handle_get_property,
  _gvfs_dbus_daemon_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
gvfs_dbus_daemon_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return gvfs_dbus_daemon_interface_info ();
}

static GDBusInterfaceVTable *
gvfs_dbus_daemon_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_gvfs_dbus_daemon_skeleton_vtable;
}

static GVariant *
gvfs_dbus_daemon_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  GVfsDBusDaemonSkeleton *skeleton = GVFS_DBUS_DAEMON_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  if (_gvfs_dbus_daemon_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _gvfs_dbus_daemon_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _gvfs_dbus_daemon_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _gvfs_dbus_daemon_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.vfs.Daemon", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
gvfs_dbus_daemon_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
}

static void gvfs_dbus_daemon_skeleton_iface_init (GVfsDBusDaemonIface *iface);
G_DEFINE_TYPE_WITH_CODE (GVfsDBusDaemonSkeleton, gvfs_dbus_daemon_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (GVfsDBusDaemonSkeleton)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_DAEMON, gvfs_dbus_daemon_skeleton_iface_init));

static void
gvfs_dbus_daemon_skeleton_finalize (GObject *object)
{
  GVfsDBusDaemonSkeleton *skeleton = GVFS_DBUS_DAEMON_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (gvfs_dbus_daemon_skeleton_parent_class)->finalize (object);
}

static void
gvfs_dbus_daemon_skeleton_init (GVfsDBusDaemonSkeleton *skeleton)
{
  skeleton->priv = gvfs_dbus_daemon_skeleton_get_instance_private (skeleton);
  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
gvfs_dbus_daemon_skeleton_class_init (GVfsDBusDaemonSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = gvfs_dbus_daemon_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = gvfs_dbus_daemon_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = gvfs_dbus_daemon_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = gvfs_dbus_daemon_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = gvfs_dbus_daemon_skeleton_dbus_interface_get_vtable;
}

static void
gvfs_dbus_daemon_skeleton_iface_init (GVfsDBusDaemonIface *iface)
{
}

/**
 * gvfs_dbus_daemon_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Daemon.top_of_page">org.gtk.vfs.Daemon</link>.
 *
 * Returns: (transfer full) (type GVfsDBusDaemonSkeleton): The skeleton object.
 */
GVfsDBusDaemon *
gvfs_dbus_daemon_skeleton_new (void)
{
  return GVFS_DBUS_DAEMON (g_object_new (GVFS_DBUS_TYPE_DAEMON_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gtk.vfs.Spawner
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:GVfsDBusSpawner
 * @title: GVfsDBusSpawner
 * @short_description: Generated C code for the org.gtk.vfs.Spawner D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gtk-vfs-Spawner.top_of_page">org.gtk.vfs.Spawner</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gtk.vfs.Spawner ---- */

static const _ExtendedGDBusArgInfo _gvfs_dbus_spawner_method_info_spawned_IN_ARG_succeeded =
{
  {
    -1,
    (gchar *) "succeeded",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_spawner_method_info_spawned_IN_ARG_error_message =
{
  {
    -1,
    (gchar *) "error_message",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_spawner_method_info_spawned_IN_ARG_error_code =
{
  {
    -1,
    (gchar *) "error_code",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_spawner_method_info_spawned_IN_ARG_pointers[] =
{
  &_gvfs_dbus_spawner_method_info_spawned_IN_ARG_succeeded,
  &_gvfs_dbus_spawner_method_info_spawned_IN_ARG_error_message,
  &_gvfs_dbus_spawner_method_info_spawned_IN_ARG_error_code,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_spawner_method_info_spawned =
{
  {
    -1,
    (gchar *) "Spawned",
    (GDBusArgInfo **) &_gvfs_dbus_spawner_method_info_spawned_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-spawned",
  FALSE
};

static const _ExtendedGDBusMethodInfo * const _gvfs_dbus_spawner_method_info_pointers[] =
{
  &_gvfs_dbus_spawner_method_info_spawned,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _gvfs_dbus_spawner_interface_info =
{
  {
    -1,
    (gchar *) "org.gtk.vfs.Spawner",
    (GDBusMethodInfo **) &_gvfs_dbus_spawner_method_info_pointers,
    NULL,
    NULL,
    NULL
  },
  "spawner",
};


/**
 * gvfs_dbus_spawner_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gtk-vfs-Spawner.top_of_page">org.gtk.vfs.Spawner</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
gvfs_dbus_spawner_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_gvfs_dbus_spawner_interface_info.parent_struct;
}

/**
 * gvfs_dbus_spawner_override_properties:
 * @klass: The class structure for a #GObject<!-- -->-derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #GVfsDBusSpawner interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
gvfs_dbus_spawner_override_properties (GObjectClass *klass, guint property_id_begin)
{
  return property_id_begin - 1;
}



/**
 * GVfsDBusSpawner:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Spawner.top_of_page">org.gtk.vfs.Spawner</link>.
 */

/**
 * GVfsDBusSpawnerIface:
 * @parent_iface: The parent interface.
 * @handle_spawned: Handler for the #GVfsDBusSpawner::handle-spawned signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Spawner.top_of_page">org.gtk.vfs.Spawner</link>.
 */

typedef GVfsDBusSpawnerIface GVfsDBusSpawnerInterface;
G_DEFINE_INTERFACE (GVfsDBusSpawner, gvfs_dbus_spawner, G_TYPE_OBJECT);

static void
gvfs_dbus_spawner_default_init (GVfsDBusSpawnerIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * GVfsDBusSpawner::handle-spawned:
   * @object: A #GVfsDBusSpawner.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_succeeded: Argument passed by remote caller.
   * @arg_error_message: Argument passed by remote caller.
   * @arg_error_code: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Spawner.Spawned">Spawned()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_spawner_complete_spawned() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-spawned",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusSpawnerIface, handle_spawned),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_BOOLEAN, G_TYPE_STRING, G_TYPE_UINT);

}

/**
 * gvfs_dbus_spawner_call_spawned:
 * @proxy: A #GVfsDBusSpawnerProxy.
 * @arg_succeeded: Argument to pass with the method invocation.
 * @arg_error_message: Argument to pass with the method invocation.
 * @arg_error_code: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Spawner.Spawned">Spawned()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_spawner_call_spawned_finish() to get the result of the operation.
 *
 * See gvfs_dbus_spawner_call_spawned_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_spawner_call_spawned (
    GVfsDBusSpawner *proxy,
    gboolean arg_succeeded,
    const gchar *arg_error_message,
    guint arg_error_code,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Spawned",
    g_variant_new ("(bsu)",
                   arg_succeeded,
                   arg_error_message,
                   arg_error_code),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_spawner_call_spawned_finish:
 * @proxy: A #GVfsDBusSpawnerProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_spawner_call_spawned().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_spawner_call_spawned().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_spawner_call_spawned_finish (
    GVfsDBusSpawner *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_spawner_call_spawned_sync:
 * @proxy: A #GVfsDBusSpawnerProxy.
 * @arg_succeeded: Argument to pass with the method invocation.
 * @arg_error_message: Argument to pass with the method invocation.
 * @arg_error_code: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Spawner.Spawned">Spawned()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_spawner_call_spawned() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_spawner_call_spawned_sync (
    GVfsDBusSpawner *proxy,
    gboolean arg_succeeded,
    const gchar *arg_error_message,
    guint arg_error_code,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Spawned",
    g_variant_new ("(bsu)",
                   arg_succeeded,
                   arg_error_message,
                   arg_error_code),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_spawner_complete_spawned:
 * @object: A #GVfsDBusSpawner.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Spawner.Spawned">Spawned()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_spawner_complete_spawned (
    GVfsDBusSpawner *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusSpawnerProxy:
 *
 * The #GVfsDBusSpawnerProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusSpawnerProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusSpawnerProxy.
 */

struct _GVfsDBusSpawnerProxyPrivate
{
  GData *qdata;
};

static void gvfs_dbus_spawner_proxy_iface_init (GVfsDBusSpawnerIface *iface);

G_DEFINE_TYPE_WITH_CODE (GVfsDBusSpawnerProxy, gvfs_dbus_spawner_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (GVfsDBusSpawnerProxy)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_SPAWNER, gvfs_dbus_spawner_proxy_iface_init));

static void
gvfs_dbus_spawner_proxy_finalize (GObject *object)
{
  GVfsDBusSpawnerProxy *proxy = GVFS_DBUS_SPAWNER_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (gvfs_dbus_spawner_proxy_parent_class)->finalize (object);
}

static void
gvfs_dbus_spawner_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_spawner_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_spawner_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_gvfs_dbus_spawner_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], GVFS_DBUS_TYPE_SPAWNER);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_SPAWNER);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
gvfs_dbus_spawner_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  GVfsDBusSpawnerProxy *proxy = GVFS_DBUS_SPAWNER_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_spawner_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_spawner_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
gvfs_dbus_spawner_proxy_init (GVfsDBusSpawnerProxy *proxy)
{
  proxy->priv = gvfs_dbus_spawner_proxy_get_instance_private (proxy);
  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), gvfs_dbus_spawner_interface_info ());
}

static void
gvfs_dbus_spawner_proxy_class_init (GVfsDBusSpawnerProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = gvfs_dbus_spawner_proxy_finalize;
  gobject_class->get_property = gvfs_dbus_spawner_proxy_get_property;
  gobject_class->set_property = gvfs_dbus_spawner_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = gvfs_dbus_spawner_proxy_g_signal;
  proxy_class->g_properties_changed = gvfs_dbus_spawner_proxy_g_properties_changed;

}

static void
gvfs_dbus_spawner_proxy_iface_init (GVfsDBusSpawnerIface *iface)
{
}

/**
 * gvfs_dbus_spawner_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Spawner.top_of_page">org.gtk.vfs.Spawner</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_spawner_proxy_new_finish() to get the result of the operation.
 *
 * See gvfs_dbus_spawner_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_spawner_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_SPAWNER_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Spawner", NULL);
}

/**
 * gvfs_dbus_spawner_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_spawner_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_spawner_proxy_new().
 *
 * Returns: (transfer full) (type GVfsDBusSpawnerProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusSpawner *
gvfs_dbus_spawner_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_SPAWNER (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_spawner_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Spawner.top_of_page">org.gtk.vfs.Spawner</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_spawner_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusSpawnerProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusSpawner *
gvfs_dbus_spawner_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_SPAWNER_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Spawner", NULL);
  if (ret != NULL)
    return GVFS_DBUS_SPAWNER (ret);
  else
    return NULL;
}


/**
 * gvfs_dbus_spawner_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like gvfs_dbus_spawner_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_spawner_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See gvfs_dbus_spawner_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_spawner_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_SPAWNER_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Spawner", NULL);
}

/**
 * gvfs_dbus_spawner_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_spawner_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_spawner_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type GVfsDBusSpawnerProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusSpawner *
gvfs_dbus_spawner_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_SPAWNER (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_spawner_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like gvfs_dbus_spawner_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_spawner_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusSpawnerProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusSpawner *
gvfs_dbus_spawner_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_SPAWNER_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Spawner", NULL);
  if (ret != NULL)
    return GVFS_DBUS_SPAWNER (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusSpawnerSkeleton:
 *
 * The #GVfsDBusSpawnerSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusSpawnerSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusSpawnerSkeleton.
 */

struct _GVfsDBusSpawnerSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_gvfs_dbus_spawner_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  GVfsDBusSpawnerSkeleton *skeleton = GVFS_DBUS_SPAWNER_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint num_extra;
  guint n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], GVFS_DBUS_TYPE_SPAWNER);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_SPAWNER);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_gvfs_dbus_spawner_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  GVfsDBusSpawnerSkeleton *skeleton = GVFS_DBUS_SPAWNER_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_spawner_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_gvfs_dbus_spawner_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  GVfsDBusSpawnerSkeleton *skeleton = GVFS_DBUS_SPAWNER_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_spawner_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _gvfs_dbus_spawner_skeleton_vtable =
{
  _gvfs_dbus_spawner_skeleton_handle_method_call,
  _gvfs_dbus_spawner_skeleton_handle_get_property,
  _gvfs_dbus_spawner_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
gvfs_dbus_spawner_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return gvfs_dbus_spawner_interface_info ();
}

static GDBusInterfaceVTable *
gvfs_dbus_spawner_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_gvfs_dbus_spawner_skeleton_vtable;
}

static GVariant *
gvfs_dbus_spawner_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  GVfsDBusSpawnerSkeleton *skeleton = GVFS_DBUS_SPAWNER_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  if (_gvfs_dbus_spawner_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _gvfs_dbus_spawner_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _gvfs_dbus_spawner_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _gvfs_dbus_spawner_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.vfs.Spawner", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
gvfs_dbus_spawner_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
}

static void gvfs_dbus_spawner_skeleton_iface_init (GVfsDBusSpawnerIface *iface);
G_DEFINE_TYPE_WITH_CODE (GVfsDBusSpawnerSkeleton, gvfs_dbus_spawner_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (GVfsDBusSpawnerSkeleton)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_SPAWNER, gvfs_dbus_spawner_skeleton_iface_init));

static void
gvfs_dbus_spawner_skeleton_finalize (GObject *object)
{
  GVfsDBusSpawnerSkeleton *skeleton = GVFS_DBUS_SPAWNER_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (gvfs_dbus_spawner_skeleton_parent_class)->finalize (object);
}

static void
gvfs_dbus_spawner_skeleton_init (GVfsDBusSpawnerSkeleton *skeleton)
{
  skeleton->priv = gvfs_dbus_spawner_skeleton_get_instance_private (skeleton);
  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
gvfs_dbus_spawner_skeleton_class_init (GVfsDBusSpawnerSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = gvfs_dbus_spawner_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = gvfs_dbus_spawner_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = gvfs_dbus_spawner_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = gvfs_dbus_spawner_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = gvfs_dbus_spawner_skeleton_dbus_interface_get_vtable;
}

static void
gvfs_dbus_spawner_skeleton_iface_init (GVfsDBusSpawnerIface *iface)
{
}

/**
 * gvfs_dbus_spawner_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Spawner.top_of_page">org.gtk.vfs.Spawner</link>.
 *
 * Returns: (transfer full) (type GVfsDBusSpawnerSkeleton): The skeleton object.
 */
GVfsDBusSpawner *
gvfs_dbus_spawner_skeleton_new (void)
{
  return GVFS_DBUS_SPAWNER (g_object_new (GVFS_DBUS_TYPE_SPAWNER_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gtk.vfs.MountTracker
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:GVfsDBusMountTracker
 * @title: GVfsDBusMountTracker
 * @short_description: Generated C code for the org.gtk.vfs.MountTracker D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gtk-vfs-MountTracker.top_of_page">org.gtk.vfs.MountTracker</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gtk.vfs.MountTracker ---- */

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_lookup_mount_IN_ARG_mount_spec =
{
  {
    -1,
    (gchar *) "mount_spec",
    (gchar *) "(aya{sv})",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_tracker_method_info_lookup_mount_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_tracker_method_info_lookup_mount_IN_ARG_mount_spec,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_lookup_mount_OUT_ARG_mount =
{
  {
    -1,
    (gchar *) "mount",
    (gchar *) "(sossssssbay(aya{sv})ay)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_tracker_method_info_lookup_mount_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_tracker_method_info_lookup_mount_OUT_ARG_mount,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_tracker_method_info_lookup_mount =
{
  {
    -1,
    (gchar *) "LookupMount",
    (GDBusArgInfo **) &_gvfs_dbus_mount_tracker_method_info_lookup_mount_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_tracker_method_info_lookup_mount_OUT_ARG_pointers,
    NULL
  },
  "handle-lookup-mount",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_lookup_mount_by_fuse_path_IN_ARG_fuse_path =
{
  {
    -1,
    (gchar *) "fuse_path",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_tracker_method_info_lookup_mount_by_fuse_path_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_tracker_method_info_lookup_mount_by_fuse_path_IN_ARG_fuse_path,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_lookup_mount_by_fuse_path_OUT_ARG_mount =
{
  {
    -1,
    (gchar *) "mount",
    (gchar *) "(sossssssbay(aya{sv})ay)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_tracker_method_info_lookup_mount_by_fuse_path_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_tracker_method_info_lookup_mount_by_fuse_path_OUT_ARG_mount,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_tracker_method_info_lookup_mount_by_fuse_path =
{
  {
    -1,
    (gchar *) "LookupMountByFusePath",
    (GDBusArgInfo **) &_gvfs_dbus_mount_tracker_method_info_lookup_mount_by_fuse_path_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_tracker_method_info_lookup_mount_by_fuse_path_OUT_ARG_pointers,
    NULL
  },
  "handle-lookup-mount-by-fuse-path",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_mount_location_IN_ARG_mount_spec =
{
  {
    -1,
    (gchar *) "mount_spec",
    (gchar *) "(aya{sv})",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_mount_location_IN_ARG_mount_source =
{
  {
    -1,
    (gchar *) "mount_source",
    (gchar *) "(so)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_tracker_method_info_mount_location_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_tracker_method_info_mount_location_IN_ARG_mount_spec,
  &_gvfs_dbus_mount_tracker_method_info_mount_location_IN_ARG_mount_source,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_tracker_method_info_mount_location =
{
  {
    -1,
    (gchar *) "MountLocation",
    (GDBusArgInfo **) &_gvfs_dbus_mount_tracker_method_info_mount_location_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-mount-location",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_list_mounts_OUT_ARG_mounts =
{
  {
    -1,
    (gchar *) "mounts",
    (gchar *) "a(sossssssbay(aya{sv})ay)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_tracker_method_info_list_mounts_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_tracker_method_info_list_mounts_OUT_ARG_mounts,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_tracker_method_info_list_mounts =
{
  {
    -1,
    (gchar *) "ListMounts",
    NULL,
    (GDBusArgInfo **) &_gvfs_dbus_mount_tracker_method_info_list_mounts_OUT_ARG_pointers,
    NULL
  },
  "handle-list-mounts",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_obj_path =
{
  {
    -1,
    (gchar *) "obj_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_display_name =
{
  {
    -1,
    (gchar *) "display_name",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_stable_name =
{
  {
    -1,
    (gchar *) "stable_name",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_x_content_types =
{
  {
    -1,
    (gchar *) "x_content_types",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_icon =
{
  {
    -1,
    (gchar *) "icon",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_symbolic_icon =
{
  {
    -1,
    (gchar *) "symbolic_icon",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_prefered_filename_encoding =
{
  {
    -1,
    (gchar *) "prefered_filename_encoding",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_user_visible =
{
  {
    -1,
    (gchar *) "user_visible",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_mount_spec =
{
  {
    -1,
    (gchar *) "mount_spec",
    (gchar *) "(aya{sv})",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_default_location =
{
  {
    -1,
    (gchar *) "default_location",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_obj_path,
  &_gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_display_name,
  &_gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_stable_name,
  &_gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_x_content_types,
  &_gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_icon,
  &_gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_symbolic_icon,
  &_gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_prefered_filename_encoding,
  &_gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_user_visible,
  &_gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_mount_spec,
  &_gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_default_location,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_tracker_method_info_register_mount =
{
  {
    -1,
    (gchar *) "RegisterMount",
    (GDBusArgInfo **) &_gvfs_dbus_mount_tracker_method_info_register_mount_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-register-mount",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_unregister_mount_IN_ARG_obj_path =
{
  {
    -1,
    (gchar *) "obj_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_tracker_method_info_unregister_mount_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_tracker_method_info_unregister_mount_IN_ARG_obj_path,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_tracker_method_info_unregister_mount =
{
  {
    -1,
    (gchar *) "UnregisterMount",
    (GDBusArgInfo **) &_gvfs_dbus_mount_tracker_method_info_unregister_mount_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-unregister-mount",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_list_mount_types_OUT_ARG_mount_types =
{
  {
    -1,
    (gchar *) "mount_types",
    (gchar *) "as",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_tracker_method_info_list_mount_types_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_tracker_method_info_list_mount_types_OUT_ARG_mount_types,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_tracker_method_info_list_mount_types =
{
  {
    -1,
    (gchar *) "ListMountTypes",
    NULL,
    (GDBusArgInfo **) &_gvfs_dbus_mount_tracker_method_info_list_mount_types_OUT_ARG_pointers,
    NULL
  },
  "handle-list-mount-types",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_method_info_list_mountable_info_OUT_ARG_mountables =
{
  {
    -1,
    (gchar *) "mountables",
    (gchar *) "a(ssasib)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_tracker_method_info_list_mountable_info_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_tracker_method_info_list_mountable_info_OUT_ARG_mountables,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_tracker_method_info_list_mountable_info =
{
  {
    -1,
    (gchar *) "ListMountableInfo",
    NULL,
    (GDBusArgInfo **) &_gvfs_dbus_mount_tracker_method_info_list_mountable_info_OUT_ARG_pointers,
    NULL
  },
  "handle-list-mountable-info",
  FALSE
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_tracker_method_info_register_fuse =
{
  {
    -1,
    (gchar *) "RegisterFuse",
    NULL,
    NULL,
    NULL
  },
  "handle-register-fuse",
  FALSE
};

static const _ExtendedGDBusMethodInfo * const _gvfs_dbus_mount_tracker_method_info_pointers[] =
{
  &_gvfs_dbus_mount_tracker_method_info_lookup_mount,
  &_gvfs_dbus_mount_tracker_method_info_lookup_mount_by_fuse_path,
  &_gvfs_dbus_mount_tracker_method_info_mount_location,
  &_gvfs_dbus_mount_tracker_method_info_list_mounts,
  &_gvfs_dbus_mount_tracker_method_info_register_mount,
  &_gvfs_dbus_mount_tracker_method_info_unregister_mount,
  &_gvfs_dbus_mount_tracker_method_info_list_mount_types,
  &_gvfs_dbus_mount_tracker_method_info_list_mountable_info,
  &_gvfs_dbus_mount_tracker_method_info_register_fuse,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_signal_info_mounted_ARG_mount =
{
  {
    -1,
    (gchar *) "mount",
    (gchar *) "(sossssssbay(aya{sv})ay)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_tracker_signal_info_mounted_ARG_pointers[] =
{
  &_gvfs_dbus_mount_tracker_signal_info_mounted_ARG_mount,
  NULL
};

static const _ExtendedGDBusSignalInfo _gvfs_dbus_mount_tracker_signal_info_mounted =
{
  {
    -1,
    (gchar *) "Mounted",
    (GDBusArgInfo **) &_gvfs_dbus_mount_tracker_signal_info_mounted_ARG_pointers,
    NULL
  },
  "mounted"
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_tracker_signal_info_unmounted_ARG_mount =
{
  {
    -1,
    (gchar *) "mount",
    (gchar *) "(sossssssbay(aya{sv})ay)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_tracker_signal_info_unmounted_ARG_pointers[] =
{
  &_gvfs_dbus_mount_tracker_signal_info_unmounted_ARG_mount,
  NULL
};

static const _ExtendedGDBusSignalInfo _gvfs_dbus_mount_tracker_signal_info_unmounted =
{
  {
    -1,
    (gchar *) "Unmounted",
    (GDBusArgInfo **) &_gvfs_dbus_mount_tracker_signal_info_unmounted_ARG_pointers,
    NULL
  },
  "unmounted"
};

static const _ExtendedGDBusSignalInfo * const _gvfs_dbus_mount_tracker_signal_info_pointers[] =
{
  &_gvfs_dbus_mount_tracker_signal_info_mounted,
  &_gvfs_dbus_mount_tracker_signal_info_unmounted,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _gvfs_dbus_mount_tracker_interface_info =
{
  {
    -1,
    (gchar *) "org.gtk.vfs.MountTracker",
    (GDBusMethodInfo **) &_gvfs_dbus_mount_tracker_method_info_pointers,
    (GDBusSignalInfo **) &_gvfs_dbus_mount_tracker_signal_info_pointers,
    NULL,
    NULL
  },
  "mount-tracker",
};


/**
 * gvfs_dbus_mount_tracker_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gtk-vfs-MountTracker.top_of_page">org.gtk.vfs.MountTracker</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
gvfs_dbus_mount_tracker_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_gvfs_dbus_mount_tracker_interface_info.parent_struct;
}

/**
 * gvfs_dbus_mount_tracker_override_properties:
 * @klass: The class structure for a #GObject<!-- -->-derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #GVfsDBusMountTracker interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
gvfs_dbus_mount_tracker_override_properties (GObjectClass *klass, guint property_id_begin)
{
  return property_id_begin - 1;
}



/**
 * GVfsDBusMountTracker:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MountTracker.top_of_page">org.gtk.vfs.MountTracker</link>.
 */

/**
 * GVfsDBusMountTrackerIface:
 * @parent_iface: The parent interface.
 * @handle_list_mount_types: Handler for the #GVfsDBusMountTracker::handle-list-mount-types signal.
 * @handle_list_mountable_info: Handler for the #GVfsDBusMountTracker::handle-list-mountable-info signal.
 * @handle_list_mounts: Handler for the #GVfsDBusMountTracker::handle-list-mounts signal.
 * @handle_lookup_mount: Handler for the #GVfsDBusMountTracker::handle-lookup-mount signal.
 * @handle_lookup_mount_by_fuse_path: Handler for the #GVfsDBusMountTracker::handle-lookup-mount-by-fuse-path signal.
 * @handle_mount_location: Handler for the #GVfsDBusMountTracker::handle-mount-location signal.
 * @handle_register_fuse: Handler for the #GVfsDBusMountTracker::handle-register-fuse signal.
 * @handle_register_mount: Handler for the #GVfsDBusMountTracker::handle-register-mount signal.
 * @handle_unregister_mount: Handler for the #GVfsDBusMountTracker::handle-unregister-mount signal.
 * @mounted: Handler for the #GVfsDBusMountTracker::mounted signal.
 * @unmounted: Handler for the #GVfsDBusMountTracker::unmounted signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MountTracker.top_of_page">org.gtk.vfs.MountTracker</link>.
 */

typedef GVfsDBusMountTrackerIface GVfsDBusMountTrackerInterface;
G_DEFINE_INTERFACE (GVfsDBusMountTracker, gvfs_dbus_mount_tracker, G_TYPE_OBJECT);

static void
gvfs_dbus_mount_tracker_default_init (GVfsDBusMountTrackerIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * GVfsDBusMountTracker::handle-lookup-mount:
   * @object: A #GVfsDBusMountTracker.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_mount_spec: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.LookupMount">LookupMount()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_tracker_complete_lookup_mount() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-lookup-mount",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountTrackerIface, handle_lookup_mount),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT);

  /**
   * GVfsDBusMountTracker::handle-lookup-mount-by-fuse-path:
   * @object: A #GVfsDBusMountTracker.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_fuse_path: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.LookupMountByFusePath">LookupMountByFusePath()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_tracker_complete_lookup_mount_by_fuse_path() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-lookup-mount-by-fuse-path",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountTrackerIface, handle_lookup_mount_by_fuse_path),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * GVfsDBusMountTracker::handle-mount-location:
   * @object: A #GVfsDBusMountTracker.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_mount_spec: Argument passed by remote caller.
   * @arg_mount_source: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.MountLocation">MountLocation()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_tracker_complete_mount_location() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-mount-location",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountTrackerIface, handle_mount_location),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT, G_TYPE_VARIANT);

  /**
   * GVfsDBusMountTracker::handle-list-mounts:
   * @object: A #GVfsDBusMountTracker.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.ListMounts">ListMounts()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_tracker_complete_list_mounts() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-list-mounts",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountTrackerIface, handle_list_mounts),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * GVfsDBusMountTracker::handle-register-mount:
   * @object: A #GVfsDBusMountTracker.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_obj_path: Argument passed by remote caller.
   * @arg_display_name: Argument passed by remote caller.
   * @arg_stable_name: Argument passed by remote caller.
   * @arg_x_content_types: Argument passed by remote caller.
   * @arg_icon: Argument passed by remote caller.
   * @arg_symbolic_icon: Argument passed by remote caller.
   * @arg_prefered_filename_encoding: Argument passed by remote caller.
   * @arg_user_visible: Argument passed by remote caller.
   * @arg_mount_spec: Argument passed by remote caller.
   * @arg_default_location: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.RegisterMount">RegisterMount()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_tracker_complete_register_mount() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-register-mount",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountTrackerIface, handle_register_mount),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    11,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN, G_TYPE_VARIANT, G_TYPE_STRING);

  /**
   * GVfsDBusMountTracker::handle-unregister-mount:
   * @object: A #GVfsDBusMountTracker.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_obj_path: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.UnregisterMount">UnregisterMount()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_tracker_complete_unregister_mount() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-unregister-mount",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountTrackerIface, handle_unregister_mount),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * GVfsDBusMountTracker::handle-list-mount-types:
   * @object: A #GVfsDBusMountTracker.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.ListMountTypes">ListMountTypes()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_tracker_complete_list_mount_types() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-list-mount-types",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountTrackerIface, handle_list_mount_types),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * GVfsDBusMountTracker::handle-list-mountable-info:
   * @object: A #GVfsDBusMountTracker.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.ListMountableInfo">ListMountableInfo()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_tracker_complete_list_mountable_info() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-list-mountable-info",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountTrackerIface, handle_list_mountable_info),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * GVfsDBusMountTracker::handle-register-fuse:
   * @object: A #GVfsDBusMountTracker.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.RegisterFuse">RegisterFuse()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_tracker_complete_register_fuse() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-register-fuse",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountTrackerIface, handle_register_fuse),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /* GObject signals for received D-Bus signals: */
  /**
   * GVfsDBusMountTracker::mounted:
   * @object: A #GVfsDBusMountTracker.
   * @arg_mount: Argument.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gtk-vfs-MountTracker.Mounted">"Mounted"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  g_signal_new ("mounted",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountTrackerIface, mounted),
    NULL,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_NONE,
    1, G_TYPE_VARIANT);

  /**
   * GVfsDBusMountTracker::unmounted:
   * @object: A #GVfsDBusMountTracker.
   * @arg_mount: Argument.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gtk-vfs-MountTracker.Unmounted">"Unmounted"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  g_signal_new ("unmounted",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountTrackerIface, unmounted),
    NULL,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_NONE,
    1, G_TYPE_VARIANT);

}

/**
 * gvfs_dbus_mount_tracker_emit_mounted:
 * @object: A #GVfsDBusMountTracker.
 * @arg_mount: Argument to pass with the signal.
 *
 * Emits the <link linkend="gdbus-signal-org-gtk-vfs-MountTracker.Mounted">"Mounted"</link> D-Bus signal.
 */
void
gvfs_dbus_mount_tracker_emit_mounted (
    GVfsDBusMountTracker *object,
    GVariant *arg_mount)
{
  g_signal_emit_by_name (object, "mounted", arg_mount);
}

/**
 * gvfs_dbus_mount_tracker_emit_unmounted:
 * @object: A #GVfsDBusMountTracker.
 * @arg_mount: Argument to pass with the signal.
 *
 * Emits the <link linkend="gdbus-signal-org-gtk-vfs-MountTracker.Unmounted">"Unmounted"</link> D-Bus signal.
 */
void
gvfs_dbus_mount_tracker_emit_unmounted (
    GVfsDBusMountTracker *object,
    GVariant *arg_mount)
{
  g_signal_emit_by_name (object, "unmounted", arg_mount);
}

/**
 * gvfs_dbus_mount_tracker_call_lookup_mount:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @arg_mount_spec: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.LookupMount">LookupMount()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_tracker_call_lookup_mount_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_tracker_call_lookup_mount_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_tracker_call_lookup_mount (
    GVfsDBusMountTracker *proxy,
    GVariant *arg_mount_spec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "LookupMount",
    g_variant_new ("(@(aya{sv}))",
                   arg_mount_spec),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_tracker_call_lookup_mount_finish:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @out_mount: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_tracker_call_lookup_mount().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_tracker_call_lookup_mount().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_lookup_mount_finish (
    GVfsDBusMountTracker *proxy,
    GVariant **out_mount,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@(sossssssbay(aya{sv})ay))",
                 out_mount);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_lookup_mount_sync:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @arg_mount_spec: Argument to pass with the method invocation.
 * @out_mount: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.LookupMount">LookupMount()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_tracker_call_lookup_mount() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_lookup_mount_sync (
    GVfsDBusMountTracker *proxy,
    GVariant *arg_mount_spec,
    GVariant **out_mount,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "LookupMount",
    g_variant_new ("(@(aya{sv}))",
                   arg_mount_spec),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@(sossssssbay(aya{sv})ay))",
                 out_mount);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_lookup_mount_by_fuse_path:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @arg_fuse_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.LookupMountByFusePath">LookupMountByFusePath()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_tracker_call_lookup_mount_by_fuse_path_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_tracker_call_lookup_mount_by_fuse_path_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_tracker_call_lookup_mount_by_fuse_path (
    GVfsDBusMountTracker *proxy,
    const gchar *arg_fuse_path,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "LookupMountByFusePath",
    g_variant_new ("(^ay)",
                   arg_fuse_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_tracker_call_lookup_mount_by_fuse_path_finish:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @out_mount: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_tracker_call_lookup_mount_by_fuse_path().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_tracker_call_lookup_mount_by_fuse_path().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_lookup_mount_by_fuse_path_finish (
    GVfsDBusMountTracker *proxy,
    GVariant **out_mount,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@(sossssssbay(aya{sv})ay))",
                 out_mount);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_lookup_mount_by_fuse_path_sync:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @arg_fuse_path: Argument to pass with the method invocation.
 * @out_mount: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.LookupMountByFusePath">LookupMountByFusePath()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_tracker_call_lookup_mount_by_fuse_path() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_lookup_mount_by_fuse_path_sync (
    GVfsDBusMountTracker *proxy,
    const gchar *arg_fuse_path,
    GVariant **out_mount,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "LookupMountByFusePath",
    g_variant_new ("(^ay)",
                   arg_fuse_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@(sossssssbay(aya{sv})ay))",
                 out_mount);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_mount_location:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @arg_mount_spec: Argument to pass with the method invocation.
 * @arg_mount_source: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.MountLocation">MountLocation()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_tracker_call_mount_location_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_tracker_call_mount_location_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_tracker_call_mount_location (
    GVfsDBusMountTracker *proxy,
    GVariant *arg_mount_spec,
    GVariant *arg_mount_source,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "MountLocation",
    g_variant_new ("(@(aya{sv})@(so))",
                   arg_mount_spec,
                   arg_mount_source),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_tracker_call_mount_location_finish:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_tracker_call_mount_location().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_tracker_call_mount_location().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_mount_location_finish (
    GVfsDBusMountTracker *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_mount_location_sync:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @arg_mount_spec: Argument to pass with the method invocation.
 * @arg_mount_source: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.MountLocation">MountLocation()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_tracker_call_mount_location() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_mount_location_sync (
    GVfsDBusMountTracker *proxy,
    GVariant *arg_mount_spec,
    GVariant *arg_mount_source,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "MountLocation",
    g_variant_new ("(@(aya{sv})@(so))",
                   arg_mount_spec,
                   arg_mount_source),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_list_mounts:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.ListMounts">ListMounts()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_tracker_call_list_mounts_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_tracker_call_list_mounts_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_tracker_call_list_mounts (
    GVfsDBusMountTracker *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ListMounts",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_tracker_call_list_mounts_finish:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @out_mounts: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_tracker_call_list_mounts().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_tracker_call_list_mounts().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_list_mounts_finish (
    GVfsDBusMountTracker *proxy,
    GVariant **out_mounts,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(sossssssbay(aya{sv})ay))",
                 out_mounts);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_list_mounts_sync:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @out_mounts: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.ListMounts">ListMounts()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_tracker_call_list_mounts() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_list_mounts_sync (
    GVfsDBusMountTracker *proxy,
    GVariant **out_mounts,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ListMounts",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(sossssssbay(aya{sv})ay))",
                 out_mounts);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_register_mount:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @arg_display_name: Argument to pass with the method invocation.
 * @arg_stable_name: Argument to pass with the method invocation.
 * @arg_x_content_types: Argument to pass with the method invocation.
 * @arg_icon: Argument to pass with the method invocation.
 * @arg_symbolic_icon: Argument to pass with the method invocation.
 * @arg_prefered_filename_encoding: Argument to pass with the method invocation.
 * @arg_user_visible: Argument to pass with the method invocation.
 * @arg_mount_spec: Argument to pass with the method invocation.
 * @arg_default_location: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.RegisterMount">RegisterMount()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_tracker_call_register_mount_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_tracker_call_register_mount_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_tracker_call_register_mount (
    GVfsDBusMountTracker *proxy,
    const gchar *arg_obj_path,
    const gchar *arg_display_name,
    const gchar *arg_stable_name,
    const gchar *arg_x_content_types,
    const gchar *arg_icon,
    const gchar *arg_symbolic_icon,
    const gchar *arg_prefered_filename_encoding,
    gboolean arg_user_visible,
    GVariant *arg_mount_spec,
    const gchar *arg_default_location,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "RegisterMount",
    g_variant_new ("(ossssssb@(aya{sv})^ay)",
                   arg_obj_path,
                   arg_display_name,
                   arg_stable_name,
                   arg_x_content_types,
                   arg_icon,
                   arg_symbolic_icon,
                   arg_prefered_filename_encoding,
                   arg_user_visible,
                   arg_mount_spec,
                   arg_default_location),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_tracker_call_register_mount_finish:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_tracker_call_register_mount().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_tracker_call_register_mount().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_register_mount_finish (
    GVfsDBusMountTracker *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_register_mount_sync:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @arg_display_name: Argument to pass with the method invocation.
 * @arg_stable_name: Argument to pass with the method invocation.
 * @arg_x_content_types: Argument to pass with the method invocation.
 * @arg_icon: Argument to pass with the method invocation.
 * @arg_symbolic_icon: Argument to pass with the method invocation.
 * @arg_prefered_filename_encoding: Argument to pass with the method invocation.
 * @arg_user_visible: Argument to pass with the method invocation.
 * @arg_mount_spec: Argument to pass with the method invocation.
 * @arg_default_location: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.RegisterMount">RegisterMount()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_tracker_call_register_mount() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_register_mount_sync (
    GVfsDBusMountTracker *proxy,
    const gchar *arg_obj_path,
    const gchar *arg_display_name,
    const gchar *arg_stable_name,
    const gchar *arg_x_content_types,
    const gchar *arg_icon,
    const gchar *arg_symbolic_icon,
    const gchar *arg_prefered_filename_encoding,
    gboolean arg_user_visible,
    GVariant *arg_mount_spec,
    const gchar *arg_default_location,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "RegisterMount",
    g_variant_new ("(ossssssb@(aya{sv})^ay)",
                   arg_obj_path,
                   arg_display_name,
                   arg_stable_name,
                   arg_x_content_types,
                   arg_icon,
                   arg_symbolic_icon,
                   arg_prefered_filename_encoding,
                   arg_user_visible,
                   arg_mount_spec,
                   arg_default_location),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_unregister_mount:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.UnregisterMount">UnregisterMount()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_tracker_call_unregister_mount_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_tracker_call_unregister_mount_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_tracker_call_unregister_mount (
    GVfsDBusMountTracker *proxy,
    const gchar *arg_obj_path,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "UnregisterMount",
    g_variant_new ("(o)",
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_tracker_call_unregister_mount_finish:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_tracker_call_unregister_mount().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_tracker_call_unregister_mount().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_unregister_mount_finish (
    GVfsDBusMountTracker *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_unregister_mount_sync:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.UnregisterMount">UnregisterMount()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_tracker_call_unregister_mount() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_unregister_mount_sync (
    GVfsDBusMountTracker *proxy,
    const gchar *arg_obj_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "UnregisterMount",
    g_variant_new ("(o)",
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_list_mount_types:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.ListMountTypes">ListMountTypes()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_tracker_call_list_mount_types_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_tracker_call_list_mount_types_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_tracker_call_list_mount_types (
    GVfsDBusMountTracker *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ListMountTypes",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_tracker_call_list_mount_types_finish:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @out_mount_types: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_tracker_call_list_mount_types().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_tracker_call_list_mount_types().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_list_mount_types_finish (
    GVfsDBusMountTracker *proxy,
    gchar ***out_mount_types,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(^as)",
                 out_mount_types);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_list_mount_types_sync:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @out_mount_types: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.ListMountTypes">ListMountTypes()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_tracker_call_list_mount_types() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_list_mount_types_sync (
    GVfsDBusMountTracker *proxy,
    gchar ***out_mount_types,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ListMountTypes",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(^as)",
                 out_mount_types);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_list_mountable_info:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.ListMountableInfo">ListMountableInfo()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_tracker_call_list_mountable_info_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_tracker_call_list_mountable_info_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_tracker_call_list_mountable_info (
    GVfsDBusMountTracker *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ListMountableInfo",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_tracker_call_list_mountable_info_finish:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @out_mountables: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_tracker_call_list_mountable_info().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_tracker_call_list_mountable_info().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_list_mountable_info_finish (
    GVfsDBusMountTracker *proxy,
    GVariant **out_mountables,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(ssasib))",
                 out_mountables);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_list_mountable_info_sync:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @out_mountables: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.ListMountableInfo">ListMountableInfo()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_tracker_call_list_mountable_info() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_list_mountable_info_sync (
    GVfsDBusMountTracker *proxy,
    GVariant **out_mountables,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ListMountableInfo",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(ssasib))",
                 out_mountables);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_register_fuse:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.RegisterFuse">RegisterFuse()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_tracker_call_register_fuse_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_tracker_call_register_fuse_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_tracker_call_register_fuse (
    GVfsDBusMountTracker *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "RegisterFuse",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_tracker_call_register_fuse_finish:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_tracker_call_register_fuse().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_tracker_call_register_fuse().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_register_fuse_finish (
    GVfsDBusMountTracker *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_call_register_fuse_sync:
 * @proxy: A #GVfsDBusMountTrackerProxy.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.RegisterFuse">RegisterFuse()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_tracker_call_register_fuse() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_tracker_call_register_fuse_sync (
    GVfsDBusMountTracker *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "RegisterFuse",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_tracker_complete_lookup_mount:
 * @object: A #GVfsDBusMountTracker.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @mount: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.LookupMount">LookupMount()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_tracker_complete_lookup_mount (
    GVfsDBusMountTracker *object,
    GDBusMethodInvocation *invocation,
    GVariant *mount)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@(sossssssbay(aya{sv})ay))",
                   mount));
}

/**
 * gvfs_dbus_mount_tracker_complete_lookup_mount_by_fuse_path:
 * @object: A #GVfsDBusMountTracker.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @mount: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.LookupMountByFusePath">LookupMountByFusePath()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_tracker_complete_lookup_mount_by_fuse_path (
    GVfsDBusMountTracker *object,
    GDBusMethodInvocation *invocation,
    GVariant *mount)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@(sossssssbay(aya{sv})ay))",
                   mount));
}

/**
 * gvfs_dbus_mount_tracker_complete_mount_location:
 * @object: A #GVfsDBusMountTracker.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.MountLocation">MountLocation()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_tracker_complete_mount_location (
    GVfsDBusMountTracker *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_tracker_complete_list_mounts:
 * @object: A #GVfsDBusMountTracker.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @mounts: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.ListMounts">ListMounts()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_tracker_complete_list_mounts (
    GVfsDBusMountTracker *object,
    GDBusMethodInvocation *invocation,
    GVariant *mounts)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@a(sossssssbay(aya{sv})ay))",
                   mounts));
}

/**
 * gvfs_dbus_mount_tracker_complete_register_mount:
 * @object: A #GVfsDBusMountTracker.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.RegisterMount">RegisterMount()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_tracker_complete_register_mount (
    GVfsDBusMountTracker *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_tracker_complete_unregister_mount:
 * @object: A #GVfsDBusMountTracker.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.UnregisterMount">UnregisterMount()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_tracker_complete_unregister_mount (
    GVfsDBusMountTracker *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_tracker_complete_list_mount_types:
 * @object: A #GVfsDBusMountTracker.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @mount_types: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.ListMountTypes">ListMountTypes()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_tracker_complete_list_mount_types (
    GVfsDBusMountTracker *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *mount_types)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(^as)",
                   mount_types));
}

/**
 * gvfs_dbus_mount_tracker_complete_list_mountable_info:
 * @object: A #GVfsDBusMountTracker.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @mountables: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.ListMountableInfo">ListMountableInfo()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_tracker_complete_list_mountable_info (
    GVfsDBusMountTracker *object,
    GDBusMethodInvocation *invocation,
    GVariant *mountables)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@a(ssasib))",
                   mountables));
}

/**
 * gvfs_dbus_mount_tracker_complete_register_fuse:
 * @object: A #GVfsDBusMountTracker.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountTracker.RegisterFuse">RegisterFuse()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_tracker_complete_register_fuse (
    GVfsDBusMountTracker *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusMountTrackerProxy:
 *
 * The #GVfsDBusMountTrackerProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusMountTrackerProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusMountTrackerProxy.
 */

struct _GVfsDBusMountTrackerProxyPrivate
{
  GData *qdata;
};

static void gvfs_dbus_mount_tracker_proxy_iface_init (GVfsDBusMountTrackerIface *iface);

G_DEFINE_TYPE_WITH_CODE (GVfsDBusMountTrackerProxy, gvfs_dbus_mount_tracker_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (GVfsDBusMountTrackerProxy)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_MOUNT_TRACKER, gvfs_dbus_mount_tracker_proxy_iface_init));

static void
gvfs_dbus_mount_tracker_proxy_finalize (GObject *object)
{
  GVfsDBusMountTrackerProxy *proxy = GVFS_DBUS_MOUNT_TRACKER_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (gvfs_dbus_mount_tracker_proxy_parent_class)->finalize (object);
}

static void
gvfs_dbus_mount_tracker_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_mount_tracker_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_mount_tracker_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_tracker_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], GVFS_DBUS_TYPE_MOUNT_TRACKER);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_MOUNT_TRACKER);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
gvfs_dbus_mount_tracker_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  GVfsDBusMountTrackerProxy *proxy = GVFS_DBUS_MOUNT_TRACKER_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_tracker_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_tracker_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
gvfs_dbus_mount_tracker_proxy_init (GVfsDBusMountTrackerProxy *proxy)
{
  proxy->priv = gvfs_dbus_mount_tracker_proxy_get_instance_private (proxy);
  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), gvfs_dbus_mount_tracker_interface_info ());
}

static void
gvfs_dbus_mount_tracker_proxy_class_init (GVfsDBusMountTrackerProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = gvfs_dbus_mount_tracker_proxy_finalize;
  gobject_class->get_property = gvfs_dbus_mount_tracker_proxy_get_property;
  gobject_class->set_property = gvfs_dbus_mount_tracker_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = gvfs_dbus_mount_tracker_proxy_g_signal;
  proxy_class->g_properties_changed = gvfs_dbus_mount_tracker_proxy_g_properties_changed;

}

static void
gvfs_dbus_mount_tracker_proxy_iface_init (GVfsDBusMountTrackerIface *iface)
{
}

/**
 * gvfs_dbus_mount_tracker_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MountTracker.top_of_page">org.gtk.vfs.MountTracker</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_tracker_proxy_new_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_tracker_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_mount_tracker_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_MOUNT_TRACKER_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.MountTracker", NULL);
}

/**
 * gvfs_dbus_mount_tracker_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_tracker_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_mount_tracker_proxy_new().
 *
 * Returns: (transfer full) (type GVfsDBusMountTrackerProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMountTracker *
gvfs_dbus_mount_tracker_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_MOUNT_TRACKER (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_mount_tracker_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MountTracker.top_of_page">org.gtk.vfs.MountTracker</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_tracker_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusMountTrackerProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMountTracker *
gvfs_dbus_mount_tracker_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_MOUNT_TRACKER_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.MountTracker", NULL);
  if (ret != NULL)
    return GVFS_DBUS_MOUNT_TRACKER (ret);
  else
    return NULL;
}


/**
 * gvfs_dbus_mount_tracker_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like gvfs_dbus_mount_tracker_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_tracker_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_tracker_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_mount_tracker_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_MOUNT_TRACKER_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.MountTracker", NULL);
}

/**
 * gvfs_dbus_mount_tracker_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_tracker_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_mount_tracker_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type GVfsDBusMountTrackerProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMountTracker *
gvfs_dbus_mount_tracker_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_MOUNT_TRACKER (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_mount_tracker_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like gvfs_dbus_mount_tracker_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_tracker_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusMountTrackerProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMountTracker *
gvfs_dbus_mount_tracker_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_MOUNT_TRACKER_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.MountTracker", NULL);
  if (ret != NULL)
    return GVFS_DBUS_MOUNT_TRACKER (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusMountTrackerSkeleton:
 *
 * The #GVfsDBusMountTrackerSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusMountTrackerSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusMountTrackerSkeleton.
 */

struct _GVfsDBusMountTrackerSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_gvfs_dbus_mount_tracker_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  GVfsDBusMountTrackerSkeleton *skeleton = GVFS_DBUS_MOUNT_TRACKER_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint num_extra;
  guint n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], GVFS_DBUS_TYPE_MOUNT_TRACKER);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_MOUNT_TRACKER);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_gvfs_dbus_mount_tracker_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  GVfsDBusMountTrackerSkeleton *skeleton = GVFS_DBUS_MOUNT_TRACKER_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_tracker_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_gvfs_dbus_mount_tracker_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  GVfsDBusMountTrackerSkeleton *skeleton = GVFS_DBUS_MOUNT_TRACKER_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_tracker_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _gvfs_dbus_mount_tracker_skeleton_vtable =
{
  _gvfs_dbus_mount_tracker_skeleton_handle_method_call,
  _gvfs_dbus_mount_tracker_skeleton_handle_get_property,
  _gvfs_dbus_mount_tracker_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
gvfs_dbus_mount_tracker_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return gvfs_dbus_mount_tracker_interface_info ();
}

static GDBusInterfaceVTable *
gvfs_dbus_mount_tracker_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_gvfs_dbus_mount_tracker_skeleton_vtable;
}

static GVariant *
gvfs_dbus_mount_tracker_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  GVfsDBusMountTrackerSkeleton *skeleton = GVFS_DBUS_MOUNT_TRACKER_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  if (_gvfs_dbus_mount_tracker_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _gvfs_dbus_mount_tracker_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _gvfs_dbus_mount_tracker_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _gvfs_dbus_mount_tracker_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.vfs.MountTracker", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
gvfs_dbus_mount_tracker_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
}

static void
_gvfs_dbus_mount_tracker_on_signal_mounted (
    GVfsDBusMountTracker *object,
    GVariant *arg_mount)
{
  GVfsDBusMountTrackerSkeleton *skeleton = GVFS_DBUS_MOUNT_TRACKER_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("(@(sossssssbay(aya{sv})ay))",
                   arg_mount));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.vfs.MountTracker", "Mounted",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void
_gvfs_dbus_mount_tracker_on_signal_unmounted (
    GVfsDBusMountTracker *object,
    GVariant *arg_mount)
{
  GVfsDBusMountTrackerSkeleton *skeleton = GVFS_DBUS_MOUNT_TRACKER_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("(@(sossssssbay(aya{sv})ay))",
                   arg_mount));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.vfs.MountTracker", "Unmounted",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void gvfs_dbus_mount_tracker_skeleton_iface_init (GVfsDBusMountTrackerIface *iface);
G_DEFINE_TYPE_WITH_CODE (GVfsDBusMountTrackerSkeleton, gvfs_dbus_mount_tracker_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (GVfsDBusMountTrackerSkeleton)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_MOUNT_TRACKER, gvfs_dbus_mount_tracker_skeleton_iface_init));

static void
gvfs_dbus_mount_tracker_skeleton_finalize (GObject *object)
{
  GVfsDBusMountTrackerSkeleton *skeleton = GVFS_DBUS_MOUNT_TRACKER_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (gvfs_dbus_mount_tracker_skeleton_parent_class)->finalize (object);
}

static void
gvfs_dbus_mount_tracker_skeleton_init (GVfsDBusMountTrackerSkeleton *skeleton)
{
  skeleton->priv = gvfs_dbus_mount_tracker_skeleton_get_instance_private (skeleton);
  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
gvfs_dbus_mount_tracker_skeleton_class_init (GVfsDBusMountTrackerSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = gvfs_dbus_mount_tracker_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = gvfs_dbus_mount_tracker_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = gvfs_dbus_mount_tracker_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = gvfs_dbus_mount_tracker_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = gvfs_dbus_mount_tracker_skeleton_dbus_interface_get_vtable;
}

static void
gvfs_dbus_mount_tracker_skeleton_iface_init (GVfsDBusMountTrackerIface *iface)
{
  iface->mounted = _gvfs_dbus_mount_tracker_on_signal_mounted;
  iface->unmounted = _gvfs_dbus_mount_tracker_on_signal_unmounted;
}

/**
 * gvfs_dbus_mount_tracker_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MountTracker.top_of_page">org.gtk.vfs.MountTracker</link>.
 *
 * Returns: (transfer full) (type GVfsDBusMountTrackerSkeleton): The skeleton object.
 */
GVfsDBusMountTracker *
gvfs_dbus_mount_tracker_skeleton_new (void)
{
  return GVFS_DBUS_MOUNT_TRACKER (g_object_new (GVFS_DBUS_TYPE_MOUNT_TRACKER_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gtk.vfs.Mountable
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:GVfsDBusMountable
 * @title: GVfsDBusMountable
 * @short_description: Generated C code for the org.gtk.vfs.Mountable D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gtk-vfs-Mountable.top_of_page">org.gtk.vfs.Mountable</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gtk.vfs.Mountable ---- */

static const _ExtendedGDBusArgInfo _gvfs_dbus_mountable_method_info_mount_IN_ARG_mount_spec =
{
  {
    -1,
    (gchar *) "mount_spec",
    (gchar *) "(aya{sv})",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mountable_method_info_mount_IN_ARG_automount =
{
  {
    -1,
    (gchar *) "automount",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mountable_method_info_mount_IN_ARG_mount_source =
{
  {
    -1,
    (gchar *) "mount_source",
    (gchar *) "(so)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mountable_method_info_mount_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mountable_method_info_mount_IN_ARG_mount_spec,
  &_gvfs_dbus_mountable_method_info_mount_IN_ARG_automount,
  &_gvfs_dbus_mountable_method_info_mount_IN_ARG_mount_source,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mountable_method_info_mount =
{
  {
    -1,
    (gchar *) "Mount",
    (GDBusArgInfo **) &_gvfs_dbus_mountable_method_info_mount_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-mount",
  FALSE
};

static const _ExtendedGDBusMethodInfo * const _gvfs_dbus_mountable_method_info_pointers[] =
{
  &_gvfs_dbus_mountable_method_info_mount,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _gvfs_dbus_mountable_interface_info =
{
  {
    -1,
    (gchar *) "org.gtk.vfs.Mountable",
    (GDBusMethodInfo **) &_gvfs_dbus_mountable_method_info_pointers,
    NULL,
    NULL,
    NULL
  },
  "mountable",
};


/**
 * gvfs_dbus_mountable_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gtk-vfs-Mountable.top_of_page">org.gtk.vfs.Mountable</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
gvfs_dbus_mountable_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_gvfs_dbus_mountable_interface_info.parent_struct;
}

/**
 * gvfs_dbus_mountable_override_properties:
 * @klass: The class structure for a #GObject<!-- -->-derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #GVfsDBusMountable interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
gvfs_dbus_mountable_override_properties (GObjectClass *klass, guint property_id_begin)
{
  return property_id_begin - 1;
}



/**
 * GVfsDBusMountable:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Mountable.top_of_page">org.gtk.vfs.Mountable</link>.
 */

/**
 * GVfsDBusMountableIface:
 * @parent_iface: The parent interface.
 * @handle_mount: Handler for the #GVfsDBusMountable::handle-mount signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Mountable.top_of_page">org.gtk.vfs.Mountable</link>.
 */

typedef GVfsDBusMountableIface GVfsDBusMountableInterface;
G_DEFINE_INTERFACE (GVfsDBusMountable, gvfs_dbus_mountable, G_TYPE_OBJECT);

static void
gvfs_dbus_mountable_default_init (GVfsDBusMountableIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * GVfsDBusMountable::handle-mount:
   * @object: A #GVfsDBusMountable.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_mount_spec: Argument passed by remote caller.
   * @arg_automount: Argument passed by remote caller.
   * @arg_mount_source: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mountable.Mount">Mount()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mountable_complete_mount() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-mount",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountableIface, handle_mount),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT, G_TYPE_BOOLEAN, G_TYPE_VARIANT);

}

/**
 * gvfs_dbus_mountable_call_mount:
 * @proxy: A #GVfsDBusMountableProxy.
 * @arg_mount_spec: Argument to pass with the method invocation.
 * @arg_automount: Argument to pass with the method invocation.
 * @arg_mount_source: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mountable.Mount">Mount()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mountable_call_mount_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mountable_call_mount_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mountable_call_mount (
    GVfsDBusMountable *proxy,
    GVariant *arg_mount_spec,
    gboolean arg_automount,
    GVariant *arg_mount_source,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Mount",
    g_variant_new ("(@(aya{sv})b@(so))",
                   arg_mount_spec,
                   arg_automount,
                   arg_mount_source),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mountable_call_mount_finish:
 * @proxy: A #GVfsDBusMountableProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mountable_call_mount().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mountable_call_mount().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mountable_call_mount_finish (
    GVfsDBusMountable *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mountable_call_mount_sync:
 * @proxy: A #GVfsDBusMountableProxy.
 * @arg_mount_spec: Argument to pass with the method invocation.
 * @arg_automount: Argument to pass with the method invocation.
 * @arg_mount_source: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mountable.Mount">Mount()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mountable_call_mount() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mountable_call_mount_sync (
    GVfsDBusMountable *proxy,
    GVariant *arg_mount_spec,
    gboolean arg_automount,
    GVariant *arg_mount_source,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Mount",
    g_variant_new ("(@(aya{sv})b@(so))",
                   arg_mount_spec,
                   arg_automount,
                   arg_mount_source),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mountable_complete_mount:
 * @object: A #GVfsDBusMountable.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mountable.Mount">Mount()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mountable_complete_mount (
    GVfsDBusMountable *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusMountableProxy:
 *
 * The #GVfsDBusMountableProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusMountableProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusMountableProxy.
 */

struct _GVfsDBusMountableProxyPrivate
{
  GData *qdata;
};

static void gvfs_dbus_mountable_proxy_iface_init (GVfsDBusMountableIface *iface);

G_DEFINE_TYPE_WITH_CODE (GVfsDBusMountableProxy, gvfs_dbus_mountable_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (GVfsDBusMountableProxy)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_MOUNTABLE, gvfs_dbus_mountable_proxy_iface_init));

static void
gvfs_dbus_mountable_proxy_finalize (GObject *object)
{
  GVfsDBusMountableProxy *proxy = GVFS_DBUS_MOUNTABLE_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (gvfs_dbus_mountable_proxy_parent_class)->finalize (object);
}

static void
gvfs_dbus_mountable_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_mountable_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_mountable_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_gvfs_dbus_mountable_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], GVFS_DBUS_TYPE_MOUNTABLE);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_MOUNTABLE);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
gvfs_dbus_mountable_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  GVfsDBusMountableProxy *proxy = GVFS_DBUS_MOUNTABLE_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mountable_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mountable_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
gvfs_dbus_mountable_proxy_init (GVfsDBusMountableProxy *proxy)
{
  proxy->priv = gvfs_dbus_mountable_proxy_get_instance_private (proxy);
  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), gvfs_dbus_mountable_interface_info ());
}

static void
gvfs_dbus_mountable_proxy_class_init (GVfsDBusMountableProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = gvfs_dbus_mountable_proxy_finalize;
  gobject_class->get_property = gvfs_dbus_mountable_proxy_get_property;
  gobject_class->set_property = gvfs_dbus_mountable_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = gvfs_dbus_mountable_proxy_g_signal;
  proxy_class->g_properties_changed = gvfs_dbus_mountable_proxy_g_properties_changed;

}

static void
gvfs_dbus_mountable_proxy_iface_init (GVfsDBusMountableIface *iface)
{
}

/**
 * gvfs_dbus_mountable_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Mountable.top_of_page">org.gtk.vfs.Mountable</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mountable_proxy_new_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mountable_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_mountable_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_MOUNTABLE_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Mountable", NULL);
}

/**
 * gvfs_dbus_mountable_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mountable_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_mountable_proxy_new().
 *
 * Returns: (transfer full) (type GVfsDBusMountableProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMountable *
gvfs_dbus_mountable_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_MOUNTABLE (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_mountable_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Mountable.top_of_page">org.gtk.vfs.Mountable</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mountable_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusMountableProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMountable *
gvfs_dbus_mountable_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_MOUNTABLE_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Mountable", NULL);
  if (ret != NULL)
    return GVFS_DBUS_MOUNTABLE (ret);
  else
    return NULL;
}


/**
 * gvfs_dbus_mountable_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like gvfs_dbus_mountable_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mountable_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mountable_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_mountable_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_MOUNTABLE_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Mountable", NULL);
}

/**
 * gvfs_dbus_mountable_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mountable_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_mountable_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type GVfsDBusMountableProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMountable *
gvfs_dbus_mountable_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_MOUNTABLE (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_mountable_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like gvfs_dbus_mountable_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mountable_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusMountableProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMountable *
gvfs_dbus_mountable_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_MOUNTABLE_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Mountable", NULL);
  if (ret != NULL)
    return GVFS_DBUS_MOUNTABLE (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusMountableSkeleton:
 *
 * The #GVfsDBusMountableSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusMountableSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusMountableSkeleton.
 */

struct _GVfsDBusMountableSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_gvfs_dbus_mountable_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  GVfsDBusMountableSkeleton *skeleton = GVFS_DBUS_MOUNTABLE_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint num_extra;
  guint n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], GVFS_DBUS_TYPE_MOUNTABLE);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_MOUNTABLE);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_gvfs_dbus_mountable_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  GVfsDBusMountableSkeleton *skeleton = GVFS_DBUS_MOUNTABLE_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mountable_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_gvfs_dbus_mountable_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  GVfsDBusMountableSkeleton *skeleton = GVFS_DBUS_MOUNTABLE_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mountable_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _gvfs_dbus_mountable_skeleton_vtable =
{
  _gvfs_dbus_mountable_skeleton_handle_method_call,
  _gvfs_dbus_mountable_skeleton_handle_get_property,
  _gvfs_dbus_mountable_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
gvfs_dbus_mountable_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return gvfs_dbus_mountable_interface_info ();
}

static GDBusInterfaceVTable *
gvfs_dbus_mountable_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_gvfs_dbus_mountable_skeleton_vtable;
}

static GVariant *
gvfs_dbus_mountable_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  GVfsDBusMountableSkeleton *skeleton = GVFS_DBUS_MOUNTABLE_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  if (_gvfs_dbus_mountable_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _gvfs_dbus_mountable_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _gvfs_dbus_mountable_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _gvfs_dbus_mountable_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.vfs.Mountable", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
gvfs_dbus_mountable_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
}

static void gvfs_dbus_mountable_skeleton_iface_init (GVfsDBusMountableIface *iface);
G_DEFINE_TYPE_WITH_CODE (GVfsDBusMountableSkeleton, gvfs_dbus_mountable_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (GVfsDBusMountableSkeleton)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_MOUNTABLE, gvfs_dbus_mountable_skeleton_iface_init));

static void
gvfs_dbus_mountable_skeleton_finalize (GObject *object)
{
  GVfsDBusMountableSkeleton *skeleton = GVFS_DBUS_MOUNTABLE_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (gvfs_dbus_mountable_skeleton_parent_class)->finalize (object);
}

static void
gvfs_dbus_mountable_skeleton_init (GVfsDBusMountableSkeleton *skeleton)
{
  skeleton->priv = gvfs_dbus_mountable_skeleton_get_instance_private (skeleton);
  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
gvfs_dbus_mountable_skeleton_class_init (GVfsDBusMountableSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = gvfs_dbus_mountable_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = gvfs_dbus_mountable_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = gvfs_dbus_mountable_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = gvfs_dbus_mountable_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = gvfs_dbus_mountable_skeleton_dbus_interface_get_vtable;
}

static void
gvfs_dbus_mountable_skeleton_iface_init (GVfsDBusMountableIface *iface)
{
}

/**
 * gvfs_dbus_mountable_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Mountable.top_of_page">org.gtk.vfs.Mountable</link>.
 *
 * Returns: (transfer full) (type GVfsDBusMountableSkeleton): The skeleton object.
 */
GVfsDBusMountable *
gvfs_dbus_mountable_skeleton_new (void)
{
  return GVFS_DBUS_MOUNTABLE (g_object_new (GVFS_DBUS_TYPE_MOUNTABLE_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gtk.vfs.MountOperation
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:GVfsDBusMountOperation
 * @title: GVfsDBusMountOperation
 * @short_description: Generated C code for the org.gtk.vfs.MountOperation D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gtk-vfs-MountOperation.top_of_page">org.gtk.vfs.MountOperation</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gtk.vfs.MountOperation ---- */

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_password_IN_ARG_message_string =
{
  {
    -1,
    (gchar *) "message_string",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_password_IN_ARG_default_user =
{
  {
    -1,
    (gchar *) "default_user",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_password_IN_ARG_default_domain =
{
  {
    -1,
    (gchar *) "default_domain",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_password_IN_ARG_flags_as_int =
{
  {
    -1,
    (gchar *) "flags_as_int",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_operation_method_info_ask_password_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_operation_method_info_ask_password_IN_ARG_message_string,
  &_gvfs_dbus_mount_operation_method_info_ask_password_IN_ARG_default_user,
  &_gvfs_dbus_mount_operation_method_info_ask_password_IN_ARG_default_domain,
  &_gvfs_dbus_mount_operation_method_info_ask_password_IN_ARG_flags_as_int,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_handled =
{
  {
    -1,
    (gchar *) "handled",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_aborted =
{
  {
    -1,
    (gchar *) "aborted",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_password =
{
  {
    -1,
    (gchar *) "password",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_username =
{
  {
    -1,
    (gchar *) "username",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_domain =
{
  {
    -1,
    (gchar *) "domain",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_anonymous =
{
  {
    -1,
    (gchar *) "anonymous",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_password_save =
{
  {
    -1,
    (gchar *) "password_save",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_handled,
  &_gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_aborted,
  &_gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_password,
  &_gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_username,
  &_gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_domain,
  &_gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_anonymous,
  &_gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_password_save,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_operation_method_info_ask_password =
{
  {
    -1,
    (gchar *) "AskPassword",
    (GDBusArgInfo **) &_gvfs_dbus_mount_operation_method_info_ask_password_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_operation_method_info_ask_password_OUT_ARG_pointers,
    NULL
  },
  "handle-ask-password",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_question_IN_ARG_message_string =
{
  {
    -1,
    (gchar *) "message_string",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_question_IN_ARG_choices =
{
  {
    -1,
    (gchar *) "choices",
    (gchar *) "as",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_operation_method_info_ask_question_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_operation_method_info_ask_question_IN_ARG_message_string,
  &_gvfs_dbus_mount_operation_method_info_ask_question_IN_ARG_choices,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_question_OUT_ARG_handled =
{
  {
    -1,
    (gchar *) "handled",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_question_OUT_ARG_aborted =
{
  {
    -1,
    (gchar *) "aborted",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_ask_question_OUT_ARG_choice =
{
  {
    -1,
    (gchar *) "choice",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_operation_method_info_ask_question_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_operation_method_info_ask_question_OUT_ARG_handled,
  &_gvfs_dbus_mount_operation_method_info_ask_question_OUT_ARG_aborted,
  &_gvfs_dbus_mount_operation_method_info_ask_question_OUT_ARG_choice,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_operation_method_info_ask_question =
{
  {
    -1,
    (gchar *) "AskQuestion",
    (GDBusArgInfo **) &_gvfs_dbus_mount_operation_method_info_ask_question_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_operation_method_info_ask_question_OUT_ARG_pointers,
    NULL
  },
  "handle-ask-question",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_show_processes_IN_ARG_message_string =
{
  {
    -1,
    (gchar *) "message_string",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_show_processes_IN_ARG_choices =
{
  {
    -1,
    (gchar *) "choices",
    (gchar *) "as",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_show_processes_IN_ARG_processes =
{
  {
    -1,
    (gchar *) "processes",
    (gchar *) "ai",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_operation_method_info_show_processes_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_operation_method_info_show_processes_IN_ARG_message_string,
  &_gvfs_dbus_mount_operation_method_info_show_processes_IN_ARG_choices,
  &_gvfs_dbus_mount_operation_method_info_show_processes_IN_ARG_processes,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_show_processes_OUT_ARG_handled =
{
  {
    -1,
    (gchar *) "handled",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_show_processes_OUT_ARG_aborted =
{
  {
    -1,
    (gchar *) "aborted",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_show_processes_OUT_ARG_choice =
{
  {
    -1,
    (gchar *) "choice",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_operation_method_info_show_processes_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_operation_method_info_show_processes_OUT_ARG_handled,
  &_gvfs_dbus_mount_operation_method_info_show_processes_OUT_ARG_aborted,
  &_gvfs_dbus_mount_operation_method_info_show_processes_OUT_ARG_choice,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_operation_method_info_show_processes =
{
  {
    -1,
    (gchar *) "ShowProcesses",
    (GDBusArgInfo **) &_gvfs_dbus_mount_operation_method_info_show_processes_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_operation_method_info_show_processes_OUT_ARG_pointers,
    NULL
  },
  "handle-show-processes",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_show_unmount_progress_IN_ARG_message_string =
{
  {
    -1,
    (gchar *) "message_string",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_show_unmount_progress_IN_ARG_time_left =
{
  {
    -1,
    (gchar *) "time_left",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_operation_method_info_show_unmount_progress_IN_ARG_bytes_left =
{
  {
    -1,
    (gchar *) "bytes_left",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_operation_method_info_show_unmount_progress_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_operation_method_info_show_unmount_progress_IN_ARG_message_string,
  &_gvfs_dbus_mount_operation_method_info_show_unmount_progress_IN_ARG_time_left,
  &_gvfs_dbus_mount_operation_method_info_show_unmount_progress_IN_ARG_bytes_left,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_operation_method_info_show_unmount_progress =
{
  {
    -1,
    (gchar *) "ShowUnmountProgress",
    (GDBusArgInfo **) &_gvfs_dbus_mount_operation_method_info_show_unmount_progress_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-show-unmount-progress",
  FALSE
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_operation_method_info_aborted =
{
  {
    -1,
    (gchar *) "Aborted",
    NULL,
    NULL,
    NULL
  },
  "handle-aborted",
  FALSE
};

static const _ExtendedGDBusMethodInfo * const _gvfs_dbus_mount_operation_method_info_pointers[] =
{
  &_gvfs_dbus_mount_operation_method_info_ask_password,
  &_gvfs_dbus_mount_operation_method_info_ask_question,
  &_gvfs_dbus_mount_operation_method_info_show_processes,
  &_gvfs_dbus_mount_operation_method_info_show_unmount_progress,
  &_gvfs_dbus_mount_operation_method_info_aborted,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _gvfs_dbus_mount_operation_interface_info =
{
  {
    -1,
    (gchar *) "org.gtk.vfs.MountOperation",
    (GDBusMethodInfo **) &_gvfs_dbus_mount_operation_method_info_pointers,
    NULL,
    NULL,
    NULL
  },
  "mount-operation",
};


/**
 * gvfs_dbus_mount_operation_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gtk-vfs-MountOperation.top_of_page">org.gtk.vfs.MountOperation</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
gvfs_dbus_mount_operation_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_gvfs_dbus_mount_operation_interface_info.parent_struct;
}

/**
 * gvfs_dbus_mount_operation_override_properties:
 * @klass: The class structure for a #GObject<!-- -->-derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #GVfsDBusMountOperation interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
gvfs_dbus_mount_operation_override_properties (GObjectClass *klass, guint property_id_begin)
{
  return property_id_begin - 1;
}



/**
 * GVfsDBusMountOperation:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MountOperation.top_of_page">org.gtk.vfs.MountOperation</link>.
 */

/**
 * GVfsDBusMountOperationIface:
 * @parent_iface: The parent interface.
 * @handle_aborted: Handler for the #GVfsDBusMountOperation::handle-aborted signal.
 * @handle_ask_password: Handler for the #GVfsDBusMountOperation::handle-ask-password signal.
 * @handle_ask_question: Handler for the #GVfsDBusMountOperation::handle-ask-question signal.
 * @handle_show_processes: Handler for the #GVfsDBusMountOperation::handle-show-processes signal.
 * @handle_show_unmount_progress: Handler for the #GVfsDBusMountOperation::handle-show-unmount-progress signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MountOperation.top_of_page">org.gtk.vfs.MountOperation</link>.
 */

typedef GVfsDBusMountOperationIface GVfsDBusMountOperationInterface;
G_DEFINE_INTERFACE (GVfsDBusMountOperation, gvfs_dbus_mount_operation, G_TYPE_OBJECT);

static void
gvfs_dbus_mount_operation_default_init (GVfsDBusMountOperationIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * GVfsDBusMountOperation::handle-ask-password:
   * @object: A #GVfsDBusMountOperation.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_message_string: Argument passed by remote caller.
   * @arg_default_user: Argument passed by remote caller.
   * @arg_default_domain: Argument passed by remote caller.
   * @arg_flags_as_int: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.AskPassword">AskPassword()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_operation_complete_ask_password() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-ask-password",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountOperationIface, handle_ask_password),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT);

  /**
   * GVfsDBusMountOperation::handle-ask-question:
   * @object: A #GVfsDBusMountOperation.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_message_string: Argument passed by remote caller.
   * @arg_choices: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.AskQuestion">AskQuestion()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_operation_complete_ask_question() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-ask-question",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountOperationIface, handle_ask_question),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRV);

  /**
   * GVfsDBusMountOperation::handle-show-processes:
   * @object: A #GVfsDBusMountOperation.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_message_string: Argument passed by remote caller.
   * @arg_choices: Argument passed by remote caller.
   * @arg_processes: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.ShowProcesses">ShowProcesses()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_operation_complete_show_processes() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-show-processes",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountOperationIface, handle_show_processes),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRV, G_TYPE_VARIANT);

  /**
   * GVfsDBusMountOperation::handle-show-unmount-progress:
   * @object: A #GVfsDBusMountOperation.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_message_string: Argument passed by remote caller.
   * @arg_time_left: Argument passed by remote caller.
   * @arg_bytes_left: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.ShowUnmountProgress">ShowUnmountProgress()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_operation_complete_show_unmount_progress() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-show-unmount-progress",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountOperationIface, handle_show_unmount_progress),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_INT64, G_TYPE_INT64);

  /**
   * GVfsDBusMountOperation::handle-aborted:
   * @object: A #GVfsDBusMountOperation.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.Aborted">Aborted()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_operation_complete_aborted() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-aborted",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountOperationIface, handle_aborted),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

}

/**
 * gvfs_dbus_mount_operation_call_ask_password:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @arg_message_string: Argument to pass with the method invocation.
 * @arg_default_user: Argument to pass with the method invocation.
 * @arg_default_domain: Argument to pass with the method invocation.
 * @arg_flags_as_int: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.AskPassword">AskPassword()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_operation_call_ask_password_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_operation_call_ask_password_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_operation_call_ask_password (
    GVfsDBusMountOperation *proxy,
    const gchar *arg_message_string,
    const gchar *arg_default_user,
    const gchar *arg_default_domain,
    guint arg_flags_as_int,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "AskPassword",
    g_variant_new ("(sssu)",
                   arg_message_string,
                   arg_default_user,
                   arg_default_domain,
                   arg_flags_as_int),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_operation_call_ask_password_finish:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @out_handled: (out): Return location for return parameter or %NULL to ignore.
 * @out_aborted: (out): Return location for return parameter or %NULL to ignore.
 * @out_password: (out): Return location for return parameter or %NULL to ignore.
 * @out_username: (out): Return location for return parameter or %NULL to ignore.
 * @out_domain: (out): Return location for return parameter or %NULL to ignore.
 * @out_anonymous: (out): Return location for return parameter or %NULL to ignore.
 * @out_password_save: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_operation_call_ask_password().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_operation_call_ask_password().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_operation_call_ask_password_finish (
    GVfsDBusMountOperation *proxy,
    gboolean *out_handled,
    gboolean *out_aborted,
    gchar **out_password,
    gchar **out_username,
    gchar **out_domain,
    gboolean *out_anonymous,
    guint *out_password_save,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bbsssbu)",
                 out_handled,
                 out_aborted,
                 out_password,
                 out_username,
                 out_domain,
                 out_anonymous,
                 out_password_save);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_operation_call_ask_password_sync:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @arg_message_string: Argument to pass with the method invocation.
 * @arg_default_user: Argument to pass with the method invocation.
 * @arg_default_domain: Argument to pass with the method invocation.
 * @arg_flags_as_int: Argument to pass with the method invocation.
 * @out_handled: (out): Return location for return parameter or %NULL to ignore.
 * @out_aborted: (out): Return location for return parameter or %NULL to ignore.
 * @out_password: (out): Return location for return parameter or %NULL to ignore.
 * @out_username: (out): Return location for return parameter or %NULL to ignore.
 * @out_domain: (out): Return location for return parameter or %NULL to ignore.
 * @out_anonymous: (out): Return location for return parameter or %NULL to ignore.
 * @out_password_save: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.AskPassword">AskPassword()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_operation_call_ask_password() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_operation_call_ask_password_sync (
    GVfsDBusMountOperation *proxy,
    const gchar *arg_message_string,
    const gchar *arg_default_user,
    const gchar *arg_default_domain,
    guint arg_flags_as_int,
    gboolean *out_handled,
    gboolean *out_aborted,
    gchar **out_password,
    gchar **out_username,
    gchar **out_domain,
    gboolean *out_anonymous,
    guint *out_password_save,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "AskPassword",
    g_variant_new ("(sssu)",
                   arg_message_string,
                   arg_default_user,
                   arg_default_domain,
                   arg_flags_as_int),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bbsssbu)",
                 out_handled,
                 out_aborted,
                 out_password,
                 out_username,
                 out_domain,
                 out_anonymous,
                 out_password_save);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_operation_call_ask_question:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @arg_message_string: Argument to pass with the method invocation.
 * @arg_choices: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.AskQuestion">AskQuestion()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_operation_call_ask_question_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_operation_call_ask_question_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_operation_call_ask_question (
    GVfsDBusMountOperation *proxy,
    const gchar *arg_message_string,
    const gchar *const *arg_choices,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "AskQuestion",
    g_variant_new ("(s^as)",
                   arg_message_string,
                   arg_choices),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_operation_call_ask_question_finish:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @out_handled: (out): Return location for return parameter or %NULL to ignore.
 * @out_aborted: (out): Return location for return parameter or %NULL to ignore.
 * @out_choice: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_operation_call_ask_question().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_operation_call_ask_question().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_operation_call_ask_question_finish (
    GVfsDBusMountOperation *proxy,
    gboolean *out_handled,
    gboolean *out_aborted,
    guint *out_choice,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bbu)",
                 out_handled,
                 out_aborted,
                 out_choice);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_operation_call_ask_question_sync:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @arg_message_string: Argument to pass with the method invocation.
 * @arg_choices: Argument to pass with the method invocation.
 * @out_handled: (out): Return location for return parameter or %NULL to ignore.
 * @out_aborted: (out): Return location for return parameter or %NULL to ignore.
 * @out_choice: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.AskQuestion">AskQuestion()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_operation_call_ask_question() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_operation_call_ask_question_sync (
    GVfsDBusMountOperation *proxy,
    const gchar *arg_message_string,
    const gchar *const *arg_choices,
    gboolean *out_handled,
    gboolean *out_aborted,
    guint *out_choice,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "AskQuestion",
    g_variant_new ("(s^as)",
                   arg_message_string,
                   arg_choices),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bbu)",
                 out_handled,
                 out_aborted,
                 out_choice);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_operation_call_show_processes:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @arg_message_string: Argument to pass with the method invocation.
 * @arg_choices: Argument to pass with the method invocation.
 * @arg_processes: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.ShowProcesses">ShowProcesses()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_operation_call_show_processes_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_operation_call_show_processes_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_operation_call_show_processes (
    GVfsDBusMountOperation *proxy,
    const gchar *arg_message_string,
    const gchar *const *arg_choices,
    GVariant *arg_processes,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ShowProcesses",
    g_variant_new ("(s^as@ai)",
                   arg_message_string,
                   arg_choices,
                   arg_processes),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_operation_call_show_processes_finish:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @out_handled: (out): Return location for return parameter or %NULL to ignore.
 * @out_aborted: (out): Return location for return parameter or %NULL to ignore.
 * @out_choice: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_operation_call_show_processes().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_operation_call_show_processes().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_operation_call_show_processes_finish (
    GVfsDBusMountOperation *proxy,
    gboolean *out_handled,
    gboolean *out_aborted,
    guint *out_choice,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bbu)",
                 out_handled,
                 out_aborted,
                 out_choice);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_operation_call_show_processes_sync:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @arg_message_string: Argument to pass with the method invocation.
 * @arg_choices: Argument to pass with the method invocation.
 * @arg_processes: Argument to pass with the method invocation.
 * @out_handled: (out): Return location for return parameter or %NULL to ignore.
 * @out_aborted: (out): Return location for return parameter or %NULL to ignore.
 * @out_choice: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.ShowProcesses">ShowProcesses()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_operation_call_show_processes() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_operation_call_show_processes_sync (
    GVfsDBusMountOperation *proxy,
    const gchar *arg_message_string,
    const gchar *const *arg_choices,
    GVariant *arg_processes,
    gboolean *out_handled,
    gboolean *out_aborted,
    guint *out_choice,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ShowProcesses",
    g_variant_new ("(s^as@ai)",
                   arg_message_string,
                   arg_choices,
                   arg_processes),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bbu)",
                 out_handled,
                 out_aborted,
                 out_choice);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_operation_call_show_unmount_progress:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @arg_message_string: Argument to pass with the method invocation.
 * @arg_time_left: Argument to pass with the method invocation.
 * @arg_bytes_left: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.ShowUnmountProgress">ShowUnmountProgress()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_operation_call_show_unmount_progress_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_operation_call_show_unmount_progress_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_operation_call_show_unmount_progress (
    GVfsDBusMountOperation *proxy,
    const gchar *arg_message_string,
    gint64 arg_time_left,
    gint64 arg_bytes_left,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ShowUnmountProgress",
    g_variant_new ("(sxx)",
                   arg_message_string,
                   arg_time_left,
                   arg_bytes_left),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_operation_call_show_unmount_progress_finish:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_operation_call_show_unmount_progress().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_operation_call_show_unmount_progress().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_operation_call_show_unmount_progress_finish (
    GVfsDBusMountOperation *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_operation_call_show_unmount_progress_sync:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @arg_message_string: Argument to pass with the method invocation.
 * @arg_time_left: Argument to pass with the method invocation.
 * @arg_bytes_left: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.ShowUnmountProgress">ShowUnmountProgress()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_operation_call_show_unmount_progress() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_operation_call_show_unmount_progress_sync (
    GVfsDBusMountOperation *proxy,
    const gchar *arg_message_string,
    gint64 arg_time_left,
    gint64 arg_bytes_left,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ShowUnmountProgress",
    g_variant_new ("(sxx)",
                   arg_message_string,
                   arg_time_left,
                   arg_bytes_left),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_operation_call_aborted:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.Aborted">Aborted()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_operation_call_aborted_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_operation_call_aborted_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_operation_call_aborted (
    GVfsDBusMountOperation *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Aborted",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_operation_call_aborted_finish:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_operation_call_aborted().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_operation_call_aborted().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_operation_call_aborted_finish (
    GVfsDBusMountOperation *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_operation_call_aborted_sync:
 * @proxy: A #GVfsDBusMountOperationProxy.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.Aborted">Aborted()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_operation_call_aborted() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_operation_call_aborted_sync (
    GVfsDBusMountOperation *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Aborted",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_operation_complete_ask_password:
 * @object: A #GVfsDBusMountOperation.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @handled: Parameter to return.
 * @aborted: Parameter to return.
 * @password: Parameter to return.
 * @username: Parameter to return.
 * @domain: Parameter to return.
 * @anonymous: Parameter to return.
 * @password_save: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.AskPassword">AskPassword()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_operation_complete_ask_password (
    GVfsDBusMountOperation *object,
    GDBusMethodInvocation *invocation,
    gboolean handled,
    gboolean aborted,
    const gchar *password,
    const gchar *username,
    const gchar *domain,
    gboolean anonymous,
    guint password_save)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(bbsssbu)",
                   handled,
                   aborted,
                   password,
                   username,
                   domain,
                   anonymous,
                   password_save));
}

/**
 * gvfs_dbus_mount_operation_complete_ask_question:
 * @object: A #GVfsDBusMountOperation.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @handled: Parameter to return.
 * @aborted: Parameter to return.
 * @choice: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.AskQuestion">AskQuestion()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_operation_complete_ask_question (
    GVfsDBusMountOperation *object,
    GDBusMethodInvocation *invocation,
    gboolean handled,
    gboolean aborted,
    guint choice)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(bbu)",
                   handled,
                   aborted,
                   choice));
}

/**
 * gvfs_dbus_mount_operation_complete_show_processes:
 * @object: A #GVfsDBusMountOperation.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @handled: Parameter to return.
 * @aborted: Parameter to return.
 * @choice: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.ShowProcesses">ShowProcesses()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_operation_complete_show_processes (
    GVfsDBusMountOperation *object,
    GDBusMethodInvocation *invocation,
    gboolean handled,
    gboolean aborted,
    guint choice)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(bbu)",
                   handled,
                   aborted,
                   choice));
}

/**
 * gvfs_dbus_mount_operation_complete_show_unmount_progress:
 * @object: A #GVfsDBusMountOperation.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.ShowUnmountProgress">ShowUnmountProgress()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_operation_complete_show_unmount_progress (
    GVfsDBusMountOperation *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_operation_complete_aborted:
 * @object: A #GVfsDBusMountOperation.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MountOperation.Aborted">Aborted()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_operation_complete_aborted (
    GVfsDBusMountOperation *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusMountOperationProxy:
 *
 * The #GVfsDBusMountOperationProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusMountOperationProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusMountOperationProxy.
 */

struct _GVfsDBusMountOperationProxyPrivate
{
  GData *qdata;
};

static void gvfs_dbus_mount_operation_proxy_iface_init (GVfsDBusMountOperationIface *iface);

G_DEFINE_TYPE_WITH_CODE (GVfsDBusMountOperationProxy, gvfs_dbus_mount_operation_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (GVfsDBusMountOperationProxy)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_MOUNT_OPERATION, gvfs_dbus_mount_operation_proxy_iface_init));

static void
gvfs_dbus_mount_operation_proxy_finalize (GObject *object)
{
  GVfsDBusMountOperationProxy *proxy = GVFS_DBUS_MOUNT_OPERATION_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (gvfs_dbus_mount_operation_proxy_parent_class)->finalize (object);
}

static void
gvfs_dbus_mount_operation_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_mount_operation_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_mount_operation_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_operation_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], GVFS_DBUS_TYPE_MOUNT_OPERATION);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_MOUNT_OPERATION);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
gvfs_dbus_mount_operation_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  GVfsDBusMountOperationProxy *proxy = GVFS_DBUS_MOUNT_OPERATION_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_operation_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_operation_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
gvfs_dbus_mount_operation_proxy_init (GVfsDBusMountOperationProxy *proxy)
{
  proxy->priv = gvfs_dbus_mount_operation_proxy_get_instance_private (proxy);
  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), gvfs_dbus_mount_operation_interface_info ());
}

static void
gvfs_dbus_mount_operation_proxy_class_init (GVfsDBusMountOperationProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = gvfs_dbus_mount_operation_proxy_finalize;
  gobject_class->get_property = gvfs_dbus_mount_operation_proxy_get_property;
  gobject_class->set_property = gvfs_dbus_mount_operation_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = gvfs_dbus_mount_operation_proxy_g_signal;
  proxy_class->g_properties_changed = gvfs_dbus_mount_operation_proxy_g_properties_changed;

}

static void
gvfs_dbus_mount_operation_proxy_iface_init (GVfsDBusMountOperationIface *iface)
{
}

/**
 * gvfs_dbus_mount_operation_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MountOperation.top_of_page">org.gtk.vfs.MountOperation</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_operation_proxy_new_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_operation_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_mount_operation_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_MOUNT_OPERATION_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.MountOperation", NULL);
}

/**
 * gvfs_dbus_mount_operation_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_operation_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_mount_operation_proxy_new().
 *
 * Returns: (transfer full) (type GVfsDBusMountOperationProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMountOperation *
gvfs_dbus_mount_operation_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_MOUNT_OPERATION (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_mount_operation_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MountOperation.top_of_page">org.gtk.vfs.MountOperation</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_operation_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusMountOperationProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMountOperation *
gvfs_dbus_mount_operation_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_MOUNT_OPERATION_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.MountOperation", NULL);
  if (ret != NULL)
    return GVFS_DBUS_MOUNT_OPERATION (ret);
  else
    return NULL;
}


/**
 * gvfs_dbus_mount_operation_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like gvfs_dbus_mount_operation_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_operation_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_operation_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_mount_operation_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_MOUNT_OPERATION_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.MountOperation", NULL);
}

/**
 * gvfs_dbus_mount_operation_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_operation_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_mount_operation_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type GVfsDBusMountOperationProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMountOperation *
gvfs_dbus_mount_operation_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_MOUNT_OPERATION (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_mount_operation_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like gvfs_dbus_mount_operation_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_operation_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusMountOperationProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMountOperation *
gvfs_dbus_mount_operation_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_MOUNT_OPERATION_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.MountOperation", NULL);
  if (ret != NULL)
    return GVFS_DBUS_MOUNT_OPERATION (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusMountOperationSkeleton:
 *
 * The #GVfsDBusMountOperationSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusMountOperationSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusMountOperationSkeleton.
 */

struct _GVfsDBusMountOperationSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_gvfs_dbus_mount_operation_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  GVfsDBusMountOperationSkeleton *skeleton = GVFS_DBUS_MOUNT_OPERATION_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint num_extra;
  guint n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], GVFS_DBUS_TYPE_MOUNT_OPERATION);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_MOUNT_OPERATION);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_gvfs_dbus_mount_operation_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  GVfsDBusMountOperationSkeleton *skeleton = GVFS_DBUS_MOUNT_OPERATION_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_operation_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_gvfs_dbus_mount_operation_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  GVfsDBusMountOperationSkeleton *skeleton = GVFS_DBUS_MOUNT_OPERATION_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_operation_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _gvfs_dbus_mount_operation_skeleton_vtable =
{
  _gvfs_dbus_mount_operation_skeleton_handle_method_call,
  _gvfs_dbus_mount_operation_skeleton_handle_get_property,
  _gvfs_dbus_mount_operation_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
gvfs_dbus_mount_operation_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return gvfs_dbus_mount_operation_interface_info ();
}

static GDBusInterfaceVTable *
gvfs_dbus_mount_operation_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_gvfs_dbus_mount_operation_skeleton_vtable;
}

static GVariant *
gvfs_dbus_mount_operation_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  GVfsDBusMountOperationSkeleton *skeleton = GVFS_DBUS_MOUNT_OPERATION_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  if (_gvfs_dbus_mount_operation_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _gvfs_dbus_mount_operation_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _gvfs_dbus_mount_operation_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _gvfs_dbus_mount_operation_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.vfs.MountOperation", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
gvfs_dbus_mount_operation_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
}

static void gvfs_dbus_mount_operation_skeleton_iface_init (GVfsDBusMountOperationIface *iface);
G_DEFINE_TYPE_WITH_CODE (GVfsDBusMountOperationSkeleton, gvfs_dbus_mount_operation_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (GVfsDBusMountOperationSkeleton)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_MOUNT_OPERATION, gvfs_dbus_mount_operation_skeleton_iface_init));

static void
gvfs_dbus_mount_operation_skeleton_finalize (GObject *object)
{
  GVfsDBusMountOperationSkeleton *skeleton = GVFS_DBUS_MOUNT_OPERATION_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (gvfs_dbus_mount_operation_skeleton_parent_class)->finalize (object);
}

static void
gvfs_dbus_mount_operation_skeleton_init (GVfsDBusMountOperationSkeleton *skeleton)
{
  skeleton->priv = gvfs_dbus_mount_operation_skeleton_get_instance_private (skeleton);
  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
gvfs_dbus_mount_operation_skeleton_class_init (GVfsDBusMountOperationSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = gvfs_dbus_mount_operation_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = gvfs_dbus_mount_operation_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = gvfs_dbus_mount_operation_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = gvfs_dbus_mount_operation_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = gvfs_dbus_mount_operation_skeleton_dbus_interface_get_vtable;
}

static void
gvfs_dbus_mount_operation_skeleton_iface_init (GVfsDBusMountOperationIface *iface)
{
}

/**
 * gvfs_dbus_mount_operation_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MountOperation.top_of_page">org.gtk.vfs.MountOperation</link>.
 *
 * Returns: (transfer full) (type GVfsDBusMountOperationSkeleton): The skeleton object.
 */
GVfsDBusMountOperation *
gvfs_dbus_mount_operation_skeleton_new (void)
{
  return GVFS_DBUS_MOUNT_OPERATION (g_object_new (GVFS_DBUS_TYPE_MOUNT_OPERATION_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gtk.vfs.Mount
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:GVfsDBusMount
 * @title: GVfsDBusMount
 * @short_description: Generated C code for the org.gtk.vfs.Mount D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gtk-vfs-Mount.top_of_page">org.gtk.vfs.Mount</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gtk.vfs.Mount ---- */

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_unmount_IN_ARG_dbus_id =
{
  {
    -1,
    (gchar *) "dbus_id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_unmount_IN_ARG_obj_path =
{
  {
    -1,
    (gchar *) "obj_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_unmount_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_unmount_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_unmount_IN_ARG_dbus_id,
  &_gvfs_dbus_mount_method_info_unmount_IN_ARG_obj_path,
  &_gvfs_dbus_mount_method_info_unmount_IN_ARG_flags,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_unmount =
{
  {
    -1,
    (gchar *) "Unmount",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_unmount_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-unmount",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_for_read_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_for_read_IN_ARG_pid =
{
  {
    -1,
    (gchar *) "pid",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_open_for_read_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_open_for_read_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_open_for_read_IN_ARG_pid,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_for_read_OUT_ARG_fd_id =
{
  {
    -1,
    (gchar *) "fd_id",
    (gchar *) "h",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_for_read_OUT_ARG_can_seek =
{
  {
    -1,
    (gchar *) "can_seek",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_open_for_read_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_open_for_read_OUT_ARG_fd_id,
  &_gvfs_dbus_mount_method_info_open_for_read_OUT_ARG_can_seek,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_open_for_read =
{
  {
    -1,
    (gchar *) "OpenForRead",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_open_for_read_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_open_for_read_OUT_ARG_pointers,
    NULL
  },
  "handle-open-for-read",
  TRUE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_for_write_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_for_write_IN_ARG_mode =
{
  {
    -1,
    (gchar *) "mode",
    (gchar *) "q",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_for_write_IN_ARG_etag =
{
  {
    -1,
    (gchar *) "etag",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_for_write_IN_ARG_make_backup =
{
  {
    -1,
    (gchar *) "make_backup",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_for_write_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_for_write_IN_ARG_pid =
{
  {
    -1,
    (gchar *) "pid",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_open_for_write_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_open_for_write_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_open_for_write_IN_ARG_mode,
  &_gvfs_dbus_mount_method_info_open_for_write_IN_ARG_etag,
  &_gvfs_dbus_mount_method_info_open_for_write_IN_ARG_make_backup,
  &_gvfs_dbus_mount_method_info_open_for_write_IN_ARG_flags,
  &_gvfs_dbus_mount_method_info_open_for_write_IN_ARG_pid,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_for_write_OUT_ARG_fd_id =
{
  {
    -1,
    (gchar *) "fd_id",
    (gchar *) "h",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_for_write_OUT_ARG_can_seek =
{
  {
    -1,
    (gchar *) "can_seek",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_for_write_OUT_ARG_initial_offset =
{
  {
    -1,
    (gchar *) "initial_offset",
    (gchar *) "t",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_open_for_write_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_open_for_write_OUT_ARG_fd_id,
  &_gvfs_dbus_mount_method_info_open_for_write_OUT_ARG_can_seek,
  &_gvfs_dbus_mount_method_info_open_for_write_OUT_ARG_initial_offset,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_open_for_write =
{
  {
    -1,
    (gchar *) "OpenForWrite",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_open_for_write_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_open_for_write_OUT_ARG_pointers,
    NULL
  },
  "handle-open-for-write",
  TRUE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_query_info_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_query_info_IN_ARG_attributes =
{
  {
    -1,
    (gchar *) "attributes",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_query_info_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_query_info_IN_ARG_uri =
{
  {
    -1,
    (gchar *) "uri",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_query_info_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_query_info_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_query_info_IN_ARG_attributes,
  &_gvfs_dbus_mount_method_info_query_info_IN_ARG_flags,
  &_gvfs_dbus_mount_method_info_query_info_IN_ARG_uri,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_query_info_OUT_ARG_info =
{
  {
    -1,
    (gchar *) "info",
    (gchar *) "a(suv)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_query_info_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_query_info_OUT_ARG_info,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_query_info =
{
  {
    -1,
    (gchar *) "QueryInfo",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_query_info_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_query_info_OUT_ARG_pointers,
    NULL
  },
  "handle-query-info",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_query_filesystem_info_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_query_filesystem_info_IN_ARG_attributes =
{
  {
    -1,
    (gchar *) "attributes",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_query_filesystem_info_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_query_filesystem_info_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_query_filesystem_info_IN_ARG_attributes,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_query_filesystem_info_OUT_ARG_info =
{
  {
    -1,
    (gchar *) "info",
    (gchar *) "a(suv)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_query_filesystem_info_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_query_filesystem_info_OUT_ARG_info,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_query_filesystem_info =
{
  {
    -1,
    (gchar *) "QueryFilesystemInfo",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_query_filesystem_info_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_query_filesystem_info_OUT_ARG_pointers,
    NULL
  },
  "handle-query-filesystem-info",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_enumerate_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_enumerate_IN_ARG_obj_path =
{
  {
    -1,
    (gchar *) "obj_path",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_enumerate_IN_ARG_attributes =
{
  {
    -1,
    (gchar *) "attributes",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_enumerate_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_enumerate_IN_ARG_uri =
{
  {
    -1,
    (gchar *) "uri",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_enumerate_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_enumerate_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_enumerate_IN_ARG_obj_path,
  &_gvfs_dbus_mount_method_info_enumerate_IN_ARG_attributes,
  &_gvfs_dbus_mount_method_info_enumerate_IN_ARG_flags,
  &_gvfs_dbus_mount_method_info_enumerate_IN_ARG_uri,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_enumerate =
{
  {
    -1,
    (gchar *) "Enumerate",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_enumerate_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-enumerate",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_create_directory_monitor_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_create_directory_monitor_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_create_directory_monitor_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_create_directory_monitor_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_create_directory_monitor_IN_ARG_flags,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_create_directory_monitor_OUT_ARG_obj_path =
{
  {
    -1,
    (gchar *) "obj_path",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_create_directory_monitor_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_create_directory_monitor_OUT_ARG_obj_path,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_create_directory_monitor =
{
  {
    -1,
    (gchar *) "CreateDirectoryMonitor",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_create_directory_monitor_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_create_directory_monitor_OUT_ARG_pointers,
    NULL
  },
  "handle-create-directory-monitor",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_create_file_monitor_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_create_file_monitor_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_create_file_monitor_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_create_file_monitor_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_create_file_monitor_IN_ARG_flags,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_create_file_monitor_OUT_ARG_obj_path =
{
  {
    -1,
    (gchar *) "obj_path",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_create_file_monitor_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_create_file_monitor_OUT_ARG_obj_path,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_create_file_monitor =
{
  {
    -1,
    (gchar *) "CreateFileMonitor",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_create_file_monitor_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_create_file_monitor_OUT_ARG_pointers,
    NULL
  },
  "handle-create-file-monitor",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_mount_mountable_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_mount_mountable_IN_ARG_dbus_id =
{
  {
    -1,
    (gchar *) "dbus_id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_mount_mountable_IN_ARG_obj_path =
{
  {
    -1,
    (gchar *) "obj_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_mount_mountable_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_mount_mountable_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_mount_mountable_IN_ARG_dbus_id,
  &_gvfs_dbus_mount_method_info_mount_mountable_IN_ARG_obj_path,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_mount_mountable_OUT_ARG_is_uri =
{
  {
    -1,
    (gchar *) "is_uri",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_mount_mountable_OUT_ARG_path =
{
  {
    -1,
    (gchar *) "path",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_mount_mountable_OUT_ARG_must_mount_location =
{
  {
    -1,
    (gchar *) "must_mount_location",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_mount_mountable_OUT_ARG_mount_spec =
{
  {
    -1,
    (gchar *) "mount_spec",
    (gchar *) "(aya{sv})",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_mount_mountable_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_mount_mountable_OUT_ARG_is_uri,
  &_gvfs_dbus_mount_method_info_mount_mountable_OUT_ARG_path,
  &_gvfs_dbus_mount_method_info_mount_mountable_OUT_ARG_must_mount_location,
  &_gvfs_dbus_mount_method_info_mount_mountable_OUT_ARG_mount_spec,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_mount_mountable =
{
  {
    -1,
    (gchar *) "MountMountable",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_mount_mountable_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_mount_mountable_OUT_ARG_pointers,
    NULL
  },
  "handle-mount-mountable",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_unmount_mountable_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_unmount_mountable_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_unmount_mountable_IN_ARG_dbus_id =
{
  {
    -1,
    (gchar *) "dbus_id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_unmount_mountable_IN_ARG_obj_path =
{
  {
    -1,
    (gchar *) "obj_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_unmount_mountable_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_unmount_mountable_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_unmount_mountable_IN_ARG_flags,
  &_gvfs_dbus_mount_method_info_unmount_mountable_IN_ARG_dbus_id,
  &_gvfs_dbus_mount_method_info_unmount_mountable_IN_ARG_obj_path,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_unmount_mountable =
{
  {
    -1,
    (gchar *) "UnmountMountable",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_unmount_mountable_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-unmount-mountable",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_eject_mountable_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_eject_mountable_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_eject_mountable_IN_ARG_dbus_id =
{
  {
    -1,
    (gchar *) "dbus_id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_eject_mountable_IN_ARG_obj_path =
{
  {
    -1,
    (gchar *) "obj_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_eject_mountable_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_eject_mountable_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_eject_mountable_IN_ARG_flags,
  &_gvfs_dbus_mount_method_info_eject_mountable_IN_ARG_dbus_id,
  &_gvfs_dbus_mount_method_info_eject_mountable_IN_ARG_obj_path,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_eject_mountable =
{
  {
    -1,
    (gchar *) "EjectMountable",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_eject_mountable_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-eject-mountable",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_start_mountable_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_start_mountable_IN_ARG_dbus_id =
{
  {
    -1,
    (gchar *) "dbus_id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_start_mountable_IN_ARG_obj_path =
{
  {
    -1,
    (gchar *) "obj_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_start_mountable_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_start_mountable_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_start_mountable_IN_ARG_dbus_id,
  &_gvfs_dbus_mount_method_info_start_mountable_IN_ARG_obj_path,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_start_mountable =
{
  {
    -1,
    (gchar *) "StartMountable",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_start_mountable_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-start-mountable",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_stop_mountable_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_stop_mountable_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_stop_mountable_IN_ARG_dbus_id =
{
  {
    -1,
    (gchar *) "dbus_id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_stop_mountable_IN_ARG_obj_path =
{
  {
    -1,
    (gchar *) "obj_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_stop_mountable_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_stop_mountable_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_stop_mountable_IN_ARG_flags,
  &_gvfs_dbus_mount_method_info_stop_mountable_IN_ARG_dbus_id,
  &_gvfs_dbus_mount_method_info_stop_mountable_IN_ARG_obj_path,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_stop_mountable =
{
  {
    -1,
    (gchar *) "StopMountable",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_stop_mountable_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-stop-mountable",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_poll_mountable_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_poll_mountable_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_poll_mountable_IN_ARG_path_data,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_poll_mountable =
{
  {
    -1,
    (gchar *) "PollMountable",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_poll_mountable_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-poll-mountable",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_set_display_name_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_set_display_name_IN_ARG_display_name =
{
  {
    -1,
    (gchar *) "display_name",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_set_display_name_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_set_display_name_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_set_display_name_IN_ARG_display_name,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_set_display_name_OUT_ARG_new_path =
{
  {
    -1,
    (gchar *) "new_path",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_set_display_name_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_set_display_name_OUT_ARG_new_path,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_set_display_name =
{
  {
    -1,
    (gchar *) "SetDisplayName",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_set_display_name_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_set_display_name_OUT_ARG_pointers,
    NULL
  },
  "handle-set-display-name",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_delete_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_delete_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_delete_IN_ARG_path_data,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_delete =
{
  {
    -1,
    (gchar *) "Delete",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_delete_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-delete",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_trash_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_trash_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_trash_IN_ARG_path_data,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_trash =
{
  {
    -1,
    (gchar *) "Trash",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_trash_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-trash",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_make_directory_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_make_directory_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_make_directory_IN_ARG_path_data,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_make_directory =
{
  {
    -1,
    (gchar *) "MakeDirectory",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_make_directory_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-make-directory",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_make_symbolic_link_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_make_symbolic_link_IN_ARG_symlink_value =
{
  {
    -1,
    (gchar *) "symlink_value",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_make_symbolic_link_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_make_symbolic_link_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_make_symbolic_link_IN_ARG_symlink_value,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_make_symbolic_link =
{
  {
    -1,
    (gchar *) "MakeSymbolicLink",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_make_symbolic_link_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-make-symbolic-link",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_copy_IN_ARG_path1_data =
{
  {
    -1,
    (gchar *) "path1_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_copy_IN_ARG_path2_data =
{
  {
    -1,
    (gchar *) "path2_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_copy_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_copy_IN_ARG_progress_obj_path =
{
  {
    -1,
    (gchar *) "progress_obj_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_copy_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_copy_IN_ARG_path1_data,
  &_gvfs_dbus_mount_method_info_copy_IN_ARG_path2_data,
  &_gvfs_dbus_mount_method_info_copy_IN_ARG_flags,
  &_gvfs_dbus_mount_method_info_copy_IN_ARG_progress_obj_path,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_copy =
{
  {
    -1,
    (gchar *) "Copy",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_copy_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-copy",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_move_IN_ARG_path1_data =
{
  {
    -1,
    (gchar *) "path1_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_move_IN_ARG_path2_data =
{
  {
    -1,
    (gchar *) "path2_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_move_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_move_IN_ARG_progress_obj_path =
{
  {
    -1,
    (gchar *) "progress_obj_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_move_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_move_IN_ARG_path1_data,
  &_gvfs_dbus_mount_method_info_move_IN_ARG_path2_data,
  &_gvfs_dbus_mount_method_info_move_IN_ARG_flags,
  &_gvfs_dbus_mount_method_info_move_IN_ARG_progress_obj_path,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_move =
{
  {
    -1,
    (gchar *) "Move",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_move_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-move",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_push_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_push_IN_ARG_local_path =
{
  {
    -1,
    (gchar *) "local_path",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_push_IN_ARG_send_progress =
{
  {
    -1,
    (gchar *) "send_progress",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_push_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_push_IN_ARG_progress_obj_path =
{
  {
    -1,
    (gchar *) "progress_obj_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_push_IN_ARG_remove_source =
{
  {
    -1,
    (gchar *) "remove_source",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_push_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_push_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_push_IN_ARG_local_path,
  &_gvfs_dbus_mount_method_info_push_IN_ARG_send_progress,
  &_gvfs_dbus_mount_method_info_push_IN_ARG_flags,
  &_gvfs_dbus_mount_method_info_push_IN_ARG_progress_obj_path,
  &_gvfs_dbus_mount_method_info_push_IN_ARG_remove_source,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_push =
{
  {
    -1,
    (gchar *) "Push",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_push_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-push",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_pull_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_pull_IN_ARG_local_path =
{
  {
    -1,
    (gchar *) "local_path",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_pull_IN_ARG_send_progress =
{
  {
    -1,
    (gchar *) "send_progress",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_pull_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_pull_IN_ARG_progress_obj_path =
{
  {
    -1,
    (gchar *) "progress_obj_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_pull_IN_ARG_remove_source =
{
  {
    -1,
    (gchar *) "remove_source",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_pull_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_pull_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_pull_IN_ARG_local_path,
  &_gvfs_dbus_mount_method_info_pull_IN_ARG_send_progress,
  &_gvfs_dbus_mount_method_info_pull_IN_ARG_flags,
  &_gvfs_dbus_mount_method_info_pull_IN_ARG_progress_obj_path,
  &_gvfs_dbus_mount_method_info_pull_IN_ARG_remove_source,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_pull =
{
  {
    -1,
    (gchar *) "Pull",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_pull_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-pull",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_set_attribute_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_set_attribute_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_set_attribute_IN_ARG_attribute =
{
  {
    -1,
    (gchar *) "attribute",
    (gchar *) "(suv)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_set_attribute_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_set_attribute_IN_ARG_path_data,
  &_gvfs_dbus_mount_method_info_set_attribute_IN_ARG_flags,
  &_gvfs_dbus_mount_method_info_set_attribute_IN_ARG_attribute,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_set_attribute =
{
  {
    -1,
    (gchar *) "SetAttribute",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_set_attribute_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-set-attribute",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_query_settable_attributes_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_query_settable_attributes_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_query_settable_attributes_IN_ARG_path_data,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_query_settable_attributes_OUT_ARG_list =
{
  {
    -1,
    (gchar *) "list",
    (gchar *) "a(suu)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_query_settable_attributes_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_query_settable_attributes_OUT_ARG_list,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_query_settable_attributes =
{
  {
    -1,
    (gchar *) "QuerySettableAttributes",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_query_settable_attributes_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_query_settable_attributes_OUT_ARG_pointers,
    NULL
  },
  "handle-query-settable-attributes",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_query_writable_namespaces_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_query_writable_namespaces_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_query_writable_namespaces_IN_ARG_path_data,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_query_writable_namespaces_OUT_ARG_list =
{
  {
    -1,
    (gchar *) "list",
    (gchar *) "a(suu)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_query_writable_namespaces_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_query_writable_namespaces_OUT_ARG_list,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_query_writable_namespaces =
{
  {
    -1,
    (gchar *) "QueryWritableNamespaces",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_query_writable_namespaces_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_query_writable_namespaces_OUT_ARG_pointers,
    NULL
  },
  "handle-query-writable-namespaces",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_icon_for_read_IN_ARG_path_data =
{
  {
    -1,
    (gchar *) "path_data",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_open_icon_for_read_IN_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_open_icon_for_read_IN_ARG_path_data,
  NULL
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_icon_for_read_OUT_ARG_fd_id =
{
  {
    -1,
    (gchar *) "fd_id",
    (gchar *) "h",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_mount_method_info_open_icon_for_read_OUT_ARG_can_seek =
{
  {
    -1,
    (gchar *) "can_seek",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_mount_method_info_open_icon_for_read_OUT_ARG_pointers[] =
{
  &_gvfs_dbus_mount_method_info_open_icon_for_read_OUT_ARG_fd_id,
  &_gvfs_dbus_mount_method_info_open_icon_for_read_OUT_ARG_can_seek,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_mount_method_info_open_icon_for_read =
{
  {
    -1,
    (gchar *) "OpenIconForRead",
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_open_icon_for_read_IN_ARG_pointers,
    (GDBusArgInfo **) &_gvfs_dbus_mount_method_info_open_icon_for_read_OUT_ARG_pointers,
    NULL
  },
  "handle-open-icon-for-read",
  TRUE
};

static const _ExtendedGDBusMethodInfo * const _gvfs_dbus_mount_method_info_pointers[] =
{
  &_gvfs_dbus_mount_method_info_unmount,
  &_gvfs_dbus_mount_method_info_open_for_read,
  &_gvfs_dbus_mount_method_info_open_for_write,
  &_gvfs_dbus_mount_method_info_query_info,
  &_gvfs_dbus_mount_method_info_query_filesystem_info,
  &_gvfs_dbus_mount_method_info_enumerate,
  &_gvfs_dbus_mount_method_info_create_directory_monitor,
  &_gvfs_dbus_mount_method_info_create_file_monitor,
  &_gvfs_dbus_mount_method_info_mount_mountable,
  &_gvfs_dbus_mount_method_info_unmount_mountable,
  &_gvfs_dbus_mount_method_info_eject_mountable,
  &_gvfs_dbus_mount_method_info_start_mountable,
  &_gvfs_dbus_mount_method_info_stop_mountable,
  &_gvfs_dbus_mount_method_info_poll_mountable,
  &_gvfs_dbus_mount_method_info_set_display_name,
  &_gvfs_dbus_mount_method_info_delete,
  &_gvfs_dbus_mount_method_info_trash,
  &_gvfs_dbus_mount_method_info_make_directory,
  &_gvfs_dbus_mount_method_info_make_symbolic_link,
  &_gvfs_dbus_mount_method_info_copy,
  &_gvfs_dbus_mount_method_info_move,
  &_gvfs_dbus_mount_method_info_push,
  &_gvfs_dbus_mount_method_info_pull,
  &_gvfs_dbus_mount_method_info_set_attribute,
  &_gvfs_dbus_mount_method_info_query_settable_attributes,
  &_gvfs_dbus_mount_method_info_query_writable_namespaces,
  &_gvfs_dbus_mount_method_info_open_icon_for_read,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _gvfs_dbus_mount_interface_info =
{
  {
    -1,
    (gchar *) "org.gtk.vfs.Mount",
    (GDBusMethodInfo **) &_gvfs_dbus_mount_method_info_pointers,
    NULL,
    NULL,
    NULL
  },
  "mount",
};


/**
 * gvfs_dbus_mount_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gtk-vfs-Mount.top_of_page">org.gtk.vfs.Mount</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
gvfs_dbus_mount_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_gvfs_dbus_mount_interface_info.parent_struct;
}

/**
 * gvfs_dbus_mount_override_properties:
 * @klass: The class structure for a #GObject<!-- -->-derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #GVfsDBusMount interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
gvfs_dbus_mount_override_properties (GObjectClass *klass, guint property_id_begin)
{
  return property_id_begin - 1;
}



/**
 * GVfsDBusMount:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Mount.top_of_page">org.gtk.vfs.Mount</link>.
 */

/**
 * GVfsDBusMountIface:
 * @parent_iface: The parent interface.
 * @handle_copy: Handler for the #GVfsDBusMount::handle-copy signal.
 * @handle_create_directory_monitor: Handler for the #GVfsDBusMount::handle-create-directory-monitor signal.
 * @handle_create_file_monitor: Handler for the #GVfsDBusMount::handle-create-file-monitor signal.
 * @handle_delete: Handler for the #GVfsDBusMount::handle-delete signal.
 * @handle_eject_mountable: Handler for the #GVfsDBusMount::handle-eject-mountable signal.
 * @handle_enumerate: Handler for the #GVfsDBusMount::handle-enumerate signal.
 * @handle_make_directory: Handler for the #GVfsDBusMount::handle-make-directory signal.
 * @handle_make_symbolic_link: Handler for the #GVfsDBusMount::handle-make-symbolic-link signal.
 * @handle_mount_mountable: Handler for the #GVfsDBusMount::handle-mount-mountable signal.
 * @handle_move: Handler for the #GVfsDBusMount::handle-move signal.
 * @handle_open_for_read: Handler for the #GVfsDBusMount::handle-open-for-read signal.
 * @handle_open_for_write: Handler for the #GVfsDBusMount::handle-open-for-write signal.
 * @handle_open_icon_for_read: Handler for the #GVfsDBusMount::handle-open-icon-for-read signal.
 * @handle_poll_mountable: Handler for the #GVfsDBusMount::handle-poll-mountable signal.
 * @handle_pull: Handler for the #GVfsDBusMount::handle-pull signal.
 * @handle_push: Handler for the #GVfsDBusMount::handle-push signal.
 * @handle_query_filesystem_info: Handler for the #GVfsDBusMount::handle-query-filesystem-info signal.
 * @handle_query_info: Handler for the #GVfsDBusMount::handle-query-info signal.
 * @handle_query_settable_attributes: Handler for the #GVfsDBusMount::handle-query-settable-attributes signal.
 * @handle_query_writable_namespaces: Handler for the #GVfsDBusMount::handle-query-writable-namespaces signal.
 * @handle_set_attribute: Handler for the #GVfsDBusMount::handle-set-attribute signal.
 * @handle_set_display_name: Handler for the #GVfsDBusMount::handle-set-display-name signal.
 * @handle_start_mountable: Handler for the #GVfsDBusMount::handle-start-mountable signal.
 * @handle_stop_mountable: Handler for the #GVfsDBusMount::handle-stop-mountable signal.
 * @handle_trash: Handler for the #GVfsDBusMount::handle-trash signal.
 * @handle_unmount: Handler for the #GVfsDBusMount::handle-unmount signal.
 * @handle_unmount_mountable: Handler for the #GVfsDBusMount::handle-unmount-mountable signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Mount.top_of_page">org.gtk.vfs.Mount</link>.
 */

typedef GVfsDBusMountIface GVfsDBusMountInterface;
G_DEFINE_INTERFACE (GVfsDBusMount, gvfs_dbus_mount, G_TYPE_OBJECT);

static void
gvfs_dbus_mount_default_init (GVfsDBusMountIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * GVfsDBusMount::handle-unmount:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_dbus_id: Argument passed by remote caller.
   * @arg_obj_path: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.Unmount">Unmount()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_unmount() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-unmount",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_unmount),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT);

  /**
   * GVfsDBusMount::handle-open-for-read:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @fd_list: (allow-none): A #GUnixFDList or %NULL.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_pid: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.OpenForRead">OpenForRead()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_open_for_read() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-open-for-read",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_open_for_read),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UNIX_FD_LIST, G_TYPE_STRING, G_TYPE_UINT);

  /**
   * GVfsDBusMount::handle-open-for-write:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @fd_list: (allow-none): A #GUnixFDList or %NULL.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_mode: Argument passed by remote caller.
   * @arg_etag: Argument passed by remote caller.
   * @arg_make_backup: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   * @arg_pid: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.OpenForWrite">OpenForWrite()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_open_for_write() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-open-for-write",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_open_for_write),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    8,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UNIX_FD_LIST, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_BOOLEAN, G_TYPE_UINT, G_TYPE_UINT);

  /**
   * GVfsDBusMount::handle-query-info:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_attributes: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   * @arg_uri: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.QueryInfo">QueryInfo()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_query_info() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-query-info",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_query_info),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-query-filesystem-info:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_attributes: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.QueryFilesystemInfo">QueryFilesystemInfo()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_query_filesystem_info() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-query-filesystem-info",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_query_filesystem_info),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-enumerate:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_obj_path: Argument passed by remote caller.
   * @arg_attributes: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   * @arg_uri: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.Enumerate">Enumerate()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_enumerate() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-enumerate",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_enumerate),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    6,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-create-directory-monitor:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.CreateDirectoryMonitor">CreateDirectoryMonitor()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_create_directory_monitor() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-create-directory-monitor",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_create_directory_monitor),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_UINT);

  /**
   * GVfsDBusMount::handle-create-file-monitor:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.CreateFileMonitor">CreateFileMonitor()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_create_file_monitor() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-create-file-monitor",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_create_file_monitor),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_UINT);

  /**
   * GVfsDBusMount::handle-mount-mountable:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_dbus_id: Argument passed by remote caller.
   * @arg_obj_path: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.MountMountable">MountMountable()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_mount_mountable() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-mount-mountable",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_mount_mountable),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-unmount-mountable:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   * @arg_dbus_id: Argument passed by remote caller.
   * @arg_obj_path: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.UnmountMountable">UnmountMountable()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_unmount_mountable() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-unmount-mountable",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_unmount_mountable),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-eject-mountable:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   * @arg_dbus_id: Argument passed by remote caller.
   * @arg_obj_path: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.EjectMountable">EjectMountable()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_eject_mountable() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-eject-mountable",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_eject_mountable),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-start-mountable:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_dbus_id: Argument passed by remote caller.
   * @arg_obj_path: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.StartMountable">StartMountable()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_start_mountable() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-start-mountable",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_start_mountable),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-stop-mountable:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   * @arg_dbus_id: Argument passed by remote caller.
   * @arg_obj_path: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.StopMountable">StopMountable()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_stop_mountable() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-stop-mountable",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_stop_mountable),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-poll-mountable:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.PollMountable">PollMountable()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_poll_mountable() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-poll-mountable",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_poll_mountable),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-set-display-name:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_display_name: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.SetDisplayName">SetDisplayName()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_set_display_name() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-display-name",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_set_display_name),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-delete:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.Delete">Delete()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_delete() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-delete",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_delete),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-trash:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.Trash">Trash()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_trash() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-trash",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_trash),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-make-directory:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.MakeDirectory">MakeDirectory()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_make_directory() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-make-directory",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_make_directory),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-make-symbolic-link:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_symlink_value: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.MakeSymbolicLink">MakeSymbolicLink()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_make_symbolic_link() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-make-symbolic-link",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_make_symbolic_link),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-copy:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path1_data: Argument passed by remote caller.
   * @arg_path2_data: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   * @arg_progress_obj_path: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.Copy">Copy()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_copy() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-copy",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_copy),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-move:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path1_data: Argument passed by remote caller.
   * @arg_path2_data: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   * @arg_progress_obj_path: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.Move">Move()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_move() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-move",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_move),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-push:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_local_path: Argument passed by remote caller.
   * @arg_send_progress: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   * @arg_progress_obj_path: Argument passed by remote caller.
   * @arg_remove_source: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.Push">Push()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_push() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-push",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_push),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    7,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_BOOLEAN);

  /**
   * GVfsDBusMount::handle-pull:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_local_path: Argument passed by remote caller.
   * @arg_send_progress: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   * @arg_progress_obj_path: Argument passed by remote caller.
   * @arg_remove_source: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.Pull">Pull()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_pull() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-pull",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_pull),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    7,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_BOOLEAN);

  /**
   * GVfsDBusMount::handle-set-attribute:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   * @arg_attribute: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.SetAttribute">SetAttribute()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_set_attribute() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-attribute",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_set_attribute),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_VARIANT);

  /**
   * GVfsDBusMount::handle-query-settable-attributes:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.QuerySettableAttributes">QuerySettableAttributes()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_query_settable_attributes() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-query-settable-attributes",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_query_settable_attributes),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-query-writable-namespaces:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_path_data: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.QueryWritableNamespaces">QueryWritableNamespaces()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_query_writable_namespaces() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-query-writable-namespaces",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_query_writable_namespaces),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * GVfsDBusMount::handle-open-icon-for-read:
   * @object: A #GVfsDBusMount.
   * @invocation: A #GDBusMethodInvocation.
   * @fd_list: (allow-none): A #GUnixFDList or %NULL.
   * @arg_path_data: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Mount.OpenIconForRead">OpenIconForRead()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_mount_complete_open_icon_for_read() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-open-icon-for-read",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMountIface, handle_open_icon_for_read),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UNIX_FD_LIST, G_TYPE_STRING);

}

/**
 * gvfs_dbus_mount_call_unmount:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Unmount">Unmount()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_unmount_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_unmount_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_unmount (
    GVfsDBusMount *proxy,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    guint arg_flags,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Unmount",
    g_variant_new ("(sou)",
                   arg_dbus_id,
                   arg_obj_path,
                   arg_flags),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_unmount_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_unmount().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_unmount().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_unmount_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_unmount_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Unmount">Unmount()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_unmount() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_unmount_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    guint arg_flags,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Unmount",
    g_variant_new ("(sou)",
                   arg_dbus_id,
                   arg_obj_path,
                   arg_flags),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_open_for_read:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_pid: Argument to pass with the method invocation.
 * @fd_list: (allow-none): A #GUnixFDList or %NULL.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.OpenForRead">OpenForRead()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_open_for_read_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_open_for_read_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_open_for_read (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_pid,
    GUnixFDList *fd_list,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call_with_unix_fd_list (G_DBUS_PROXY (proxy),
    "OpenForRead",
    g_variant_new ("(^ayu)",
                   arg_path_data,
                   arg_pid),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_open_for_read_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @out_fd_id: (out): Return location for return parameter or %NULL to ignore.
 * @out_can_seek: (out): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out): Return location for a #GUnixFDList or %NULL.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_open_for_read().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_open_for_read().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_open_for_read_finish (
    GVfsDBusMount *proxy,
    GVariant **out_fd_id,
    gboolean *out_can_seek,
    GUnixFDList **out_fd_list,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_finish (G_DBUS_PROXY (proxy), out_fd_list, res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@hb)",
                 out_fd_id,
                 out_can_seek);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_open_for_read_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_pid: Argument to pass with the method invocation.
 * @fd_list: (allow-none): A #GUnixFDList or %NULL.
 * @out_fd_id: (out): Return location for return parameter or %NULL to ignore.
 * @out_can_seek: (out): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out): Return location for a #GUnixFDList or %NULL.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.OpenForRead">OpenForRead()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_open_for_read() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_open_for_read_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_pid,
    GUnixFDList  *fd_list,
    GVariant **out_fd_id,
    gboolean *out_can_seek,
    GUnixFDList **out_fd_list,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_sync (G_DBUS_PROXY (proxy),
    "OpenForRead",
    g_variant_new ("(^ayu)",
                   arg_path_data,
                   arg_pid),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    out_fd_list,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@hb)",
                 out_fd_id,
                 out_can_seek);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_open_for_write:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_mode: Argument to pass with the method invocation.
 * @arg_etag: Argument to pass with the method invocation.
 * @arg_make_backup: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_pid: Argument to pass with the method invocation.
 * @fd_list: (allow-none): A #GUnixFDList or %NULL.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.OpenForWrite">OpenForWrite()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_open_for_write_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_open_for_write_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_open_for_write (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint16 arg_mode,
    const gchar *arg_etag,
    gboolean arg_make_backup,
    guint arg_flags,
    guint arg_pid,
    GUnixFDList *fd_list,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call_with_unix_fd_list (G_DBUS_PROXY (proxy),
    "OpenForWrite",
    g_variant_new ("(^ayqsbuu)",
                   arg_path_data,
                   arg_mode,
                   arg_etag,
                   arg_make_backup,
                   arg_flags,
                   arg_pid),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_open_for_write_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @out_fd_id: (out): Return location for return parameter or %NULL to ignore.
 * @out_can_seek: (out): Return location for return parameter or %NULL to ignore.
 * @out_initial_offset: (out): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out): Return location for a #GUnixFDList or %NULL.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_open_for_write().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_open_for_write().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_open_for_write_finish (
    GVfsDBusMount *proxy,
    GVariant **out_fd_id,
    gboolean *out_can_seek,
    guint64 *out_initial_offset,
    GUnixFDList **out_fd_list,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_finish (G_DBUS_PROXY (proxy), out_fd_list, res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@hbt)",
                 out_fd_id,
                 out_can_seek,
                 out_initial_offset);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_open_for_write_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_mode: Argument to pass with the method invocation.
 * @arg_etag: Argument to pass with the method invocation.
 * @arg_make_backup: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_pid: Argument to pass with the method invocation.
 * @fd_list: (allow-none): A #GUnixFDList or %NULL.
 * @out_fd_id: (out): Return location for return parameter or %NULL to ignore.
 * @out_can_seek: (out): Return location for return parameter or %NULL to ignore.
 * @out_initial_offset: (out): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out): Return location for a #GUnixFDList or %NULL.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.OpenForWrite">OpenForWrite()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_open_for_write() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_open_for_write_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint16 arg_mode,
    const gchar *arg_etag,
    gboolean arg_make_backup,
    guint arg_flags,
    guint arg_pid,
    GUnixFDList  *fd_list,
    GVariant **out_fd_id,
    gboolean *out_can_seek,
    guint64 *out_initial_offset,
    GUnixFDList **out_fd_list,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_sync (G_DBUS_PROXY (proxy),
    "OpenForWrite",
    g_variant_new ("(^ayqsbuu)",
                   arg_path_data,
                   arg_mode,
                   arg_etag,
                   arg_make_backup,
                   arg_flags,
                   arg_pid),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    out_fd_list,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@hbt)",
                 out_fd_id,
                 out_can_seek,
                 out_initial_offset);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_query_info:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_attributes: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_uri: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.QueryInfo">QueryInfo()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_query_info_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_query_info_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_query_info (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_attributes,
    guint arg_flags,
    const gchar *arg_uri,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "QueryInfo",
    g_variant_new ("(^aysus)",
                   arg_path_data,
                   arg_attributes,
                   arg_flags,
                   arg_uri),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_query_info_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @out_info: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_query_info().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_query_info().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_query_info_finish (
    GVfsDBusMount *proxy,
    GVariant **out_info,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(suv))",
                 out_info);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_query_info_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_attributes: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_uri: Argument to pass with the method invocation.
 * @out_info: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.QueryInfo">QueryInfo()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_query_info() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_query_info_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_attributes,
    guint arg_flags,
    const gchar *arg_uri,
    GVariant **out_info,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "QueryInfo",
    g_variant_new ("(^aysus)",
                   arg_path_data,
                   arg_attributes,
                   arg_flags,
                   arg_uri),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(suv))",
                 out_info);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_query_filesystem_info:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_attributes: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.QueryFilesystemInfo">QueryFilesystemInfo()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_query_filesystem_info_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_query_filesystem_info_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_query_filesystem_info (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_attributes,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "QueryFilesystemInfo",
    g_variant_new ("(^ays)",
                   arg_path_data,
                   arg_attributes),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_query_filesystem_info_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @out_info: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_query_filesystem_info().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_query_filesystem_info().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_query_filesystem_info_finish (
    GVfsDBusMount *proxy,
    GVariant **out_info,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(suv))",
                 out_info);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_query_filesystem_info_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_attributes: Argument to pass with the method invocation.
 * @out_info: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.QueryFilesystemInfo">QueryFilesystemInfo()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_query_filesystem_info() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_query_filesystem_info_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_attributes,
    GVariant **out_info,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "QueryFilesystemInfo",
    g_variant_new ("(^ays)",
                   arg_path_data,
                   arg_attributes),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(suv))",
                 out_info);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_enumerate:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @arg_attributes: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_uri: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Enumerate">Enumerate()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_enumerate_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_enumerate_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_enumerate (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_obj_path,
    const gchar *arg_attributes,
    guint arg_flags,
    const gchar *arg_uri,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Enumerate",
    g_variant_new ("(^ayssus)",
                   arg_path_data,
                   arg_obj_path,
                   arg_attributes,
                   arg_flags,
                   arg_uri),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_enumerate_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_enumerate().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_enumerate().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_enumerate_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_enumerate_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @arg_attributes: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_uri: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Enumerate">Enumerate()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_enumerate() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_enumerate_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_obj_path,
    const gchar *arg_attributes,
    guint arg_flags,
    const gchar *arg_uri,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Enumerate",
    g_variant_new ("(^ayssus)",
                   arg_path_data,
                   arg_obj_path,
                   arg_attributes,
                   arg_flags,
                   arg_uri),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_create_directory_monitor:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.CreateDirectoryMonitor">CreateDirectoryMonitor()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_create_directory_monitor_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_create_directory_monitor_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_create_directory_monitor (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_flags,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "CreateDirectoryMonitor",
    g_variant_new ("(^ayu)",
                   arg_path_data,
                   arg_flags),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_create_directory_monitor_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @out_obj_path: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_create_directory_monitor().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_create_directory_monitor().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_create_directory_monitor_finish (
    GVfsDBusMount *proxy,
    gchar **out_obj_path,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(s)",
                 out_obj_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_create_directory_monitor_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @out_obj_path: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.CreateDirectoryMonitor">CreateDirectoryMonitor()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_create_directory_monitor() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_create_directory_monitor_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_flags,
    gchar **out_obj_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "CreateDirectoryMonitor",
    g_variant_new ("(^ayu)",
                   arg_path_data,
                   arg_flags),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(s)",
                 out_obj_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_create_file_monitor:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.CreateFileMonitor">CreateFileMonitor()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_create_file_monitor_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_create_file_monitor_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_create_file_monitor (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_flags,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "CreateFileMonitor",
    g_variant_new ("(^ayu)",
                   arg_path_data,
                   arg_flags),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_create_file_monitor_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @out_obj_path: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_create_file_monitor().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_create_file_monitor().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_create_file_monitor_finish (
    GVfsDBusMount *proxy,
    gchar **out_obj_path,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(s)",
                 out_obj_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_create_file_monitor_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @out_obj_path: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.CreateFileMonitor">CreateFileMonitor()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_create_file_monitor() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_create_file_monitor_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_flags,
    gchar **out_obj_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "CreateFileMonitor",
    g_variant_new ("(^ayu)",
                   arg_path_data,
                   arg_flags),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(s)",
                 out_obj_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_mount_mountable:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.MountMountable">MountMountable()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_mount_mountable_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_mount_mountable_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_mount_mountable (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "MountMountable",
    g_variant_new ("(^ayso)",
                   arg_path_data,
                   arg_dbus_id,
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_mount_mountable_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @out_is_uri: (out): Return location for return parameter or %NULL to ignore.
 * @out_path: (out): Return location for return parameter or %NULL to ignore.
 * @out_must_mount_location: (out): Return location for return parameter or %NULL to ignore.
 * @out_mount_spec: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_mount_mountable().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_mount_mountable().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_mount_mountable_finish (
    GVfsDBusMount *proxy,
    gboolean *out_is_uri,
    gchar **out_path,
    gboolean *out_must_mount_location,
    GVariant **out_mount_spec,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(b^ayb@(aya{sv}))",
                 out_is_uri,
                 out_path,
                 out_must_mount_location,
                 out_mount_spec);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_mount_mountable_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @out_is_uri: (out): Return location for return parameter or %NULL to ignore.
 * @out_path: (out): Return location for return parameter or %NULL to ignore.
 * @out_must_mount_location: (out): Return location for return parameter or %NULL to ignore.
 * @out_mount_spec: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.MountMountable">MountMountable()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_mount_mountable() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_mount_mountable_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    gboolean *out_is_uri,
    gchar **out_path,
    gboolean *out_must_mount_location,
    GVariant **out_mount_spec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "MountMountable",
    g_variant_new ("(^ayso)",
                   arg_path_data,
                   arg_dbus_id,
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(b^ayb@(aya{sv}))",
                 out_is_uri,
                 out_path,
                 out_must_mount_location,
                 out_mount_spec);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_unmount_mountable:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.UnmountMountable">UnmountMountable()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_unmount_mountable_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_unmount_mountable_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_unmount_mountable (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_flags,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "UnmountMountable",
    g_variant_new ("(^ayuso)",
                   arg_path_data,
                   arg_flags,
                   arg_dbus_id,
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_unmount_mountable_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_unmount_mountable().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_unmount_mountable().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_unmount_mountable_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_unmount_mountable_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.UnmountMountable">UnmountMountable()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_unmount_mountable() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_unmount_mountable_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_flags,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "UnmountMountable",
    g_variant_new ("(^ayuso)",
                   arg_path_data,
                   arg_flags,
                   arg_dbus_id,
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_eject_mountable:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.EjectMountable">EjectMountable()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_eject_mountable_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_eject_mountable_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_eject_mountable (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_flags,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "EjectMountable",
    g_variant_new ("(^ayuso)",
                   arg_path_data,
                   arg_flags,
                   arg_dbus_id,
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_eject_mountable_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_eject_mountable().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_eject_mountable().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_eject_mountable_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_eject_mountable_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.EjectMountable">EjectMountable()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_eject_mountable() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_eject_mountable_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_flags,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "EjectMountable",
    g_variant_new ("(^ayuso)",
                   arg_path_data,
                   arg_flags,
                   arg_dbus_id,
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_start_mountable:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.StartMountable">StartMountable()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_start_mountable_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_start_mountable_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_start_mountable (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "StartMountable",
    g_variant_new ("(^ayso)",
                   arg_path_data,
                   arg_dbus_id,
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_start_mountable_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_start_mountable().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_start_mountable().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_start_mountable_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_start_mountable_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.StartMountable">StartMountable()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_start_mountable() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_start_mountable_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "StartMountable",
    g_variant_new ("(^ayso)",
                   arg_path_data,
                   arg_dbus_id,
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_stop_mountable:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.StopMountable">StopMountable()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_stop_mountable_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_stop_mountable_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_stop_mountable (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_flags,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "StopMountable",
    g_variant_new ("(^ayuso)",
                   arg_path_data,
                   arg_flags,
                   arg_dbus_id,
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_stop_mountable_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_stop_mountable().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_stop_mountable().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_stop_mountable_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_stop_mountable_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_dbus_id: Argument to pass with the method invocation.
 * @arg_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.StopMountable">StopMountable()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_stop_mountable() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_stop_mountable_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_flags,
    const gchar *arg_dbus_id,
    const gchar *arg_obj_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "StopMountable",
    g_variant_new ("(^ayuso)",
                   arg_path_data,
                   arg_flags,
                   arg_dbus_id,
                   arg_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_poll_mountable:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.PollMountable">PollMountable()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_poll_mountable_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_poll_mountable_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_poll_mountable (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "PollMountable",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_poll_mountable_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_poll_mountable().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_poll_mountable().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_poll_mountable_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_poll_mountable_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.PollMountable">PollMountable()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_poll_mountable() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_poll_mountable_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "PollMountable",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_set_display_name:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_display_name: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.SetDisplayName">SetDisplayName()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_set_display_name_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_set_display_name_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_set_display_name (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_display_name,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetDisplayName",
    g_variant_new ("(^ays)",
                   arg_path_data,
                   arg_display_name),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_set_display_name_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @out_new_path: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_set_display_name().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_set_display_name().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_set_display_name_finish (
    GVfsDBusMount *proxy,
    gchar **out_new_path,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(^ay)",
                 out_new_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_set_display_name_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_display_name: Argument to pass with the method invocation.
 * @out_new_path: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.SetDisplayName">SetDisplayName()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_set_display_name() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_set_display_name_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_display_name,
    gchar **out_new_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetDisplayName",
    g_variant_new ("(^ays)",
                   arg_path_data,
                   arg_display_name),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(^ay)",
                 out_new_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_delete:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Delete">Delete()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_delete_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_delete_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_delete (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Delete",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_delete_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_delete().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_delete().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_delete_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_delete_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Delete">Delete()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_delete() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_delete_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Delete",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_trash:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Trash">Trash()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_trash_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_trash_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_trash (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Trash",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_trash_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_trash().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_trash().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_trash_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_trash_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Trash">Trash()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_trash() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_trash_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Trash",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_make_directory:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.MakeDirectory">MakeDirectory()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_make_directory_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_make_directory_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_make_directory (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "MakeDirectory",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_make_directory_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_make_directory().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_make_directory().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_make_directory_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_make_directory_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.MakeDirectory">MakeDirectory()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_make_directory() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_make_directory_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "MakeDirectory",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_make_symbolic_link:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_symlink_value: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.MakeSymbolicLink">MakeSymbolicLink()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_make_symbolic_link_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_make_symbolic_link_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_make_symbolic_link (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_symlink_value,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "MakeSymbolicLink",
    g_variant_new ("(^ay^ay)",
                   arg_path_data,
                   arg_symlink_value),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_make_symbolic_link_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_make_symbolic_link().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_make_symbolic_link().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_make_symbolic_link_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_make_symbolic_link_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_symlink_value: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.MakeSymbolicLink">MakeSymbolicLink()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_make_symbolic_link() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_make_symbolic_link_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_symlink_value,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "MakeSymbolicLink",
    g_variant_new ("(^ay^ay)",
                   arg_path_data,
                   arg_symlink_value),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_copy:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path1_data: Argument to pass with the method invocation.
 * @arg_path2_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_progress_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Copy">Copy()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_copy_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_copy_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_copy (
    GVfsDBusMount *proxy,
    const gchar *arg_path1_data,
    const gchar *arg_path2_data,
    guint arg_flags,
    const gchar *arg_progress_obj_path,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Copy",
    g_variant_new ("(^ay^ayuo)",
                   arg_path1_data,
                   arg_path2_data,
                   arg_flags,
                   arg_progress_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_copy_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_copy().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_copy().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_copy_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_copy_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path1_data: Argument to pass with the method invocation.
 * @arg_path2_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_progress_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Copy">Copy()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_copy() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_copy_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path1_data,
    const gchar *arg_path2_data,
    guint arg_flags,
    const gchar *arg_progress_obj_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Copy",
    g_variant_new ("(^ay^ayuo)",
                   arg_path1_data,
                   arg_path2_data,
                   arg_flags,
                   arg_progress_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_move:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path1_data: Argument to pass with the method invocation.
 * @arg_path2_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_progress_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Move">Move()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_move_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_move_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_move (
    GVfsDBusMount *proxy,
    const gchar *arg_path1_data,
    const gchar *arg_path2_data,
    guint arg_flags,
    const gchar *arg_progress_obj_path,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Move",
    g_variant_new ("(^ay^ayuo)",
                   arg_path1_data,
                   arg_path2_data,
                   arg_flags,
                   arg_progress_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_move_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_move().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_move().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_move_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_move_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path1_data: Argument to pass with the method invocation.
 * @arg_path2_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_progress_obj_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Move">Move()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_move() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_move_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path1_data,
    const gchar *arg_path2_data,
    guint arg_flags,
    const gchar *arg_progress_obj_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Move",
    g_variant_new ("(^ay^ayuo)",
                   arg_path1_data,
                   arg_path2_data,
                   arg_flags,
                   arg_progress_obj_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_push:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_local_path: Argument to pass with the method invocation.
 * @arg_send_progress: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_progress_obj_path: Argument to pass with the method invocation.
 * @arg_remove_source: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Push">Push()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_push_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_push_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_push (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_local_path,
    gboolean arg_send_progress,
    guint arg_flags,
    const gchar *arg_progress_obj_path,
    gboolean arg_remove_source,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Push",
    g_variant_new ("(^ay^aybuob)",
                   arg_path_data,
                   arg_local_path,
                   arg_send_progress,
                   arg_flags,
                   arg_progress_obj_path,
                   arg_remove_source),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_push_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_push().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_push().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_push_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_push_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_local_path: Argument to pass with the method invocation.
 * @arg_send_progress: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_progress_obj_path: Argument to pass with the method invocation.
 * @arg_remove_source: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Push">Push()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_push() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_push_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_local_path,
    gboolean arg_send_progress,
    guint arg_flags,
    const gchar *arg_progress_obj_path,
    gboolean arg_remove_source,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Push",
    g_variant_new ("(^ay^aybuob)",
                   arg_path_data,
                   arg_local_path,
                   arg_send_progress,
                   arg_flags,
                   arg_progress_obj_path,
                   arg_remove_source),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_pull:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_local_path: Argument to pass with the method invocation.
 * @arg_send_progress: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_progress_obj_path: Argument to pass with the method invocation.
 * @arg_remove_source: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Pull">Pull()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_pull_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_pull_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_pull (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_local_path,
    gboolean arg_send_progress,
    guint arg_flags,
    const gchar *arg_progress_obj_path,
    gboolean arg_remove_source,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Pull",
    g_variant_new ("(^ay^aybuob)",
                   arg_path_data,
                   arg_local_path,
                   arg_send_progress,
                   arg_flags,
                   arg_progress_obj_path,
                   arg_remove_source),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_pull_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_pull().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_pull().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_pull_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_pull_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_local_path: Argument to pass with the method invocation.
 * @arg_send_progress: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_progress_obj_path: Argument to pass with the method invocation.
 * @arg_remove_source: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.Pull">Pull()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_pull() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_pull_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    const gchar *arg_local_path,
    gboolean arg_send_progress,
    guint arg_flags,
    const gchar *arg_progress_obj_path,
    gboolean arg_remove_source,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Pull",
    g_variant_new ("(^ay^aybuob)",
                   arg_path_data,
                   arg_local_path,
                   arg_send_progress,
                   arg_flags,
                   arg_progress_obj_path,
                   arg_remove_source),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_set_attribute:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_attribute: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.SetAttribute">SetAttribute()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_set_attribute_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_set_attribute_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_set_attribute (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_flags,
    GVariant *arg_attribute,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetAttribute",
    g_variant_new ("(^ayu@(suv))",
                   arg_path_data,
                   arg_flags,
                   arg_attribute),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_set_attribute_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_set_attribute().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_set_attribute().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_set_attribute_finish (
    GVfsDBusMount *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_set_attribute_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @arg_attribute: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.SetAttribute">SetAttribute()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_set_attribute() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_set_attribute_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    guint arg_flags,
    GVariant *arg_attribute,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetAttribute",
    g_variant_new ("(^ayu@(suv))",
                   arg_path_data,
                   arg_flags,
                   arg_attribute),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_query_settable_attributes:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.QuerySettableAttributes">QuerySettableAttributes()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_query_settable_attributes_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_query_settable_attributes_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_query_settable_attributes (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "QuerySettableAttributes",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_query_settable_attributes_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @out_list: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_query_settable_attributes().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_query_settable_attributes().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_query_settable_attributes_finish (
    GVfsDBusMount *proxy,
    GVariant **out_list,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(suu))",
                 out_list);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_query_settable_attributes_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @out_list: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.QuerySettableAttributes">QuerySettableAttributes()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_query_settable_attributes() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_query_settable_attributes_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GVariant **out_list,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "QuerySettableAttributes",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(suu))",
                 out_list);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_query_writable_namespaces:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.QueryWritableNamespaces">QueryWritableNamespaces()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_query_writable_namespaces_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_query_writable_namespaces_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_query_writable_namespaces (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "QueryWritableNamespaces",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_query_writable_namespaces_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @out_list: (out): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_query_writable_namespaces().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_query_writable_namespaces().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_query_writable_namespaces_finish (
    GVfsDBusMount *proxy,
    GVariant **out_list,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(suu))",
                 out_list);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_query_writable_namespaces_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @out_list: (out): Return location for return parameter or %NULL to ignore.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.QueryWritableNamespaces">QueryWritableNamespaces()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_query_writable_namespaces() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_query_writable_namespaces_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GVariant **out_list,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "QueryWritableNamespaces",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(suu))",
                 out_list);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_open_icon_for_read:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @fd_list: (allow-none): A #GUnixFDList or %NULL.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.OpenIconForRead">OpenIconForRead()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_call_open_icon_for_read_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_call_open_icon_for_read_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_mount_call_open_icon_for_read (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GUnixFDList *fd_list,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call_with_unix_fd_list (G_DBUS_PROXY (proxy),
    "OpenIconForRead",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_mount_call_open_icon_for_read_finish:
 * @proxy: A #GVfsDBusMountProxy.
 * @out_fd_id: (out): Return location for return parameter or %NULL to ignore.
 * @out_can_seek: (out): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out): Return location for a #GUnixFDList or %NULL.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_call_open_icon_for_read().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_mount_call_open_icon_for_read().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_open_icon_for_read_finish (
    GVfsDBusMount *proxy,
    GVariant **out_fd_id,
    gboolean *out_can_seek,
    GUnixFDList **out_fd_list,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_finish (G_DBUS_PROXY (proxy), out_fd_list, res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@hb)",
                 out_fd_id,
                 out_can_seek);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_call_open_icon_for_read_sync:
 * @proxy: A #GVfsDBusMountProxy.
 * @arg_path_data: Argument to pass with the method invocation.
 * @fd_list: (allow-none): A #GUnixFDList or %NULL.
 * @out_fd_id: (out): Return location for return parameter or %NULL to ignore.
 * @out_can_seek: (out): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out): Return location for a #GUnixFDList or %NULL.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Mount.OpenIconForRead">OpenIconForRead()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_call_open_icon_for_read() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_mount_call_open_icon_for_read_sync (
    GVfsDBusMount *proxy,
    const gchar *arg_path_data,
    GUnixFDList  *fd_list,
    GVariant **out_fd_id,
    gboolean *out_can_seek,
    GUnixFDList **out_fd_list,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_sync (G_DBUS_PROXY (proxy),
    "OpenIconForRead",
    g_variant_new ("(^ay)",
                   arg_path_data),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    out_fd_list,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@hb)",
                 out_fd_id,
                 out_can_seek);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_mount_complete_unmount:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.Unmount">Unmount()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_unmount (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_open_for_read:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @fd_list: (allow-none): A #GUnixFDList or %NULL.
 * @fd_id: Parameter to return.
 * @can_seek: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.OpenForRead">OpenForRead()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_open_for_read (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    GVariant *fd_id,
    gboolean can_seek)
{
  g_dbus_method_invocation_return_value_with_unix_fd_list (invocation,
    g_variant_new ("(@hb)",
                   fd_id,
                   can_seek),
    fd_list);
}

/**
 * gvfs_dbus_mount_complete_open_for_write:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @fd_list: (allow-none): A #GUnixFDList or %NULL.
 * @fd_id: Parameter to return.
 * @can_seek: Parameter to return.
 * @initial_offset: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.OpenForWrite">OpenForWrite()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_open_for_write (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    GVariant *fd_id,
    gboolean can_seek,
    guint64 initial_offset)
{
  g_dbus_method_invocation_return_value_with_unix_fd_list (invocation,
    g_variant_new ("(@hbt)",
                   fd_id,
                   can_seek,
                   initial_offset),
    fd_list);
}

/**
 * gvfs_dbus_mount_complete_query_info:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @info: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.QueryInfo">QueryInfo()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_query_info (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation,
    GVariant *info)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@a(suv))",
                   info));
}

/**
 * gvfs_dbus_mount_complete_query_filesystem_info:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @info: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.QueryFilesystemInfo">QueryFilesystemInfo()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_query_filesystem_info (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation,
    GVariant *info)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@a(suv))",
                   info));
}

/**
 * gvfs_dbus_mount_complete_enumerate:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.Enumerate">Enumerate()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_enumerate (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_create_directory_monitor:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @obj_path: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.CreateDirectoryMonitor">CreateDirectoryMonitor()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_create_directory_monitor (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation,
    const gchar *obj_path)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(s)",
                   obj_path));
}

/**
 * gvfs_dbus_mount_complete_create_file_monitor:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @obj_path: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.CreateFileMonitor">CreateFileMonitor()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_create_file_monitor (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation,
    const gchar *obj_path)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(s)",
                   obj_path));
}

/**
 * gvfs_dbus_mount_complete_mount_mountable:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @is_uri: Parameter to return.
 * @path: Parameter to return.
 * @must_mount_location: Parameter to return.
 * @mount_spec: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.MountMountable">MountMountable()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_mount_mountable (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation,
    gboolean is_uri,
    const gchar *path,
    gboolean must_mount_location,
    GVariant *mount_spec)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(b^ayb@(aya{sv}))",
                   is_uri,
                   path,
                   must_mount_location,
                   mount_spec));
}

/**
 * gvfs_dbus_mount_complete_unmount_mountable:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.UnmountMountable">UnmountMountable()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_unmount_mountable (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_eject_mountable:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.EjectMountable">EjectMountable()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_eject_mountable (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_start_mountable:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.StartMountable">StartMountable()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_start_mountable (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_stop_mountable:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.StopMountable">StopMountable()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_stop_mountable (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_poll_mountable:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.PollMountable">PollMountable()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_poll_mountable (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_set_display_name:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @new_path: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.SetDisplayName">SetDisplayName()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_set_display_name (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation,
    const gchar *new_path)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(^ay)",
                   new_path));
}

/**
 * gvfs_dbus_mount_complete_delete:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.Delete">Delete()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_delete (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_trash:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.Trash">Trash()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_trash (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_make_directory:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.MakeDirectory">MakeDirectory()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_make_directory (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_make_symbolic_link:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.MakeSymbolicLink">MakeSymbolicLink()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_make_symbolic_link (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_copy:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.Copy">Copy()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_copy (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_move:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.Move">Move()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_move (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_push:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.Push">Push()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_push (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_pull:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.Pull">Pull()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_pull (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_set_attribute:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.SetAttribute">SetAttribute()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_set_attribute (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_mount_complete_query_settable_attributes:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @list: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.QuerySettableAttributes">QuerySettableAttributes()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_query_settable_attributes (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation,
    GVariant *list)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@a(suu))",
                   list));
}

/**
 * gvfs_dbus_mount_complete_query_writable_namespaces:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @list: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.QueryWritableNamespaces">QueryWritableNamespaces()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_query_writable_namespaces (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation,
    GVariant *list)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@a(suu))",
                   list));
}

/**
 * gvfs_dbus_mount_complete_open_icon_for_read:
 * @object: A #GVfsDBusMount.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @fd_list: (allow-none): A #GUnixFDList or %NULL.
 * @fd_id: Parameter to return.
 * @can_seek: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Mount.OpenIconForRead">OpenIconForRead()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_mount_complete_open_icon_for_read (
    GVfsDBusMount *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    GVariant *fd_id,
    gboolean can_seek)
{
  g_dbus_method_invocation_return_value_with_unix_fd_list (invocation,
    g_variant_new ("(@hb)",
                   fd_id,
                   can_seek),
    fd_list);
}

/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusMountProxy:
 *
 * The #GVfsDBusMountProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusMountProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusMountProxy.
 */

struct _GVfsDBusMountProxyPrivate
{
  GData *qdata;
};

static void gvfs_dbus_mount_proxy_iface_init (GVfsDBusMountIface *iface);

G_DEFINE_TYPE_WITH_CODE (GVfsDBusMountProxy, gvfs_dbus_mount_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (GVfsDBusMountProxy)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_MOUNT, gvfs_dbus_mount_proxy_iface_init));

static void
gvfs_dbus_mount_proxy_finalize (GObject *object)
{
  GVfsDBusMountProxy *proxy = GVFS_DBUS_MOUNT_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (gvfs_dbus_mount_proxy_parent_class)->finalize (object);
}

static void
gvfs_dbus_mount_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_mount_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_mount_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], GVFS_DBUS_TYPE_MOUNT);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_MOUNT);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
gvfs_dbus_mount_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  GVfsDBusMountProxy *proxy = GVFS_DBUS_MOUNT_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
gvfs_dbus_mount_proxy_init (GVfsDBusMountProxy *proxy)
{
  proxy->priv = gvfs_dbus_mount_proxy_get_instance_private (proxy);
  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), gvfs_dbus_mount_interface_info ());
}

static void
gvfs_dbus_mount_proxy_class_init (GVfsDBusMountProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = gvfs_dbus_mount_proxy_finalize;
  gobject_class->get_property = gvfs_dbus_mount_proxy_get_property;
  gobject_class->set_property = gvfs_dbus_mount_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = gvfs_dbus_mount_proxy_g_signal;
  proxy_class->g_properties_changed = gvfs_dbus_mount_proxy_g_properties_changed;

}

static void
gvfs_dbus_mount_proxy_iface_init (GVfsDBusMountIface *iface)
{
}

/**
 * gvfs_dbus_mount_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Mount.top_of_page">org.gtk.vfs.Mount</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_proxy_new_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_mount_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_MOUNT_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Mount", NULL);
}

/**
 * gvfs_dbus_mount_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_mount_proxy_new().
 *
 * Returns: (transfer full) (type GVfsDBusMountProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMount *
gvfs_dbus_mount_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_MOUNT (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_mount_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Mount.top_of_page">org.gtk.vfs.Mount</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusMountProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMount *
gvfs_dbus_mount_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_MOUNT_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Mount", NULL);
  if (ret != NULL)
    return GVFS_DBUS_MOUNT (ret);
  else
    return NULL;
}


/**
 * gvfs_dbus_mount_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like gvfs_dbus_mount_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_mount_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See gvfs_dbus_mount_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_mount_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_MOUNT_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Mount", NULL);
}

/**
 * gvfs_dbus_mount_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_mount_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_mount_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type GVfsDBusMountProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMount *
gvfs_dbus_mount_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_MOUNT (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_mount_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like gvfs_dbus_mount_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_mount_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusMountProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMount *
gvfs_dbus_mount_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_MOUNT_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Mount", NULL);
  if (ret != NULL)
    return GVFS_DBUS_MOUNT (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusMountSkeleton:
 *
 * The #GVfsDBusMountSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusMountSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusMountSkeleton.
 */

struct _GVfsDBusMountSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_gvfs_dbus_mount_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  GVfsDBusMountSkeleton *skeleton = GVFS_DBUS_MOUNT_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint num_extra;
  guint n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], GVFS_DBUS_TYPE_MOUNT);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_MOUNT);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_gvfs_dbus_mount_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  GVfsDBusMountSkeleton *skeleton = GVFS_DBUS_MOUNT_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_gvfs_dbus_mount_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  GVfsDBusMountSkeleton *skeleton = GVFS_DBUS_MOUNT_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_mount_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _gvfs_dbus_mount_skeleton_vtable =
{
  _gvfs_dbus_mount_skeleton_handle_method_call,
  _gvfs_dbus_mount_skeleton_handle_get_property,
  _gvfs_dbus_mount_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
gvfs_dbus_mount_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return gvfs_dbus_mount_interface_info ();
}

static GDBusInterfaceVTable *
gvfs_dbus_mount_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_gvfs_dbus_mount_skeleton_vtable;
}

static GVariant *
gvfs_dbus_mount_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  GVfsDBusMountSkeleton *skeleton = GVFS_DBUS_MOUNT_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  if (_gvfs_dbus_mount_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _gvfs_dbus_mount_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _gvfs_dbus_mount_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _gvfs_dbus_mount_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.vfs.Mount", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
gvfs_dbus_mount_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
}

static void gvfs_dbus_mount_skeleton_iface_init (GVfsDBusMountIface *iface);
G_DEFINE_TYPE_WITH_CODE (GVfsDBusMountSkeleton, gvfs_dbus_mount_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (GVfsDBusMountSkeleton)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_MOUNT, gvfs_dbus_mount_skeleton_iface_init));

static void
gvfs_dbus_mount_skeleton_finalize (GObject *object)
{
  GVfsDBusMountSkeleton *skeleton = GVFS_DBUS_MOUNT_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (gvfs_dbus_mount_skeleton_parent_class)->finalize (object);
}

static void
gvfs_dbus_mount_skeleton_init (GVfsDBusMountSkeleton *skeleton)
{
  skeleton->priv = gvfs_dbus_mount_skeleton_get_instance_private (skeleton);
  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
gvfs_dbus_mount_skeleton_class_init (GVfsDBusMountSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = gvfs_dbus_mount_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = gvfs_dbus_mount_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = gvfs_dbus_mount_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = gvfs_dbus_mount_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = gvfs_dbus_mount_skeleton_dbus_interface_get_vtable;
}

static void
gvfs_dbus_mount_skeleton_iface_init (GVfsDBusMountIface *iface)
{
}

/**
 * gvfs_dbus_mount_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Mount.top_of_page">org.gtk.vfs.Mount</link>.
 *
 * Returns: (transfer full) (type GVfsDBusMountSkeleton): The skeleton object.
 */
GVfsDBusMount *
gvfs_dbus_mount_skeleton_new (void)
{
  return GVFS_DBUS_MOUNT (g_object_new (GVFS_DBUS_TYPE_MOUNT_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gtk.vfs.Enumerator
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:GVfsDBusEnumerator
 * @title: GVfsDBusEnumerator
 * @short_description: Generated C code for the org.gtk.vfs.Enumerator D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gtk-vfs-Enumerator.top_of_page">org.gtk.vfs.Enumerator</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gtk.vfs.Enumerator ---- */

static const _ExtendedGDBusMethodInfo _gvfs_dbus_enumerator_method_info_done =
{
  {
    -1,
    (gchar *) "Done",
    NULL,
    NULL,
    NULL
  },
  "handle-done",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_enumerator_method_info_got_info_IN_ARG_infos =
{
  {
    -1,
    (gchar *) "infos",
    (gchar *) "aa(suv)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_enumerator_method_info_got_info_IN_ARG_pointers[] =
{
  &_gvfs_dbus_enumerator_method_info_got_info_IN_ARG_infos,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_enumerator_method_info_got_info =
{
  {
    -1,
    (gchar *) "GotInfo",
    (GDBusArgInfo **) &_gvfs_dbus_enumerator_method_info_got_info_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-got-info",
  FALSE
};

static const _ExtendedGDBusMethodInfo * const _gvfs_dbus_enumerator_method_info_pointers[] =
{
  &_gvfs_dbus_enumerator_method_info_done,
  &_gvfs_dbus_enumerator_method_info_got_info,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _gvfs_dbus_enumerator_interface_info =
{
  {
    -1,
    (gchar *) "org.gtk.vfs.Enumerator",
    (GDBusMethodInfo **) &_gvfs_dbus_enumerator_method_info_pointers,
    NULL,
    NULL,
    NULL
  },
  "enumerator",
};


/**
 * gvfs_dbus_enumerator_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gtk-vfs-Enumerator.top_of_page">org.gtk.vfs.Enumerator</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
gvfs_dbus_enumerator_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_gvfs_dbus_enumerator_interface_info.parent_struct;
}

/**
 * gvfs_dbus_enumerator_override_properties:
 * @klass: The class structure for a #GObject<!-- -->-derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #GVfsDBusEnumerator interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
gvfs_dbus_enumerator_override_properties (GObjectClass *klass, guint property_id_begin)
{
  return property_id_begin - 1;
}



/**
 * GVfsDBusEnumerator:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Enumerator.top_of_page">org.gtk.vfs.Enumerator</link>.
 */

/**
 * GVfsDBusEnumeratorIface:
 * @parent_iface: The parent interface.
 * @handle_done: Handler for the #GVfsDBusEnumerator::handle-done signal.
 * @handle_got_info: Handler for the #GVfsDBusEnumerator::handle-got-info signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Enumerator.top_of_page">org.gtk.vfs.Enumerator</link>.
 */

typedef GVfsDBusEnumeratorIface GVfsDBusEnumeratorInterface;
G_DEFINE_INTERFACE (GVfsDBusEnumerator, gvfs_dbus_enumerator, G_TYPE_OBJECT);

static void
gvfs_dbus_enumerator_default_init (GVfsDBusEnumeratorIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * GVfsDBusEnumerator::handle-done:
   * @object: A #GVfsDBusEnumerator.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Enumerator.Done">Done()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_enumerator_complete_done() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-done",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusEnumeratorIface, handle_done),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * GVfsDBusEnumerator::handle-got-info:
   * @object: A #GVfsDBusEnumerator.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_infos: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Enumerator.GotInfo">GotInfo()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_enumerator_complete_got_info() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-got-info",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusEnumeratorIface, handle_got_info),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT);

}

/**
 * gvfs_dbus_enumerator_call_done:
 * @proxy: A #GVfsDBusEnumeratorProxy.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Enumerator.Done">Done()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_enumerator_call_done_finish() to get the result of the operation.
 *
 * See gvfs_dbus_enumerator_call_done_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_enumerator_call_done (
    GVfsDBusEnumerator *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Done",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_enumerator_call_done_finish:
 * @proxy: A #GVfsDBusEnumeratorProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_enumerator_call_done().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_enumerator_call_done().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_enumerator_call_done_finish (
    GVfsDBusEnumerator *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_enumerator_call_done_sync:
 * @proxy: A #GVfsDBusEnumeratorProxy.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Enumerator.Done">Done()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_enumerator_call_done() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_enumerator_call_done_sync (
    GVfsDBusEnumerator *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Done",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_enumerator_call_got_info:
 * @proxy: A #GVfsDBusEnumeratorProxy.
 * @arg_infos: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Enumerator.GotInfo">GotInfo()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_enumerator_call_got_info_finish() to get the result of the operation.
 *
 * See gvfs_dbus_enumerator_call_got_info_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_enumerator_call_got_info (
    GVfsDBusEnumerator *proxy,
    GVariant *arg_infos,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "GotInfo",
    g_variant_new ("(@aa(suv))",
                   arg_infos),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_enumerator_call_got_info_finish:
 * @proxy: A #GVfsDBusEnumeratorProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_enumerator_call_got_info().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_enumerator_call_got_info().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_enumerator_call_got_info_finish (
    GVfsDBusEnumerator *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_enumerator_call_got_info_sync:
 * @proxy: A #GVfsDBusEnumeratorProxy.
 * @arg_infos: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Enumerator.GotInfo">GotInfo()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_enumerator_call_got_info() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_enumerator_call_got_info_sync (
    GVfsDBusEnumerator *proxy,
    GVariant *arg_infos,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "GotInfo",
    g_variant_new ("(@aa(suv))",
                   arg_infos),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_enumerator_complete_done:
 * @object: A #GVfsDBusEnumerator.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Enumerator.Done">Done()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_enumerator_complete_done (
    GVfsDBusEnumerator *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_enumerator_complete_got_info:
 * @object: A #GVfsDBusEnumerator.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Enumerator.GotInfo">GotInfo()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_enumerator_complete_got_info (
    GVfsDBusEnumerator *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusEnumeratorProxy:
 *
 * The #GVfsDBusEnumeratorProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusEnumeratorProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusEnumeratorProxy.
 */

struct _GVfsDBusEnumeratorProxyPrivate
{
  GData *qdata;
};

static void gvfs_dbus_enumerator_proxy_iface_init (GVfsDBusEnumeratorIface *iface);

G_DEFINE_TYPE_WITH_CODE (GVfsDBusEnumeratorProxy, gvfs_dbus_enumerator_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (GVfsDBusEnumeratorProxy)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_ENUMERATOR, gvfs_dbus_enumerator_proxy_iface_init));

static void
gvfs_dbus_enumerator_proxy_finalize (GObject *object)
{
  GVfsDBusEnumeratorProxy *proxy = GVFS_DBUS_ENUMERATOR_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (gvfs_dbus_enumerator_proxy_parent_class)->finalize (object);
}

static void
gvfs_dbus_enumerator_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_enumerator_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_enumerator_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_gvfs_dbus_enumerator_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], GVFS_DBUS_TYPE_ENUMERATOR);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_ENUMERATOR);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
gvfs_dbus_enumerator_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  GVfsDBusEnumeratorProxy *proxy = GVFS_DBUS_ENUMERATOR_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_enumerator_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_enumerator_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
gvfs_dbus_enumerator_proxy_init (GVfsDBusEnumeratorProxy *proxy)
{
  proxy->priv = gvfs_dbus_enumerator_proxy_get_instance_private (proxy);
  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), gvfs_dbus_enumerator_interface_info ());
}

static void
gvfs_dbus_enumerator_proxy_class_init (GVfsDBusEnumeratorProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = gvfs_dbus_enumerator_proxy_finalize;
  gobject_class->get_property = gvfs_dbus_enumerator_proxy_get_property;
  gobject_class->set_property = gvfs_dbus_enumerator_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = gvfs_dbus_enumerator_proxy_g_signal;
  proxy_class->g_properties_changed = gvfs_dbus_enumerator_proxy_g_properties_changed;

}

static void
gvfs_dbus_enumerator_proxy_iface_init (GVfsDBusEnumeratorIface *iface)
{
}

/**
 * gvfs_dbus_enumerator_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Enumerator.top_of_page">org.gtk.vfs.Enumerator</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_enumerator_proxy_new_finish() to get the result of the operation.
 *
 * See gvfs_dbus_enumerator_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_enumerator_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_ENUMERATOR_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Enumerator", NULL);
}

/**
 * gvfs_dbus_enumerator_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_enumerator_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_enumerator_proxy_new().
 *
 * Returns: (transfer full) (type GVfsDBusEnumeratorProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusEnumerator *
gvfs_dbus_enumerator_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_ENUMERATOR (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_enumerator_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Enumerator.top_of_page">org.gtk.vfs.Enumerator</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_enumerator_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusEnumeratorProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusEnumerator *
gvfs_dbus_enumerator_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_ENUMERATOR_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Enumerator", NULL);
  if (ret != NULL)
    return GVFS_DBUS_ENUMERATOR (ret);
  else
    return NULL;
}


/**
 * gvfs_dbus_enumerator_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like gvfs_dbus_enumerator_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_enumerator_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See gvfs_dbus_enumerator_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_enumerator_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_ENUMERATOR_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Enumerator", NULL);
}

/**
 * gvfs_dbus_enumerator_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_enumerator_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_enumerator_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type GVfsDBusEnumeratorProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusEnumerator *
gvfs_dbus_enumerator_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_ENUMERATOR (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_enumerator_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like gvfs_dbus_enumerator_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_enumerator_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusEnumeratorProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusEnumerator *
gvfs_dbus_enumerator_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_ENUMERATOR_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Enumerator", NULL);
  if (ret != NULL)
    return GVFS_DBUS_ENUMERATOR (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusEnumeratorSkeleton:
 *
 * The #GVfsDBusEnumeratorSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusEnumeratorSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusEnumeratorSkeleton.
 */

struct _GVfsDBusEnumeratorSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_gvfs_dbus_enumerator_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  GVfsDBusEnumeratorSkeleton *skeleton = GVFS_DBUS_ENUMERATOR_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint num_extra;
  guint n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], GVFS_DBUS_TYPE_ENUMERATOR);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_ENUMERATOR);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_gvfs_dbus_enumerator_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  GVfsDBusEnumeratorSkeleton *skeleton = GVFS_DBUS_ENUMERATOR_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_enumerator_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_gvfs_dbus_enumerator_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  GVfsDBusEnumeratorSkeleton *skeleton = GVFS_DBUS_ENUMERATOR_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_enumerator_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _gvfs_dbus_enumerator_skeleton_vtable =
{
  _gvfs_dbus_enumerator_skeleton_handle_method_call,
  _gvfs_dbus_enumerator_skeleton_handle_get_property,
  _gvfs_dbus_enumerator_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
gvfs_dbus_enumerator_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return gvfs_dbus_enumerator_interface_info ();
}

static GDBusInterfaceVTable *
gvfs_dbus_enumerator_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_gvfs_dbus_enumerator_skeleton_vtable;
}

static GVariant *
gvfs_dbus_enumerator_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  GVfsDBusEnumeratorSkeleton *skeleton = GVFS_DBUS_ENUMERATOR_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  if (_gvfs_dbus_enumerator_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _gvfs_dbus_enumerator_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _gvfs_dbus_enumerator_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _gvfs_dbus_enumerator_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.vfs.Enumerator", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
gvfs_dbus_enumerator_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
}

static void gvfs_dbus_enumerator_skeleton_iface_init (GVfsDBusEnumeratorIface *iface);
G_DEFINE_TYPE_WITH_CODE (GVfsDBusEnumeratorSkeleton, gvfs_dbus_enumerator_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (GVfsDBusEnumeratorSkeleton)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_ENUMERATOR, gvfs_dbus_enumerator_skeleton_iface_init));

static void
gvfs_dbus_enumerator_skeleton_finalize (GObject *object)
{
  GVfsDBusEnumeratorSkeleton *skeleton = GVFS_DBUS_ENUMERATOR_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (gvfs_dbus_enumerator_skeleton_parent_class)->finalize (object);
}

static void
gvfs_dbus_enumerator_skeleton_init (GVfsDBusEnumeratorSkeleton *skeleton)
{
  skeleton->priv = gvfs_dbus_enumerator_skeleton_get_instance_private (skeleton);
  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
gvfs_dbus_enumerator_skeleton_class_init (GVfsDBusEnumeratorSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = gvfs_dbus_enumerator_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = gvfs_dbus_enumerator_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = gvfs_dbus_enumerator_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = gvfs_dbus_enumerator_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = gvfs_dbus_enumerator_skeleton_dbus_interface_get_vtable;
}

static void
gvfs_dbus_enumerator_skeleton_iface_init (GVfsDBusEnumeratorIface *iface)
{
}

/**
 * gvfs_dbus_enumerator_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Enumerator.top_of_page">org.gtk.vfs.Enumerator</link>.
 *
 * Returns: (transfer full) (type GVfsDBusEnumeratorSkeleton): The skeleton object.
 */
GVfsDBusEnumerator *
gvfs_dbus_enumerator_skeleton_new (void)
{
  return GVFS_DBUS_ENUMERATOR (g_object_new (GVFS_DBUS_TYPE_ENUMERATOR_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gtk.vfs.Progress
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:GVfsDBusProgress
 * @title: GVfsDBusProgress
 * @short_description: Generated C code for the org.gtk.vfs.Progress D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gtk-vfs-Progress.top_of_page">org.gtk.vfs.Progress</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gtk.vfs.Progress ---- */

static const _ExtendedGDBusArgInfo _gvfs_dbus_progress_method_info_progress_IN_ARG_current =
{
  {
    -1,
    (gchar *) "current",
    (gchar *) "t",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_progress_method_info_progress_IN_ARG_total =
{
  {
    -1,
    (gchar *) "total",
    (gchar *) "t",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_progress_method_info_progress_IN_ARG_pointers[] =
{
  &_gvfs_dbus_progress_method_info_progress_IN_ARG_current,
  &_gvfs_dbus_progress_method_info_progress_IN_ARG_total,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_progress_method_info_progress =
{
  {
    -1,
    (gchar *) "Progress",
    (GDBusArgInfo **) &_gvfs_dbus_progress_method_info_progress_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-progress",
  FALSE
};

static const _ExtendedGDBusMethodInfo * const _gvfs_dbus_progress_method_info_pointers[] =
{
  &_gvfs_dbus_progress_method_info_progress,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _gvfs_dbus_progress_interface_info =
{
  {
    -1,
    (gchar *) "org.gtk.vfs.Progress",
    (GDBusMethodInfo **) &_gvfs_dbus_progress_method_info_pointers,
    NULL,
    NULL,
    NULL
  },
  "progress",
};


/**
 * gvfs_dbus_progress_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gtk-vfs-Progress.top_of_page">org.gtk.vfs.Progress</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
gvfs_dbus_progress_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_gvfs_dbus_progress_interface_info.parent_struct;
}

/**
 * gvfs_dbus_progress_override_properties:
 * @klass: The class structure for a #GObject<!-- -->-derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #GVfsDBusProgress interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
gvfs_dbus_progress_override_properties (GObjectClass *klass, guint property_id_begin)
{
  return property_id_begin - 1;
}



/**
 * GVfsDBusProgress:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Progress.top_of_page">org.gtk.vfs.Progress</link>.
 */

/**
 * GVfsDBusProgressIface:
 * @parent_iface: The parent interface.
 * @handle_progress: Handler for the #GVfsDBusProgress::handle-progress signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Progress.top_of_page">org.gtk.vfs.Progress</link>.
 */

typedef GVfsDBusProgressIface GVfsDBusProgressInterface;
G_DEFINE_INTERFACE (GVfsDBusProgress, gvfs_dbus_progress, G_TYPE_OBJECT);

static void
gvfs_dbus_progress_default_init (GVfsDBusProgressIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * GVfsDBusProgress::handle-progress:
   * @object: A #GVfsDBusProgress.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_current: Argument passed by remote caller.
   * @arg_total: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Progress.Progress">Progress()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_progress_complete_progress() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-progress",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusProgressIface, handle_progress),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT64, G_TYPE_UINT64);

}

/**
 * gvfs_dbus_progress_call_progress:
 * @proxy: A #GVfsDBusProgressProxy.
 * @arg_current: Argument to pass with the method invocation.
 * @arg_total: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Progress.Progress">Progress()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_progress_call_progress_finish() to get the result of the operation.
 *
 * See gvfs_dbus_progress_call_progress_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_progress_call_progress (
    GVfsDBusProgress *proxy,
    guint64 arg_current,
    guint64 arg_total,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Progress",
    g_variant_new ("(tt)",
                   arg_current,
                   arg_total),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_progress_call_progress_finish:
 * @proxy: A #GVfsDBusProgressProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_progress_call_progress().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_progress_call_progress().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_progress_call_progress_finish (
    GVfsDBusProgress *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_progress_call_progress_sync:
 * @proxy: A #GVfsDBusProgressProxy.
 * @arg_current: Argument to pass with the method invocation.
 * @arg_total: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Progress.Progress">Progress()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_progress_call_progress() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_progress_call_progress_sync (
    GVfsDBusProgress *proxy,
    guint64 arg_current,
    guint64 arg_total,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Progress",
    g_variant_new ("(tt)",
                   arg_current,
                   arg_total),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_progress_complete_progress:
 * @object: A #GVfsDBusProgress.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Progress.Progress">Progress()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_progress_complete_progress (
    GVfsDBusProgress *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusProgressProxy:
 *
 * The #GVfsDBusProgressProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusProgressProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusProgressProxy.
 */

struct _GVfsDBusProgressProxyPrivate
{
  GData *qdata;
};

static void gvfs_dbus_progress_proxy_iface_init (GVfsDBusProgressIface *iface);

G_DEFINE_TYPE_WITH_CODE (GVfsDBusProgressProxy, gvfs_dbus_progress_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (GVfsDBusProgressProxy)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_PROGRESS, gvfs_dbus_progress_proxy_iface_init));

static void
gvfs_dbus_progress_proxy_finalize (GObject *object)
{
  GVfsDBusProgressProxy *proxy = GVFS_DBUS_PROGRESS_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (gvfs_dbus_progress_proxy_parent_class)->finalize (object);
}

static void
gvfs_dbus_progress_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_progress_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_progress_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_gvfs_dbus_progress_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], GVFS_DBUS_TYPE_PROGRESS);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_PROGRESS);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
gvfs_dbus_progress_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  GVfsDBusProgressProxy *proxy = GVFS_DBUS_PROGRESS_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_progress_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_progress_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
gvfs_dbus_progress_proxy_init (GVfsDBusProgressProxy *proxy)
{
  proxy->priv = gvfs_dbus_progress_proxy_get_instance_private (proxy);
  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), gvfs_dbus_progress_interface_info ());
}

static void
gvfs_dbus_progress_proxy_class_init (GVfsDBusProgressProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = gvfs_dbus_progress_proxy_finalize;
  gobject_class->get_property = gvfs_dbus_progress_proxy_get_property;
  gobject_class->set_property = gvfs_dbus_progress_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = gvfs_dbus_progress_proxy_g_signal;
  proxy_class->g_properties_changed = gvfs_dbus_progress_proxy_g_properties_changed;

}

static void
gvfs_dbus_progress_proxy_iface_init (GVfsDBusProgressIface *iface)
{
}

/**
 * gvfs_dbus_progress_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Progress.top_of_page">org.gtk.vfs.Progress</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_progress_proxy_new_finish() to get the result of the operation.
 *
 * See gvfs_dbus_progress_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_progress_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_PROGRESS_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Progress", NULL);
}

/**
 * gvfs_dbus_progress_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_progress_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_progress_proxy_new().
 *
 * Returns: (transfer full) (type GVfsDBusProgressProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusProgress *
gvfs_dbus_progress_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_PROGRESS (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_progress_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Progress.top_of_page">org.gtk.vfs.Progress</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_progress_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusProgressProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusProgress *
gvfs_dbus_progress_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_PROGRESS_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Progress", NULL);
  if (ret != NULL)
    return GVFS_DBUS_PROGRESS (ret);
  else
    return NULL;
}


/**
 * gvfs_dbus_progress_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like gvfs_dbus_progress_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_progress_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See gvfs_dbus_progress_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_progress_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_PROGRESS_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Progress", NULL);
}

/**
 * gvfs_dbus_progress_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_progress_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_progress_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type GVfsDBusProgressProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusProgress *
gvfs_dbus_progress_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_PROGRESS (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_progress_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like gvfs_dbus_progress_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_progress_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusProgressProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusProgress *
gvfs_dbus_progress_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_PROGRESS_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Progress", NULL);
  if (ret != NULL)
    return GVFS_DBUS_PROGRESS (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusProgressSkeleton:
 *
 * The #GVfsDBusProgressSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusProgressSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusProgressSkeleton.
 */

struct _GVfsDBusProgressSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_gvfs_dbus_progress_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  GVfsDBusProgressSkeleton *skeleton = GVFS_DBUS_PROGRESS_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint num_extra;
  guint n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], GVFS_DBUS_TYPE_PROGRESS);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_PROGRESS);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_gvfs_dbus_progress_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  GVfsDBusProgressSkeleton *skeleton = GVFS_DBUS_PROGRESS_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_progress_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_gvfs_dbus_progress_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  GVfsDBusProgressSkeleton *skeleton = GVFS_DBUS_PROGRESS_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_progress_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _gvfs_dbus_progress_skeleton_vtable =
{
  _gvfs_dbus_progress_skeleton_handle_method_call,
  _gvfs_dbus_progress_skeleton_handle_get_property,
  _gvfs_dbus_progress_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
gvfs_dbus_progress_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return gvfs_dbus_progress_interface_info ();
}

static GDBusInterfaceVTable *
gvfs_dbus_progress_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_gvfs_dbus_progress_skeleton_vtable;
}

static GVariant *
gvfs_dbus_progress_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  GVfsDBusProgressSkeleton *skeleton = GVFS_DBUS_PROGRESS_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  if (_gvfs_dbus_progress_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _gvfs_dbus_progress_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _gvfs_dbus_progress_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _gvfs_dbus_progress_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.vfs.Progress", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
gvfs_dbus_progress_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
}

static void gvfs_dbus_progress_skeleton_iface_init (GVfsDBusProgressIface *iface);
G_DEFINE_TYPE_WITH_CODE (GVfsDBusProgressSkeleton, gvfs_dbus_progress_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (GVfsDBusProgressSkeleton)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_PROGRESS, gvfs_dbus_progress_skeleton_iface_init));

static void
gvfs_dbus_progress_skeleton_finalize (GObject *object)
{
  GVfsDBusProgressSkeleton *skeleton = GVFS_DBUS_PROGRESS_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (gvfs_dbus_progress_skeleton_parent_class)->finalize (object);
}

static void
gvfs_dbus_progress_skeleton_init (GVfsDBusProgressSkeleton *skeleton)
{
  skeleton->priv = gvfs_dbus_progress_skeleton_get_instance_private (skeleton);
  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
gvfs_dbus_progress_skeleton_class_init (GVfsDBusProgressSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = gvfs_dbus_progress_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = gvfs_dbus_progress_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = gvfs_dbus_progress_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = gvfs_dbus_progress_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = gvfs_dbus_progress_skeleton_dbus_interface_get_vtable;
}

static void
gvfs_dbus_progress_skeleton_iface_init (GVfsDBusProgressIface *iface)
{
}

/**
 * gvfs_dbus_progress_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Progress.top_of_page">org.gtk.vfs.Progress</link>.
 *
 * Returns: (transfer full) (type GVfsDBusProgressSkeleton): The skeleton object.
 */
GVfsDBusProgress *
gvfs_dbus_progress_skeleton_new (void)
{
  return GVFS_DBUS_PROGRESS (g_object_new (GVFS_DBUS_TYPE_PROGRESS_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gtk.vfs.Monitor
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:GVfsDBusMonitor
 * @title: GVfsDBusMonitor
 * @short_description: Generated C code for the org.gtk.vfs.Monitor D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gtk-vfs-Monitor.top_of_page">org.gtk.vfs.Monitor</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gtk.vfs.Monitor ---- */

static const _ExtendedGDBusArgInfo _gvfs_dbus_monitor_method_info_subscribe_IN_ARG_object_path =
{
  {
    -1,
    (gchar *) "object_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_monitor_method_info_subscribe_IN_ARG_pointers[] =
{
  &_gvfs_dbus_monitor_method_info_subscribe_IN_ARG_object_path,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_monitor_method_info_subscribe =
{
  {
    -1,
    (gchar *) "Subscribe",
    (GDBusArgInfo **) &_gvfs_dbus_monitor_method_info_subscribe_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-subscribe",
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_monitor_method_info_unsubscribe_IN_ARG_object_path =
{
  {
    -1,
    (gchar *) "object_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_monitor_method_info_unsubscribe_IN_ARG_pointers[] =
{
  &_gvfs_dbus_monitor_method_info_unsubscribe_IN_ARG_object_path,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_monitor_method_info_unsubscribe =
{
  {
    -1,
    (gchar *) "Unsubscribe",
    (GDBusArgInfo **) &_gvfs_dbus_monitor_method_info_unsubscribe_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-unsubscribe",
  FALSE
};

static const _ExtendedGDBusMethodInfo * const _gvfs_dbus_monitor_method_info_pointers[] =
{
  &_gvfs_dbus_monitor_method_info_subscribe,
  &_gvfs_dbus_monitor_method_info_unsubscribe,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _gvfs_dbus_monitor_interface_info =
{
  {
    -1,
    (gchar *) "org.gtk.vfs.Monitor",
    (GDBusMethodInfo **) &_gvfs_dbus_monitor_method_info_pointers,
    NULL,
    NULL,
    NULL
  },
  "monitor",
};


/**
 * gvfs_dbus_monitor_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gtk-vfs-Monitor.top_of_page">org.gtk.vfs.Monitor</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
gvfs_dbus_monitor_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_gvfs_dbus_monitor_interface_info.parent_struct;
}

/**
 * gvfs_dbus_monitor_override_properties:
 * @klass: The class structure for a #GObject<!-- -->-derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #GVfsDBusMonitor interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
gvfs_dbus_monitor_override_properties (GObjectClass *klass, guint property_id_begin)
{
  return property_id_begin - 1;
}



/**
 * GVfsDBusMonitor:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Monitor.top_of_page">org.gtk.vfs.Monitor</link>.
 */

/**
 * GVfsDBusMonitorIface:
 * @parent_iface: The parent interface.
 * @handle_subscribe: Handler for the #GVfsDBusMonitor::handle-subscribe signal.
 * @handle_unsubscribe: Handler for the #GVfsDBusMonitor::handle-unsubscribe signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Monitor.top_of_page">org.gtk.vfs.Monitor</link>.
 */

typedef GVfsDBusMonitorIface GVfsDBusMonitorInterface;
G_DEFINE_INTERFACE (GVfsDBusMonitor, gvfs_dbus_monitor, G_TYPE_OBJECT);

static void
gvfs_dbus_monitor_default_init (GVfsDBusMonitorIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * GVfsDBusMonitor::handle-subscribe:
   * @object: A #GVfsDBusMonitor.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_object_path: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Monitor.Subscribe">Subscribe()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_monitor_complete_subscribe() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-subscribe",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMonitorIface, handle_subscribe),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * GVfsDBusMonitor::handle-unsubscribe:
   * @object: A #GVfsDBusMonitor.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_object_path: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-Monitor.Unsubscribe">Unsubscribe()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_monitor_complete_unsubscribe() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-unsubscribe",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMonitorIface, handle_unsubscribe),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

}

/**
 * gvfs_dbus_monitor_call_subscribe:
 * @proxy: A #GVfsDBusMonitorProxy.
 * @arg_object_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Monitor.Subscribe">Subscribe()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_monitor_call_subscribe_finish() to get the result of the operation.
 *
 * See gvfs_dbus_monitor_call_subscribe_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_monitor_call_subscribe (
    GVfsDBusMonitor *proxy,
    const gchar *arg_object_path,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Subscribe",
    g_variant_new ("(o)",
                   arg_object_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_monitor_call_subscribe_finish:
 * @proxy: A #GVfsDBusMonitorProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_monitor_call_subscribe().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_monitor_call_subscribe().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_monitor_call_subscribe_finish (
    GVfsDBusMonitor *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_monitor_call_subscribe_sync:
 * @proxy: A #GVfsDBusMonitorProxy.
 * @arg_object_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Monitor.Subscribe">Subscribe()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_monitor_call_subscribe() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_monitor_call_subscribe_sync (
    GVfsDBusMonitor *proxy,
    const gchar *arg_object_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Subscribe",
    g_variant_new ("(o)",
                   arg_object_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_monitor_call_unsubscribe:
 * @proxy: A #GVfsDBusMonitorProxy.
 * @arg_object_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Monitor.Unsubscribe">Unsubscribe()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_monitor_call_unsubscribe_finish() to get the result of the operation.
 *
 * See gvfs_dbus_monitor_call_unsubscribe_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_monitor_call_unsubscribe (
    GVfsDBusMonitor *proxy,
    const gchar *arg_object_path,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Unsubscribe",
    g_variant_new ("(o)",
                   arg_object_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_monitor_call_unsubscribe_finish:
 * @proxy: A #GVfsDBusMonitorProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_monitor_call_unsubscribe().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_monitor_call_unsubscribe().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_monitor_call_unsubscribe_finish (
    GVfsDBusMonitor *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_monitor_call_unsubscribe_sync:
 * @proxy: A #GVfsDBusMonitorProxy.
 * @arg_object_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-Monitor.Unsubscribe">Unsubscribe()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_monitor_call_unsubscribe() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_monitor_call_unsubscribe_sync (
    GVfsDBusMonitor *proxy,
    const gchar *arg_object_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Unsubscribe",
    g_variant_new ("(o)",
                   arg_object_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_monitor_complete_subscribe:
 * @object: A #GVfsDBusMonitor.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Monitor.Subscribe">Subscribe()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_monitor_complete_subscribe (
    GVfsDBusMonitor *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * gvfs_dbus_monitor_complete_unsubscribe:
 * @object: A #GVfsDBusMonitor.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-Monitor.Unsubscribe">Unsubscribe()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_monitor_complete_unsubscribe (
    GVfsDBusMonitor *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusMonitorProxy:
 *
 * The #GVfsDBusMonitorProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusMonitorProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusMonitorProxy.
 */

struct _GVfsDBusMonitorProxyPrivate
{
  GData *qdata;
};

static void gvfs_dbus_monitor_proxy_iface_init (GVfsDBusMonitorIface *iface);

G_DEFINE_TYPE_WITH_CODE (GVfsDBusMonitorProxy, gvfs_dbus_monitor_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (GVfsDBusMonitorProxy)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_MONITOR, gvfs_dbus_monitor_proxy_iface_init));

static void
gvfs_dbus_monitor_proxy_finalize (GObject *object)
{
  GVfsDBusMonitorProxy *proxy = GVFS_DBUS_MONITOR_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (gvfs_dbus_monitor_proxy_parent_class)->finalize (object);
}

static void
gvfs_dbus_monitor_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_monitor_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_monitor_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_gvfs_dbus_monitor_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], GVFS_DBUS_TYPE_MONITOR);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_MONITOR);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
gvfs_dbus_monitor_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  GVfsDBusMonitorProxy *proxy = GVFS_DBUS_MONITOR_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_monitor_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_monitor_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
gvfs_dbus_monitor_proxy_init (GVfsDBusMonitorProxy *proxy)
{
  proxy->priv = gvfs_dbus_monitor_proxy_get_instance_private (proxy);
  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), gvfs_dbus_monitor_interface_info ());
}

static void
gvfs_dbus_monitor_proxy_class_init (GVfsDBusMonitorProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = gvfs_dbus_monitor_proxy_finalize;
  gobject_class->get_property = gvfs_dbus_monitor_proxy_get_property;
  gobject_class->set_property = gvfs_dbus_monitor_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = gvfs_dbus_monitor_proxy_g_signal;
  proxy_class->g_properties_changed = gvfs_dbus_monitor_proxy_g_properties_changed;

}

static void
gvfs_dbus_monitor_proxy_iface_init (GVfsDBusMonitorIface *iface)
{
}

/**
 * gvfs_dbus_monitor_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Monitor.top_of_page">org.gtk.vfs.Monitor</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_monitor_proxy_new_finish() to get the result of the operation.
 *
 * See gvfs_dbus_monitor_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_monitor_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_MONITOR_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Monitor", NULL);
}

/**
 * gvfs_dbus_monitor_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_monitor_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_monitor_proxy_new().
 *
 * Returns: (transfer full) (type GVfsDBusMonitorProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMonitor *
gvfs_dbus_monitor_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_MONITOR (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_monitor_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Monitor.top_of_page">org.gtk.vfs.Monitor</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_monitor_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusMonitorProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMonitor *
gvfs_dbus_monitor_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_MONITOR_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Monitor", NULL);
  if (ret != NULL)
    return GVFS_DBUS_MONITOR (ret);
  else
    return NULL;
}


/**
 * gvfs_dbus_monitor_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like gvfs_dbus_monitor_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_monitor_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See gvfs_dbus_monitor_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_monitor_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_MONITOR_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Monitor", NULL);
}

/**
 * gvfs_dbus_monitor_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_monitor_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_monitor_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type GVfsDBusMonitorProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMonitor *
gvfs_dbus_monitor_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_MONITOR (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_monitor_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like gvfs_dbus_monitor_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_monitor_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusMonitorProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMonitor *
gvfs_dbus_monitor_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_MONITOR_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.Monitor", NULL);
  if (ret != NULL)
    return GVFS_DBUS_MONITOR (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusMonitorSkeleton:
 *
 * The #GVfsDBusMonitorSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusMonitorSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusMonitorSkeleton.
 */

struct _GVfsDBusMonitorSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_gvfs_dbus_monitor_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  GVfsDBusMonitorSkeleton *skeleton = GVFS_DBUS_MONITOR_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint num_extra;
  guint n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], GVFS_DBUS_TYPE_MONITOR);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_MONITOR);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_gvfs_dbus_monitor_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  GVfsDBusMonitorSkeleton *skeleton = GVFS_DBUS_MONITOR_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_monitor_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_gvfs_dbus_monitor_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  GVfsDBusMonitorSkeleton *skeleton = GVFS_DBUS_MONITOR_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_monitor_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _gvfs_dbus_monitor_skeleton_vtable =
{
  _gvfs_dbus_monitor_skeleton_handle_method_call,
  _gvfs_dbus_monitor_skeleton_handle_get_property,
  _gvfs_dbus_monitor_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
gvfs_dbus_monitor_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return gvfs_dbus_monitor_interface_info ();
}

static GDBusInterfaceVTable *
gvfs_dbus_monitor_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_gvfs_dbus_monitor_skeleton_vtable;
}

static GVariant *
gvfs_dbus_monitor_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  GVfsDBusMonitorSkeleton *skeleton = GVFS_DBUS_MONITOR_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  if (_gvfs_dbus_monitor_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _gvfs_dbus_monitor_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _gvfs_dbus_monitor_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _gvfs_dbus_monitor_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.vfs.Monitor", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
gvfs_dbus_monitor_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
}

static void gvfs_dbus_monitor_skeleton_iface_init (GVfsDBusMonitorIface *iface);
G_DEFINE_TYPE_WITH_CODE (GVfsDBusMonitorSkeleton, gvfs_dbus_monitor_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (GVfsDBusMonitorSkeleton)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_MONITOR, gvfs_dbus_monitor_skeleton_iface_init));

static void
gvfs_dbus_monitor_skeleton_finalize (GObject *object)
{
  GVfsDBusMonitorSkeleton *skeleton = GVFS_DBUS_MONITOR_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (gvfs_dbus_monitor_skeleton_parent_class)->finalize (object);
}

static void
gvfs_dbus_monitor_skeleton_init (GVfsDBusMonitorSkeleton *skeleton)
{
  skeleton->priv = gvfs_dbus_monitor_skeleton_get_instance_private (skeleton);
  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
gvfs_dbus_monitor_skeleton_class_init (GVfsDBusMonitorSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = gvfs_dbus_monitor_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = gvfs_dbus_monitor_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = gvfs_dbus_monitor_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = gvfs_dbus_monitor_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = gvfs_dbus_monitor_skeleton_dbus_interface_get_vtable;
}

static void
gvfs_dbus_monitor_skeleton_iface_init (GVfsDBusMonitorIface *iface)
{
}

/**
 * gvfs_dbus_monitor_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-Monitor.top_of_page">org.gtk.vfs.Monitor</link>.
 *
 * Returns: (transfer full) (type GVfsDBusMonitorSkeleton): The skeleton object.
 */
GVfsDBusMonitor *
gvfs_dbus_monitor_skeleton_new (void)
{
  return GVFS_DBUS_MONITOR (g_object_new (GVFS_DBUS_TYPE_MONITOR_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gtk.vfs.MonitorClient
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:GVfsDBusMonitorClient
 * @title: GVfsDBusMonitorClient
 * @short_description: Generated C code for the org.gtk.vfs.MonitorClient D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gtk-vfs-MonitorClient.top_of_page">org.gtk.vfs.MonitorClient</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gtk.vfs.MonitorClient ---- */

static const _ExtendedGDBusArgInfo _gvfs_dbus_monitor_client_method_info_changed_IN_ARG_event_type =
{
  {
    -1,
    (gchar *) "event_type",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_monitor_client_method_info_changed_IN_ARG_mount_spec =
{
  {
    -1,
    (gchar *) "mount_spec",
    (gchar *) "(aya{sv})",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_monitor_client_method_info_changed_IN_ARG_file_path =
{
  {
    -1,
    (gchar *) "file_path",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_monitor_client_method_info_changed_IN_ARG_other_mount_spec =
{
  {
    -1,
    (gchar *) "other_mount_spec",
    (gchar *) "(aya{sv})",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _gvfs_dbus_monitor_client_method_info_changed_IN_ARG_other_file_path =
{
  {
    -1,
    (gchar *) "other_file_path",
    (gchar *) "ay",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo * const _gvfs_dbus_monitor_client_method_info_changed_IN_ARG_pointers[] =
{
  &_gvfs_dbus_monitor_client_method_info_changed_IN_ARG_event_type,
  &_gvfs_dbus_monitor_client_method_info_changed_IN_ARG_mount_spec,
  &_gvfs_dbus_monitor_client_method_info_changed_IN_ARG_file_path,
  &_gvfs_dbus_monitor_client_method_info_changed_IN_ARG_other_mount_spec,
  &_gvfs_dbus_monitor_client_method_info_changed_IN_ARG_other_file_path,
  NULL
};

static const _ExtendedGDBusMethodInfo _gvfs_dbus_monitor_client_method_info_changed =
{
  {
    -1,
    (gchar *) "Changed",
    (GDBusArgInfo **) &_gvfs_dbus_monitor_client_method_info_changed_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-changed",
  FALSE
};

static const _ExtendedGDBusMethodInfo * const _gvfs_dbus_monitor_client_method_info_pointers[] =
{
  &_gvfs_dbus_monitor_client_method_info_changed,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _gvfs_dbus_monitor_client_interface_info =
{
  {
    -1,
    (gchar *) "org.gtk.vfs.MonitorClient",
    (GDBusMethodInfo **) &_gvfs_dbus_monitor_client_method_info_pointers,
    NULL,
    NULL,
    NULL
  },
  "monitor-client",
};


/**
 * gvfs_dbus_monitor_client_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gtk-vfs-MonitorClient.top_of_page">org.gtk.vfs.MonitorClient</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
gvfs_dbus_monitor_client_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_gvfs_dbus_monitor_client_interface_info.parent_struct;
}

/**
 * gvfs_dbus_monitor_client_override_properties:
 * @klass: The class structure for a #GObject<!-- -->-derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #GVfsDBusMonitorClient interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
gvfs_dbus_monitor_client_override_properties (GObjectClass *klass, guint property_id_begin)
{
  return property_id_begin - 1;
}



/**
 * GVfsDBusMonitorClient:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MonitorClient.top_of_page">org.gtk.vfs.MonitorClient</link>.
 */

/**
 * GVfsDBusMonitorClientIface:
 * @parent_iface: The parent interface.
 * @handle_changed: Handler for the #GVfsDBusMonitorClient::handle-changed signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MonitorClient.top_of_page">org.gtk.vfs.MonitorClient</link>.
 */

typedef GVfsDBusMonitorClientIface GVfsDBusMonitorClientInterface;
G_DEFINE_INTERFACE (GVfsDBusMonitorClient, gvfs_dbus_monitor_client, G_TYPE_OBJECT);

static void
gvfs_dbus_monitor_client_default_init (GVfsDBusMonitorClientIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * GVfsDBusMonitorClient::handle-changed:
   * @object: A #GVfsDBusMonitorClient.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_event_type: Argument passed by remote caller.
   * @arg_mount_spec: Argument passed by remote caller.
   * @arg_file_path: Argument passed by remote caller.
   * @arg_other_mount_spec: Argument passed by remote caller.
   * @arg_other_file_path: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-vfs-MonitorClient.Changed">Changed()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call gvfs_dbus_monitor_client_complete_changed() or e.g. g_dbus_method_invocation_return_error() on it) and no order signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %TRUE if the invocation was handled, %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-changed",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (GVfsDBusMonitorClientIface, handle_changed),
    g_signal_accumulator_true_handled,
    NULL,
    g_cclosure_marshal_generic,
    G_TYPE_BOOLEAN,
    6,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT, G_TYPE_VARIANT, G_TYPE_STRING, G_TYPE_VARIANT, G_TYPE_STRING);

}

/**
 * gvfs_dbus_monitor_client_call_changed:
 * @proxy: A #GVfsDBusMonitorClientProxy.
 * @arg_event_type: Argument to pass with the method invocation.
 * @arg_mount_spec: Argument to pass with the method invocation.
 * @arg_file_path: Argument to pass with the method invocation.
 * @arg_other_mount_spec: Argument to pass with the method invocation.
 * @arg_other_file_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MonitorClient.Changed">Changed()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_monitor_client_call_changed_finish() to get the result of the operation.
 *
 * See gvfs_dbus_monitor_client_call_changed_sync() for the synchronous, blocking version of this method.
 */
void
gvfs_dbus_monitor_client_call_changed (
    GVfsDBusMonitorClient *proxy,
    guint arg_event_type,
    GVariant *arg_mount_spec,
    const gchar *arg_file_path,
    GVariant *arg_other_mount_spec,
    const gchar *arg_other_file_path,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Changed",
    g_variant_new ("(u@(aya{sv})^ay@(aya{sv})^ay)",
                   arg_event_type,
                   arg_mount_spec,
                   arg_file_path,
                   arg_other_mount_spec,
                   arg_other_file_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * gvfs_dbus_monitor_client_call_changed_finish:
 * @proxy: A #GVfsDBusMonitorClientProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_monitor_client_call_changed().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with gvfs_dbus_monitor_client_call_changed().
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_monitor_client_call_changed_finish (
    GVfsDBusMonitorClient *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_monitor_client_call_changed_sync:
 * @proxy: A #GVfsDBusMonitorClientProxy.
 * @arg_event_type: Argument to pass with the method invocation.
 * @arg_mount_spec: Argument to pass with the method invocation.
 * @arg_file_path: Argument to pass with the method invocation.
 * @arg_other_mount_spec: Argument to pass with the method invocation.
 * @arg_other_file_path: Argument to pass with the method invocation.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-vfs-MonitorClient.Changed">Changed()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_monitor_client_call_changed() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeded, %FALSE if @error is set.
 */
gboolean
gvfs_dbus_monitor_client_call_changed_sync (
    GVfsDBusMonitorClient *proxy,
    guint arg_event_type,
    GVariant *arg_mount_spec,
    const gchar *arg_file_path,
    GVariant *arg_other_mount_spec,
    const gchar *arg_other_file_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Changed",
    g_variant_new ("(u@(aya{sv})^ay@(aya{sv})^ay)",
                   arg_event_type,
                   arg_mount_spec,
                   arg_file_path,
                   arg_other_mount_spec,
                   arg_other_file_path),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * gvfs_dbus_monitor_client_complete_changed:
 * @object: A #GVfsDBusMonitorClient.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-vfs-MonitorClient.Changed">Changed()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
gvfs_dbus_monitor_client_complete_changed (
    GVfsDBusMonitorClient *object,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusMonitorClientProxy:
 *
 * The #GVfsDBusMonitorClientProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusMonitorClientProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusMonitorClientProxy.
 */

struct _GVfsDBusMonitorClientProxyPrivate
{
  GData *qdata;
};

static void gvfs_dbus_monitor_client_proxy_iface_init (GVfsDBusMonitorClientIface *iface);

G_DEFINE_TYPE_WITH_CODE (GVfsDBusMonitorClientProxy, gvfs_dbus_monitor_client_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (GVfsDBusMonitorClientProxy)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_MONITOR_CLIENT, gvfs_dbus_monitor_client_proxy_iface_init));

static void
gvfs_dbus_monitor_client_proxy_finalize (GObject *object)
{
  GVfsDBusMonitorClientProxy *proxy = GVFS_DBUS_MONITOR_CLIENT_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (gvfs_dbus_monitor_client_proxy_parent_class)->finalize (object);
}

static void
gvfs_dbus_monitor_client_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_monitor_client_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
gvfs_dbus_monitor_client_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_gvfs_dbus_monitor_client_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], GVFS_DBUS_TYPE_MONITOR_CLIENT);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_MONITOR_CLIENT);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
gvfs_dbus_monitor_client_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  GVfsDBusMonitorClientProxy *proxy = GVFS_DBUS_MONITOR_CLIENT_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_monitor_client_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_monitor_client_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
gvfs_dbus_monitor_client_proxy_init (GVfsDBusMonitorClientProxy *proxy)
{
  proxy->priv = gvfs_dbus_monitor_client_proxy_get_instance_private (proxy);
  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), gvfs_dbus_monitor_client_interface_info ());
}

static void
gvfs_dbus_monitor_client_proxy_class_init (GVfsDBusMonitorClientProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = gvfs_dbus_monitor_client_proxy_finalize;
  gobject_class->get_property = gvfs_dbus_monitor_client_proxy_get_property;
  gobject_class->set_property = gvfs_dbus_monitor_client_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = gvfs_dbus_monitor_client_proxy_g_signal;
  proxy_class->g_properties_changed = gvfs_dbus_monitor_client_proxy_g_properties_changed;

}

static void
gvfs_dbus_monitor_client_proxy_iface_init (GVfsDBusMonitorClientIface *iface)
{
}

/**
 * gvfs_dbus_monitor_client_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MonitorClient.top_of_page">org.gtk.vfs.MonitorClient</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_monitor_client_proxy_new_finish() to get the result of the operation.
 *
 * See gvfs_dbus_monitor_client_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_monitor_client_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_MONITOR_CLIENT_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.MonitorClient", NULL);
}

/**
 * gvfs_dbus_monitor_client_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_monitor_client_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_monitor_client_proxy_new().
 *
 * Returns: (transfer full) (type GVfsDBusMonitorClientProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMonitorClient *
gvfs_dbus_monitor_client_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_MONITOR_CLIENT (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_monitor_client_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (allow-none): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MonitorClient.top_of_page">org.gtk.vfs.MonitorClient</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_monitor_client_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusMonitorClientProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMonitorClient *
gvfs_dbus_monitor_client_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_MONITOR_CLIENT_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.MonitorClient", NULL);
  if (ret != NULL)
    return GVFS_DBUS_MONITOR_CLIENT (ret);
  else
    return NULL;
}


/**
 * gvfs_dbus_monitor_client_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like gvfs_dbus_monitor_client_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the <link linkend="g-main-context-push-thread-default">thread-default main loop</link> of the thread you are calling this method from.
 * You can then call gvfs_dbus_monitor_client_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See gvfs_dbus_monitor_client_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
gvfs_dbus_monitor_client_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (GVFS_DBUS_TYPE_MONITOR_CLIENT_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.MonitorClient", NULL);
}

/**
 * gvfs_dbus_monitor_client_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to gvfs_dbus_monitor_client_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with gvfs_dbus_monitor_client_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type GVfsDBusMonitorClientProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMonitorClient *
gvfs_dbus_monitor_client_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return GVFS_DBUS_MONITOR_CLIENT (ret);
  else
    return NULL;
}

/**
 * gvfs_dbus_monitor_client_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (allow-none): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like gvfs_dbus_monitor_client_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See gvfs_dbus_monitor_client_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type GVfsDBusMonitorClientProxy): The constructed proxy object or %NULL if @error is set.
 */
GVfsDBusMonitorClient *
gvfs_dbus_monitor_client_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (GVFS_DBUS_TYPE_MONITOR_CLIENT_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.vfs.MonitorClient", NULL);
  if (ret != NULL)
    return GVFS_DBUS_MONITOR_CLIENT (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * GVfsDBusMonitorClientSkeleton:
 *
 * The #GVfsDBusMonitorClientSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * GVfsDBusMonitorClientSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #GVfsDBusMonitorClientSkeleton.
 */

struct _GVfsDBusMonitorClientSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_gvfs_dbus_monitor_client_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  GVfsDBusMonitorClientSkeleton *skeleton = GVFS_DBUS_MONITOR_CLIENT_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  guint num_params;
  guint num_extra;
  guint n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], GVFS_DBUS_TYPE_MONITOR_CLIENT);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, GVFS_DBUS_TYPE_MONITOR_CLIENT);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_gvfs_dbus_monitor_client_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  GVfsDBusMonitorClientSkeleton *skeleton = GVFS_DBUS_MONITOR_CLIENT_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_monitor_client_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_gvfs_dbus_monitor_client_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  GVfsDBusMonitorClientSkeleton *skeleton = GVFS_DBUS_MONITOR_CLIENT_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_gvfs_dbus_monitor_client_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _gvfs_dbus_monitor_client_skeleton_vtable =
{
  _gvfs_dbus_monitor_client_skeleton_handle_method_call,
  _gvfs_dbus_monitor_client_skeleton_handle_get_property,
  _gvfs_dbus_monitor_client_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
gvfs_dbus_monitor_client_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return gvfs_dbus_monitor_client_interface_info ();
}

static GDBusInterfaceVTable *
gvfs_dbus_monitor_client_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_gvfs_dbus_monitor_client_skeleton_vtable;
}

static GVariant *
gvfs_dbus_monitor_client_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  GVfsDBusMonitorClientSkeleton *skeleton = GVFS_DBUS_MONITOR_CLIENT_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  if (_gvfs_dbus_monitor_client_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _gvfs_dbus_monitor_client_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _gvfs_dbus_monitor_client_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _gvfs_dbus_monitor_client_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.vfs.MonitorClient", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
gvfs_dbus_monitor_client_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
}

static void gvfs_dbus_monitor_client_skeleton_iface_init (GVfsDBusMonitorClientIface *iface);
G_DEFINE_TYPE_WITH_CODE (GVfsDBusMonitorClientSkeleton, gvfs_dbus_monitor_client_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (GVfsDBusMonitorClientSkeleton)
                         G_IMPLEMENT_INTERFACE (GVFS_DBUS_TYPE_MONITOR_CLIENT, gvfs_dbus_monitor_client_skeleton_iface_init));

static void
gvfs_dbus_monitor_client_skeleton_finalize (GObject *object)
{
  GVfsDBusMonitorClientSkeleton *skeleton = GVFS_DBUS_MONITOR_CLIENT_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (gvfs_dbus_monitor_client_skeleton_parent_class)->finalize (object);
}

static void
gvfs_dbus_monitor_client_skeleton_init (GVfsDBusMonitorClientSkeleton *skeleton)
{
  skeleton->priv = gvfs_dbus_monitor_client_skeleton_get_instance_private (skeleton);
  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
gvfs_dbus_monitor_client_skeleton_class_init (GVfsDBusMonitorClientSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = gvfs_dbus_monitor_client_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = gvfs_dbus_monitor_client_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = gvfs_dbus_monitor_client_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = gvfs_dbus_monitor_client_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = gvfs_dbus_monitor_client_skeleton_dbus_interface_get_vtable;
}

static void
gvfs_dbus_monitor_client_skeleton_iface_init (GVfsDBusMonitorClientIface *iface)
{
}

/**
 * gvfs_dbus_monitor_client_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gtk-vfs-MonitorClient.top_of_page">org.gtk.vfs.MonitorClient</link>.
 *
 * Returns: (transfer full) (type GVfsDBusMonitorClientSkeleton): The skeleton object.
 */
GVfsDBusMonitorClient *
gvfs_dbus_monitor_client_skeleton_new (void)
{
  return GVFS_DBUS_MONITOR_CLIENT (g_object_new (GVFS_DBUS_TYPE_MONITOR_CLIENT_SKELETON, NULL));
}

